//
// This file is part of the PlaJA code base (2019 - 2019).
//

#ifndef PLAJA_DESTINATION_H
#define PLAJA_DESTINATION_H

#include <list>
#include "../../search/states/forward_states.h"
#include "../using_parser.h"
#include "expression/forward_expression.h"
#include "iterators/ast_iterator.h"
#include "ast_element.h"
#include "commentable.h"
#include "forward_ast.h"

class Destination final: public AstElement, public Commentable {
private:
    std::string destinationLocation;
    VariableIndex_type destLocIndex; // variable index (w.r.t. location) of the underlying automaton
    PLAJA::integer destLocVal;

    // probability
    std::unique_ptr<Expression> probabilityExpression;
    FIELD_IF_COMMENT_PARSING(std::string probabilityComment;)

    // assignments
    std::vector<std::unique_ptr<Assignment>> assignments;
    std::vector<std::list<const Assignment*>> assignmentsPerSeq;

public:
    explicit Destination();
    ~Destination() override;
    DELETE_CONSTRUCTOR(Destination)

    /* construction */

    /**
     * No effect, if num_seq is smaller-equal get_number_assignments.
     * @param num_seq, the maximal number of required sequences.
     */
    [[maybe_unused]] void add_sequences(std::size_t num_seq);

    void reserve(std::size_t assignments_cap);

    void add_assignment(std::unique_ptr<Assignment>&& assignment);

    /* setter */

    inline void set_destinationLocation(std::string dest_loc) { destinationLocation = std::move(dest_loc); }

    inline void set_destLocIndex(VariableIndex_type dest_loc_index) { destLocIndex = dest_loc_index; }

    inline void set_destLocVal(PLAJA::integer dest_loc_val) { destLocVal = dest_loc_val; }

    void set_probabilityExpression(std::unique_ptr<Expression>&& prob_exp);

    inline void set_probabilityComment(const std::string& prob_comment) { SET_IF_COMMENT_PARSING(probabilityComment, prob_comment) }

    void set_assignment(std::unique_ptr<Assignment>&& assignment, std::size_t index);

    /* getter */

    [[nodiscard]] inline const std::string& get_destinationLocation() const { return destinationLocation; }

    [[nodiscard]] inline VariableIndex_type get_destLocIndex() const { return destLocIndex; }

    [[nodiscard]] inline PLAJA::integer get_destinationLocVal() const { return destLocVal; }

    [[nodiscard]] inline Expression* get_probabilityExpression() { return probabilityExpression.get(); }

    [[nodiscard]] inline const Expression* get_probabilityExpression() const { return probabilityExpression.get(); }

    [[nodiscard]] const std::string& get_probabilityComment() const { return GET_IF_COMMENT_PARSING(probabilityComment); }

    [[nodiscard]] inline std::size_t get_number_assignments() const { return assignments.size(); }

    [[nodiscard]] inline Assignment* get_assignment(std::size_t index) {
        PLAJA_ASSERT(index < assignments.size())
        return assignments[index].get();
    }

    [[nodiscard]] inline const Assignment* get_assignment(std::size_t index) const {
        PLAJA_ASSERT(index < assignments.size())
        return assignments[index].get();
    }

    [[nodiscard]] inline std::size_t get_number_sequences() const { return assignmentsPerSeq.size(); }

    [[nodiscard]] inline std::size_t get_number_assignments_per_seq(SequenceIndex_type index) const {
        PLAJA_ASSERT(index < assignmentsPerSeq.size())
        return assignmentsPerSeq[index].size();
    }

    [[nodiscard]] inline const std::list<const Assignment*>& get_assignments_per_seq(SequenceIndex_type seq_index) const {
        PLAJA_ASSERT(seq_index < assignmentsPerSeq.size())
        return assignmentsPerSeq[seq_index];
    }

    /* */

    /** Deep copy of a destination. */
    [[nodiscard]] std::unique_ptr<Destination> deepCopy() const;

    void evaluate(const StateBase* current, StateBase* target, SequenceIndex_type seq_index) const;

    /* override */
    void accept(AstVisitor* ast_visitor) override;
    void accept(AstVisitorConst* ast_visitor) const override;

    /* iterators */

    [[nodiscard]] inline AstConstIterator<Assignment> assignmentIterator() const { return AstConstIterator(assignments); }

    //

    class AssignmentIterator {
    private:
        Destination& destination;
        std::size_t index;

    public:
        explicit AssignmentIterator(Destination& destination_):
            destination(destination_)
            , index(0) {}

        inline void operator++() { ++index; }

        [[nodiscard]] inline bool end() const { return index == destination.get_number_assignments(); }

        [[nodiscard]] inline Assignment* operator()() const { return destination.get_assignment(index); }

        [[nodiscard]] inline Assignment* operator->() const { return destination.get_assignment(index); }

        [[nodiscard]] inline Assignment& operator*() const { return *destination.get_assignment(index); }

        inline void set(std::unique_ptr<Assignment>&& assignment) const { destination.set_assignment(std::move(assignment), index); }

    };

    [[nodiscard]] inline AssignmentIterator assignmentIterator() { return AssignmentIterator(*this); }

    //

    class AssignmentIteratorPerSequence {
        friend Destination;
    private:
        const std::list<const Assignment*>& assignments;
        std::list<const Assignment*>::const_iterator it;
        const std::list<const Assignment*>::const_iterator itEnd;

        explicit AssignmentIteratorPerSequence(const std::list<const Assignment*>& assignments_):
            assignments(assignments_)
            , it(assignments.cbegin())
            , itEnd(assignments.end()) {
        }

    public:
        inline void operator++() { ++it; }

        [[nodiscard]] inline bool end() const { return it == itEnd; }

        inline const Assignment* operator->() { return *it; }
    };

    [[nodiscard]] inline AssignmentIteratorPerSequence assignmentIterator(SequenceIndex_type seq_index) const {
        return AssignmentIteratorPerSequence(get_assignments_per_seq(seq_index));
    }
};

#endif //PLAJA_DESTINATION_H
