//
// This file is part of the PlaJA code base (2019 - 2019).
//

#ifndef PLAJA_EDGE_H
#define PLAJA_EDGE_H

#include "../using_parser.h"
#include "expression/forward_expression.h"
#include "iterators/ast_iterator.h"
#include "ast_element.h"
#include "commentable.h"
#include "forward_ast.h"

class Edge final: public AstElement, public Commentable {

private:
    EdgeID_type id; // id for this edge (under edges in an automaton); defaults to EDGE::nullEdge, which is invalid for edge instances
    std::string sourceLocation;
    VariableIndex_type srcLocIndex; // variable index (w.r.t. location) of the underlying automaton
    PLAJA::integer srcLocVal; // value of the location, i.e., the index in the corresponding automaton's location's range, -1 iff unspecified
    std::string actionLabel;
    ActionID_type actionID; // defaults to ACTION::silentAction

    // rate
    std::unique_ptr<Expression> rateExpression;
    FIELD_IF_COMMENT_PARSING(std::string rateComment;)

    // guard
    std::unique_ptr<Expression> guardExpression;
    FIELD_IF_COMMENT_PARSING(std::string guardComment;)

    // destinations
    std::vector<std::unique_ptr<Destination>> destinations;

public:
    explicit Edge();
    ~Edge() override;
    DELETE_CONSTRUCTOR(Edge)

    /* construction */

    void reserve(std::size_t dest_cap);

    void add_destination(std::unique_ptr<Destination>&& dest);

    /* setter */

    inline void set_id(EdgeID_type edge_id) { id = edge_id; }

    inline void set_sourceLocation(std::string src_loc) { sourceLocation = std::move(src_loc); }

    inline void set_srcLocIndex(VariableIndex_type src_loc_index) { srcLocIndex = src_loc_index; }

    inline void set_srcLocVal(PLAJA::integer src_loc_id) { srcLocVal = src_loc_id; }

    inline void set_action(std::string action) { actionLabel = std::move(action); }

    inline void set_actionID(ActionID_type action_id) { actionID = action_id; }

    void set_rateExpression(std::unique_ptr<Expression>&& rate_exp);

    inline void set_rateComment(const std::string& rate_comment) { SET_IF_COMMENT_PARSING(rateComment, rate_comment) }

    void set_guardExpression(std::unique_ptr<Expression>&& guard_exp);

    void set_guardComment(const std::string& guard_comment) { SET_IF_COMMENT_PARSING(guardComment, guard_comment) }

    void set_destination(std::unique_ptr<Destination>&& dest, std::size_t index);

    void set_destinations(std::vector<std::unique_ptr<Destination>>&& destinations);

    /* getter */

    [[nodiscard]] inline EdgeID_type get_id() const { return id; }

    [[nodiscard]] inline const std::string& get_sourceLocation() const { return sourceLocation; }

    [[nodiscard]] inline VariableIndex_type get_srcLocIndex() const { return srcLocIndex; }

    [[nodiscard]] inline PLAJA::integer get_srcLocVal() const { return srcLocVal; }

    [[nodiscard]] inline const std::string& get_action() const { return actionLabel; }

    [[nodiscard]] inline ActionID_type get_actionID() const { return actionID; }

    [[nodiscard]] inline Expression* get_rateExpression() { return rateExpression.get(); }

    [[nodiscard]] inline const Expression* get_rateExpression() const { return rateExpression.get(); }

    [[nodiscard]] inline const std::string& get_rateComment() const { return GET_IF_COMMENT_PARSING(rateComment); }

    [[nodiscard]] inline Expression* get_guardExpression() { return guardExpression.get(); }

    [[nodiscard]] inline const Expression* get_guardExpression() const { return guardExpression.get(); }

    [[nodiscard]] inline const std::string& get_guardComment() const { return GET_IF_COMMENT_PARSING(guardComment); }

    [[nodiscard]] inline std::size_t get_number_destinations() const { return destinations.size(); }

    [[nodiscard]] inline Destination* get_destination(std::size_t index) {
        PLAJA_ASSERT(index < destinations.size())
        return destinations[index].get();
    }

    [[nodiscard]] inline const Destination* get_destination(std::size_t index) const {
        PLAJA_ASSERT(index < destinations.size())
        return destinations[index].get();
    }

    [[nodiscard]] inline AstConstIterator<Destination> destinationIterator() const { return AstConstIterator(destinations); }

    [[nodiscard]] inline AstIterator<Destination> destinationIterator() { return AstIterator(destinations); }

    /** Deep copy of an edge. */
    [[nodiscard]] std::unique_ptr<Edge> deepCopy() const;

    /* override */
    void accept(AstVisitor* ast_visitor) override;
    void accept(AstVisitorConst* ast_visitor) const override;

};

#endif //PLAJA_EDGE_H
