//
// This file is part of the PlaJA code base (2019 - 2019).
//

#include "set_variable_index.h"
#include "../../exception/semantics_exception.h"
#include "../../search/information/model_information.h"
#include "../ast/expression/non_standard/variable_value_expression.h"
#include "../ast/expression/derivative_expression.h"
#include "../ast/expression/variable_expression.h"
#include "../ast/iterators/model_iterator.h"
#include "../ast/destination.h"
#include "../ast/edge.h"

SetVarIndexBase::SetVarIndexBase() = default;

SetVarIndexBase::~SetVarIndexBase() = default;

void SetVarIndexBase::visit(Destination* destination) {
    PLAJA_ASSERT(automaton_index >= 0)
    destination->set_destLocIndex(automaton_index); // set location index, note this assumes that automaton_index = location variable index
    AstVisitor::visit(destination);
}

void SetVarIndexBase::visit(Edge* edge) {
    PLAJA_ASSERT(automaton_index >= 0)
    edge->set_srcLocIndex(automaton_index); // set location index, note this assumes that automaton_index = location variable index
    AstVisitor::visit(edge);
}

void SetVarIndexBase::visit(LocationValueExpression* /*exp*/) { /* NOOP */ }

/**********************************************************************************************************************/

SetVarIndexByID::SetVarIndexByID(const std::vector<VariableIndex_type>& id_to_index, const ModelInformation* model_info):
    idToIndex(&id_to_index)
    , model(nullptr)
    , modelInfo(model_info) {}

SetVarIndexByID::SetVarIndexByID(const Model& model):
    idToIndex(nullptr)
    , model(&model)
    , modelInfo(&model.get_model_information()) {}

SetVarIndexByID::~SetVarIndexByID() = default;

void SetVarIndexByID::visit(Destination* destination) { SetVarIndexBase::visit(destination); }

void SetVarIndexByID::visit(Edge* edge) { SetVarIndexBase::visit(edge); }

void SetVarIndexByID::visit(LocationValueExpression* exp) { SetVarIndexBase::visit(exp); }

void SetVarIndexByID::visit(DerivativeExpression* exp) {
    if (idToIndex) { exp->set_index((*idToIndex)[exp->get_id()]); }
    else {
        PLAJA_ASSERT(model)
        exp->set_index(model->get_variable_index_by_id(exp->get_id()));
    }
}

void SetVarIndexByID::visit(VariableExpression* exp) {
    if (idToIndex) { exp->set_index((*idToIndex)[exp->get_id()]); }
    else {
        PLAJA_ASSERT(model)
        exp->set_index(model->get_variable_index_by_id(exp->get_id()));
    }
}

void SetVarIndexByID::visit(VariableValueExpression* exp) {
    AstVisitor::visit(exp);

    // check bounds if model info present
    if (modelInfo) {
        const VariableIndex_type var_index = exp->get_var()->get_variable_index();
        PLAJA_ASSERT(exp->get_val()->is_constant())
        if (not modelInfo->is_valid(var_index, exp->get_val()->evaluate_floating_const())) { throw SemanticsException(exp->to_string()); }
    }
}

void SetVarIndexByID::set_vars(const std::vector<VariableIndex_type>& id_to_index, AstElement* ast_el, const ModelInformation* model_info) {
    SetVarIndexByID visitor(id_to_index, model_info);
    ast_el->accept(&visitor);
}

void SetVarIndexByID::set_vars(const Model& model, AstElement& ast_el) {
    SetVarIndexByID visitor(model);
    ast_el.accept(&visitor);
}

/**********************************************************************************************************************/

SetVarIndexByIndex::SetVarIndexByIndex(const std::vector<int>& index_to_index):
    indexToIndex(index_to_index) {}

SetVarIndexByIndex::~SetVarIndexByIndex() = default;

void SetVarIndexByIndex::visit(Destination* destination) { SetVarIndexBase::visit(destination); }

void SetVarIndexByIndex::visit(Edge* edge) { SetVarIndexBase::visit(edge); }

void SetVarIndexByIndex::visit(LocationValueExpression* exp) { SetVarIndexBase::visit(exp); }

void SetVarIndexByIndex::visit(DerivativeExpression* exp) { exp->set_index(indexToIndex.at(exp->get_index())); }

void SetVarIndexByIndex::visit(VariableExpression* exp) { exp->set_index(indexToIndex.at(exp->get_index())); }

void SetVarIndexByIndex::set_vars(const std::vector<int>& index_to_index, AstElement* ast_el) {
    SetVarIndexByIndex visitor(index_to_index);
    ast_el->accept(&visitor);
}

