//
// This file is part of the PlaJA code base (2019 - 2020).
//

#include <algorithm>
#include "variables_of.h"
#include "../ast/expression/non_standard/location_value_expression.h"
#include "../ast/expression/non_standard/variable_value_expression.h"
#include "../ast/expression/array_access_expression.h"
#include "../ast/assignment.h"
#include "../ast/destination.h"
#include "../ast/edge.h"

VariablesOf::VariablesOf(bool ignore_lvalues):
    vars()
    , ignoreLValues(ignore_lvalues) {

}

VariablesOf::~VariablesOf() = default;

std::unordered_set<VariableID_type> VariablesOf::vars_of(const AstElement* ast_element, bool ignore_lvalues) {
    PLAJA_ASSERT(ast_element)
    VariablesOf vars_of_visitor(ignore_lvalues);
    ast_element->accept(&vars_of_visitor);
    return std::move(vars_of_visitor.vars);
}

// visits:

void VariablesOf::visit(const Assignment* assignment) {
    JANI_ASSERT(assignment->get_ref())

    if (ignoreLValues) {
        const auto* array_index = assignment->get_ref()->get_array_index();
        if (array_index) { array_index->accept(this); } // array index is an rValue ...
    } else { assignment->get_ref()->accept(this); }

    if (assignment->is_deterministic()) {
        JANI_ASSERT(assignment->get_value())
        assignment->get_value()->accept(this);
    } else {
        PLAJA_ASSERT(assignment->is_non_deterministic())
        JANI_ASSERT(assignment->get_lower_bound() and assignment->get_upper_bound())
        assignment->get_lower_bound()->accept(this);
        assignment->get_upper_bound()->accept(this);
    }
}

void VariablesOf::visit(const VariableExpression* exp) { vars.insert(exp->get_variable_id()); }

/**********************************************************************************************************************/

NonDetUpdatedVarsOf::NonDetUpdatedVarsOf() = default;

NonDetUpdatedVarsOf::~NonDetUpdatedVarsOf() = default;

std::unordered_set<VariableID_type> NonDetUpdatedVarsOf::vars_of(const AstElement* ast_element) {
    NonDetUpdatedVarsOf non_det_updated_vars_of;
    ast_element->accept(&non_det_updated_vars_of);
    return std::move(non_det_updated_vars_of.vars);
}

void NonDetUpdatedVarsOf::visit(const Assignment* assignment) {
    JANI_ASSERT(assignment->get_ref())
    if (assignment->is_non_deterministic()) { vars.insert(assignment->get_ref()->get_variable_id()); }
}

/**********************************************************************************************************************/

VariableIndexesOf::VariableIndexesOf(bool ignore_lvalues, bool ignore_locations):
    varIndexes()
    , ignoreLValues(ignore_lvalues)
    , ignoreLocations(ignore_locations) {
}

VariableIndexesOf::~VariableIndexesOf() = default;

std::unordered_set<VariableID_type> VariableIndexesOf::var_indexes_of(const AstElement* ast_element, bool ignore_lvalues, bool ignore_locations) {
    VariableIndexesOf var_indexes_of_visitor(ignore_lvalues, ignore_locations);
    ast_element->accept(&var_indexes_of_visitor);
    return std::move(var_indexes_of_visitor.varIndexes);
}

// visits:

void VariableIndexesOf::visit(const Assignment* assignment) {
    JANI_ASSERT(assignment->get_ref())

    if (ignoreLValues) {
        const auto* array_index = assignment->get_ref()->get_array_index();
        if (array_index) { array_index->accept(this); } // array index is an rValue ...
    } else { assignment->get_ref()->accept(this); }

    if (assignment->is_deterministic()) {
        JANI_ASSERT(assignment->get_value())
        assignment->get_value()->accept(this);
    } else {
        PLAJA_ASSERT(assignment->is_non_deterministic())
        JANI_ASSERT(assignment->get_lower_bound() and assignment->get_upper_bound())
        assignment->get_lower_bound()->accept(this);
        assignment->get_upper_bound()->accept(this);
    }
}

void VariableIndexesOf::visit(const Destination* dest) {
    AstVisitorConst::visit(dest);
    if (not ignoreLocations) { varIndexes.insert(dest->get_destLocIndex()); }
}

void VariableIndexesOf::visit(const Edge* edge) {
    AstVisitorConst::visit(edge);
    if (not ignoreLocations) { varIndexes.insert(edge->get_srcLocIndex()); }
}

void VariableIndexesOf::visit(const ArrayAccessExpression* exp) {
    PLAJA_ASSERT(exp->get_index()->is_constant()) // else: // TODO need to add all state indexes -> need model info
    varIndexes.insert(exp->get_variable_index());
}

void VariableIndexesOf::visit(const VariableExpression* exp) { varIndexes.insert(exp->get_variable_index()); }

// non-standard

void VariableIndexesOf::visit(const LocationValueExpression* exp) { if (not ignoreLocations) { varIndexes.insert(exp->get_location_index()); } }

/*********************************************************************************************************************/

// for external usage:

namespace VARIABLES_OF {

    std::unordered_set<VariableID_type> vars_of(const AstElement* ast_element, bool ignore_lvalues) { return VariablesOf::vars_of(ast_element, ignore_lvalues); }

    std::unordered_set<VariableID_type> vars_of(const Expression* expression, bool ignore_lvalues) { return VariablesOf::vars_of(expression, ignore_lvalues); }

    std::list<VariableID_type> list_vars_of(const AstElement* ast_element, bool ignore_lvalues) {
        auto vars = vars_of(ast_element, ignore_lvalues);
        return { vars.cbegin(), vars.cend() };
    }

    std::unordered_set<VariableID_type> non_det_updated_vars_of(const AstElement* ast_element) { return NonDetUpdatedVarsOf::vars_of(ast_element); }

    std::unordered_set<VariableIndex_type> state_indexes_of(const AstElement* ast_element, bool ignore_lvalues, bool ignore_locations) { return VariableIndexesOf::var_indexes_of(ast_element, ignore_lvalues, ignore_locations); }

    std::unordered_set<VariableIndex_type> state_indexes_of(const Expression* expression, bool ignore_lvalues, bool ignore_locations) { return VariableIndexesOf::var_indexes_of(expression, ignore_lvalues, ignore_locations); }

    bool contains(const AstElement* ast_element, const std::unordered_set<VariableID_type>& var_set, bool ignore_lvalues) {
        auto vars = VARIABLES_OF::vars_of(ast_element, ignore_lvalues);
        return std::any_of(var_set.cbegin(), var_set.cend(), [vars](VariableID_type var_id) { return vars.count(var_id); });
    }

    bool contains(const AstElement* ast_element, VariableID_type var_id, bool ignore_lvalues) { return VARIABLES_OF::vars_of(ast_element, ignore_lvalues).count(var_id); }

    bool contains_state_indexes(const AstElement* ast_element, const std::unordered_set<VariableIndex_type>& state_indexes_set, bool ignore_lvalues, bool ignore_locations) {
        auto state_indexes = VARIABLES_OF::state_indexes_of(ast_element, ignore_lvalues, ignore_locations);
        return std::any_of(state_indexes_set.cbegin(), state_indexes_set.cend(), [state_indexes](VariableIndex_type state_index) { return state_indexes.count(state_index); });
    }

    bool contains_state_indexes(const AstElement* ast_element, VariableIndex_type state_index, bool ignore_lvalues, bool ignore_locations) { return VARIABLES_OF::state_indexes_of(ast_element, ignore_lvalues, ignore_locations).count(state_index); }

}