//
// This file is part of the PlaJA code base (2019 - 2020).
//

#ifndef PLAJA_VARIABLES_OF_H
#define PLAJA_VARIABLES_OF_H

#include <list>
#include <unordered_set>
#include "ast_visitor_const.h"
#include "../../search/states/forward_states.h"
#include "../ast/expression/forward_expression.h"
#include "../using_parser.h"

class AstElement;

/**
 * Collect the (constant) variable ids occurring as r-value (and optionally l-value) in the ast element.
 */
class VariablesOf: public AstVisitorConst {

private:
    std::unordered_set<VariableID_type> vars;
    bool ignoreLValues;

    explicit VariablesOf(bool ignore_lvalues);
public:
    ~VariablesOf() override;
    DELETE_CONSTRUCTOR(VariablesOf)

    static std::unordered_set<VariableID_type> vars_of(const AstElement* ast_element, bool ignore_lvalues);

    // visits:
    void visit(const Assignment* assignment) override;
    void visit(const VariableExpression* exp) override;

};

/**********************************************************************************************************************/

class NonDetUpdatedVarsOf: public AstVisitorConst {

private:
    std::unordered_set<VariableID_type> vars;

    NonDetUpdatedVarsOf();
public:
    ~NonDetUpdatedVarsOf() override;
    DELETE_CONSTRUCTOR(NonDetUpdatedVarsOf)

    static std::unordered_set<VariableID_type> vars_of(const AstElement* ast_element);

    // visits:
    void visit(const Assignment* assignment) override;

};

/**********************************************************************************************************************/

/**
 * Collect the (constant) variable indexes occurring as r-value (and optionally l-value) in the ast element,
 * \ie, the indexes of value variables
 * but also those of array indexes occurring as constant-index accesses,
 * and optionally also location indexes.
 */
class VariableIndexesOf: public AstVisitorConst {

private:
    std::unordered_set<VariableIndex_type> varIndexes;
    bool ignoreLValues;
    bool ignoreLocations;

    VariableIndexesOf(bool ignore_lvalues, bool ignore_locations);
public:
    ~VariableIndexesOf() override;
    DELETE_CONSTRUCTOR(VariableIndexesOf)

    /**
     * May only be called on structures not including non-constant array-accesses.
     * @param ast_element
     * @param ignore_lvalues
     * @return
     */
    static std::unordered_set<VariableIndex_type> var_indexes_of(const AstElement* ast_element, bool ignore_lvalues, bool ignore_locations);

    // visits:
    void visit(const Assignment* assignment) override;
    void visit(const Destination* dest) override;
    void visit(const Edge* edge) override;
    void visit(const ArrayAccessExpression* exp) override;
    void visit(const VariableExpression* exp) override;

    // non-standard
    void visit(const LocationValueExpression* exp) override;
};

/**********************************************************************************************************************/

// for external usage:
namespace VARIABLES_OF {

    extern std::unordered_set<VariableID_type> vars_of(const AstElement* ast_element, bool ignore_lvalues);
    extern std::unordered_set<VariableID_type> vars_of(const Expression* expression, bool ignore_lvalues); // auxiliary
    extern std::list<VariableID_type> list_vars_of(const AstElement* ast_element, bool ignore_lvalues);

    extern std::unordered_set<VariableID_type> non_det_updated_vars_of(const AstElement* ast_element);

    extern std::unordered_set<VariableIndex_type> state_indexes_of(const AstElement* ast_element, bool ignore_lvalues, bool ignore_locations);
    extern std::unordered_set<VariableIndex_type> state_indexes_of(const Expression* expression, bool ignore_lvalues, bool ignore_locations);

    extern bool contains(const AstElement* ast_element, const std::unordered_set<VariableID_type>& var_set, bool ignore_lvalues);
    extern bool contains(const AstElement* ast_element, VariableID_type var_id, bool ignore_lvalues);
    extern bool contains_state_indexes(const AstElement* ast_element, const std::unordered_set<VariableIndex_type>& state_indexes_set, bool ignore_lvalues, bool ignore_locations);
    extern bool contains_state_indexes(const AstElement* ast_element, VariableIndex_type state_index, bool ignore_lvalues, bool ignore_locations);

}

#endif //PLAJA_VARIABLES_OF_H
