//
// This file is part of the PlaJA code base (2019 - 2020).
//

#include <fstream>
#include <torch/nn/modules/linear.h>
#include "neural_network.h"
#include "../../parser/nn_parser/neural_network.h"
#include "../../utils/utils.h"
#include "../../assertions.h"

TORCH_IN_PLAJA::NeuralNetwork::NeuralNetwork(std::vector<unsigned int> layers, std::vector<TORCH_IN_PLAJA::Floating> lower_bounds, std::vector<TORCH_IN_PLAJA::Floating> upper_bounds):
    torch::nn::Module()
    , layerSizes(std::move(layers))
    , lowerBounds(std::move(lower_bounds))
    , upperBounds(std::move(upper_bounds)) {

    auto num_layers = layerSizes.size();

    PLAJA_ASSERT(num_layers >= 3) // Input, output + at least one hidden layer.
    inputSize = layerSizes[0];
    outputSize = layerSizes[num_layers - 1];
    PLAJA_ASSERT(inputSize == lowerBounds.size())
    PLAJA_ASSERT(inputSize == upperBounds.size())

    /* Initialize forwarding functions. */
    for (std::size_t i = 0; i < num_layers - 1; ++i) {
        auto pre_layer_size = layerSizes[i];
        auto post_layer_size = layerSizes[i + 1];
        forwardingFuncs.emplace_back(register_module("forwarding_" + std::to_string(i) + "_" + std::to_string(i + 1), torch::nn::Linear(pre_layer_size, post_layer_size)));
    }

}

TORCH_IN_PLAJA::NeuralNetwork::~NeuralNetwork() = default;

/**********************************************************************************************************************/

torch::Tensor TORCH_IN_PLAJA::NeuralNetwork::forward(torch::Tensor x) {
    std::size_t output_forwarding = forwardingFuncs.size() - 1;
    for (std::size_t i = 0; i < output_forwarding; ++i) {
        x = torch::relu(forwardingFuncs[i]->forward(x));
    }
    PLAJA_ASSERT(!forwardingFuncs.empty())
    return forwardingFuncs[output_forwarding]->forward(x);
}

/**********************************************************************************************************************/

void TORCH_IN_PLAJA::NeuralNetwork::load_from_nnet(const std::string& file_name) {
    torch::NoGradGuard no_grad; // Simulates enclosing with:
    // torch::autograd::GradMode::set_enabled(false);
    // torch::autograd::GradMode::set_enabled(true);

    PLAJA::NeuralNetwork nn_nnet(file_name);
    PLAJA_ASSERT(nn_nnet.getNumLayers() == forwardingFuncs.size()) // input layer is not counted
    auto src_layer_index = 0;
    auto target_layer_index = 1;

    for (auto layer_fun: forwardingFuncs) {

        auto src_layer_size = nn_nnet.getLayerSize(src_layer_index);
        auto target_layer_size = nn_nnet.getLayerSize(target_layer_index);

        for (auto target_neuron_index = 0; target_neuron_index < target_layer_size; ++target_neuron_index) {

            /* Bias. */
            layer_fun->bias[target_neuron_index] = nn_nnet.getBias(target_layer_index, target_neuron_index);

            /* Weights. */
            for (auto src_neuron_index = 0; src_neuron_index < src_layer_size; ++src_neuron_index) {
                layer_fun->weight[target_neuron_index][src_neuron_index] = nn_nnet.getWeight(src_layer_index, src_neuron_index, target_neuron_index);
            }
        }

        ++src_layer_index;
        ++target_layer_index;
    }

}

/**********************************************************************************************************************/

namespace TORCH_IN_PLAJA_NEURAL_NETWORK {

    void print_weight(std::ofstream& nnet_file, const torch::Tensor& weight) {
        for (int64_t row_i = 0; row_i < weight.size(0); ++row_i) {
            for (int64_t col_i = 0; col_i < weight.size(1); ++col_i) {
                nnet_file << weight[row_i][col_i].item<TORCH_IN_PLAJA::Floating>() << PLAJA_UTILS::commaString;
            }
            nnet_file << PLAJA_UTILS::lineBreakString;
        }
    }

    void print_bias(std::ofstream& nnet_file, const torch::Tensor& bias) {
        for (int64_t i = 0; i < bias.size(0); ++i) {
            nnet_file << bias[i].item<TORCH_IN_PLAJA::Floating>() << PLAJA_UTILS::commaString << PLAJA_UTILS::lineBreakString;
        }
    }

}

void TORCH_IN_PLAJA::NeuralNetwork::to_nnet(const std::string& file_name) {
    std::ofstream nnet_file;
    nnet_file.open(file_name, std::ios::trunc);
    // nnet_file << "// The contents of this file are licensed under the Creative Commons" << PLAJA_UTILS::lineBreakString;
    // nnet_file << "// Attribution 4.0 International License: https://creativecommons.org/licenses/by/4.0/" << PLAJA_UTILS::lineBreakString;
    nnet_file << "// Neural Network File Format by Kyle Julian, Stanford 2016" << PLAJA_UTILS::lineBreakString;

    /* #layers (in terms of #forwarding), #inputs, #outputs, maximum layer size (including input layer; output layer anyway). */
    nnet_file << forwardingFuncs.size()
              << PLAJA_UTILS::commaString << inputSize
              << PLAJA_UTILS::commaString << outputSize
              << PLAJA_UTILS::commaString << *std::max_element(layerSizes.begin(), layerSizes.end()) << PLAJA_UTILS::commaString << PLAJA_UTILS::lineBreakString;
    /* Network layer sizes. */
    for (auto layer_size: layerSizes) { nnet_file << layer_size << PLAJA_UTILS::commaString; }
    nnet_file << PLAJA_UTILS::lineBreakString;
    /* Deprecated flag. */
    nnet_file << 0 << PLAJA_UTILS::commaString << PLAJA_UTILS::lineBreakString;

    /* Lower bounds per input. */
    if (lowerBounds.empty()) {
        for (std::size_t i = 0; i < inputSize; ++i) { nnet_file << -std::numeric_limits<double>::max() << PLAJA_UTILS::commaString; }
    } else {
        for (auto lower_bound: lowerBounds) { nnet_file << lower_bound << PLAJA_UTILS::commaString; }
    }
    nnet_file << PLAJA_UTILS::lineBreakString;

    /* Upper bounds per output. */
    if (lowerBounds.empty()) {
        for (std::size_t i = 0; i < inputSize; ++i) { nnet_file << std::numeric_limits<double>::max() << PLAJA_UTILS::commaString; }
    } else {
        for (auto upper_bound: upperBounds) { nnet_file << upper_bound << PLAJA_UTILS::commaString; }
    }
    nnet_file << PLAJA_UTILS::lineBreakString;

    /* Means (inputs + one entry for outputs). */
    for (std::size_t i = 0; i <= inputSize; ++i) { nnet_file << 0 << PLAJA_UTILS::commaString; }
    nnet_file << PLAJA_UTILS::lineBreakString;
    /* Ranges (inputs + one entry for outputs). */
    for (std::size_t i = 0; i <= inputSize; ++i) { nnet_file << 1 << PLAJA_UTILS::commaString; }
    nnet_file << PLAJA_UTILS::lineBreakString;

    /* Forwarding weights and biases. */
    nnet_file << std::fixed << std::setprecision(8);
    for (const auto& forwarding: forwardingFuncs) {
        TORCH_IN_PLAJA_NEURAL_NETWORK::print_weight(nnet_file, forwarding->weight);
        TORCH_IN_PLAJA_NEURAL_NETWORK::print_bias(nnet_file, forwarding->bias);
    }

}

void TORCH_IN_PLAJA::NeuralNetwork::dump() const {
    const std::size_t l = forwardingFuncs.size();
    for (std::size_t i = 0; i < l; ++i) {
        std::cout << "Forwarding " << i << " to " << (i + 1) << ": " << PLAJA_UTILS::lineBreakString;
        std::cout << forwardingFuncs[i]->weight << PLAJA_UTILS::lineBreakString;
        std::cout << forwardingFuncs[i]->bias << PLAJA_UTILS::lineBreakString;
    }
}

// Useful:
// neural_network.named_parameters();
// torch::Tensor tensor = torch::rand({2, 3});
// auto options = torch::TensorOptions()
//                .requires_grad(true);
