//
// This file is part of the PlaJA code base (2019 - 2021).
//

#include <cmath>
#include "ql_agent.h"
#include "../exception/plaja_exception.h"
#include "../parser/ast/expression/non_standard/objective_expression.h"
#include "../parser/ast/model.h"
#include "../search/factories/policy_learning/ql_options.h"
#include "../search/factories/configuration.h"
#include "../search/fd_adaptions/timer.h"
#include "../search/fd_adaptions/state.h"
#include "../search/information/jani2nnet/jani_2_nnet.h"
#include "../search/information/model_information.h"
#include "../search/information/property_information.h"
#include "../search/non_prob_search/initial_states_enumerator.h"
#include "../search/successor_generation/simulation_environment.h"
#include "../utils/rng.h"
#include "../globals.h"
#include "learning_statistics.h"
#include "policy_teacher.h"
#include "ql_agent_torch.h"
#include "slicing_window.h"

// TODO possible future configurations:
//  - non-negative goal potential,
//  - break cycles greedily,
//  - heuristic functions: PDB, monotonicity abstraction

namespace QL_AGENT {

    extern std::unique_ptr<SearchEngine> construct(const PLAJA::Configuration& config);

    std::unique_ptr<SearchEngine> construct(const PLAJA::Configuration& config) { return std::make_unique<QlAgent>(config); }

}

QlAgent::QlAgent(const PLAJA::Configuration& config):
    SearchEngine(config)
    , evaluationMode(config.is_flag_set(PLAJA_OPTION::evaluation_mode))
    , numEpisodes(config.get_int_option(PLAJA_OPTION::num_episodes))
    , maxLengthEpisode(config.get_int_option(PLAJA_OPTION::max_length_episode))
    , epsilonDecay(config.get_double_option(PLAJA_OPTION::epsilon_decay))
    , epsilonEnd(evaluationMode ? 0 : config.get_double_option(PLAJA_OPTION::epsilon_end))
    , goalReward(PLAJA_UTILS::cast_numeric<USING_POLICY_LEARNING::Reward_type>(config.get_double_option(PLAJA_OPTION::goal_reward)))
    , avoidReward(PLAJA_UTILS::cast_numeric<USING_POLICY_LEARNING::Reward_type>(config.get_double_option(PLAJA_OPTION::avoid_reward)))
    , stallingReward(PLAJA_UTILS::cast_numeric<USING_POLICY_LEARNING::Reward_type>(config.get_double_option(PLAJA_OPTION::stalling_reward)))
    , terminalReward(PLAJA_UTILS::cast_numeric<USING_POLICY_LEARNING::Reward_type>(config.get_double_option(PLAJA_OPTION::terminal_reward)))
    , terminateCycles(config.is_flag_set(PLAJA_OPTION::terminate_cycles))
    , saveNNet(nullptr)
    , saveBackup(config.has_value_option(PLAJA_OPTION::save_backup) ? new std::string(config.get_value_option_string(PLAJA_OPTION::save_backup)) : nullptr)
    , episodesCounter(0)
    , epsilon(evaluationMode ? 0 : config.get_double_option(PLAJA_OPTION::epsilon_start))
    , bestScore(config.get_double_option(PLAJA_OPTION::minimal_score))
    , learningStats(new LearningStatistics())
    , scoreWindow(evaluationMode ? nullptr : new SlicingWindow(config.get_int_option(PLAJA_OPTION::slicingSize)))
    , episodeLength(evaluationMode ? nullptr : new SlicingWindow(config.get_int_option(PLAJA_OPTION::slicingSize)))
    , groundDonePotential(config.get_bool_option(PLAJA_OPTION::groundDonePotential))
    , policyTeacher(nullptr)
    , teacherGreediness(0)
    , rng(PLAJA_GLOBAL::rng.get())
    , teacherEpisode(false)
    , cycleDetected(false) {
    learningStats->add_basic_stats(*searchStatistics);

    // disable csv generation in main engine
    if (intermediateStatsFile) { learningStatsFile = std::move(intermediateStatsFile); }

    // set goal and avoid expressions
    PLAJA_ASSERT(propertyInfo->get_learning_objective())
    PLAJA_ASSERT(propertyInfo->get_reach())
    objective = propertyInfo->get_learning_objective(); // objective is goal
    avoid = propertyInfo->get_reach(); // reach is avoid, i.e., the safety property to check

    // load NN interface
    jani2NNet = propertyInfo->get_nn_interface();
    PLAJA_ASSERT(jani2NNet)
    simulatorEnv = std::make_unique<SimulationEnvironment>(config, *model);
    PLAJA_GLOBAL::randomizeNonDetEval = true; // For simulation.
    for (OutputIndex_type index = 0; index < jani2NNet->get_num_output_features(); ++index) {
        learningStats->set_epsilon_greedy_selection(model->get_action_name(jani2NNet->get_output_label(index)));
    }
    if (config.is_flag_set(PLAJA_OPTION::save_nnet)) { saveNNet = &jani2NNet->get_nn_file(); }

    // start states
    // primary initial state
    startStates.push_back(simulatorEnv->get_initial_state().get_id_value());
    std::unordered_set<StateID_type> start_states;
    start_states.insert(startStates.back());
    searchStatistics->inc_attr_unsigned(PLAJA::StatsUnsigned::START_STATES);
    // additional start states
    const auto* start = propertyInfo->get_start();
    if (start) {
        for (const auto& start_state_values: InitialStatesEnumerator::compute_initial_states(*model, *start)) {
            const StateID_type start_id = simulatorEnv->get_state(start_state_values).get_id_value();
            if (start_states.insert(start_id).second) {
                startStates.push_back(start_id);
                searchStatistics->inc_attr_unsigned(PLAJA::StatsUnsigned::START_STATES);
            }
        }
        if (config.get_bool_option(PLAJA_OPTION::groundStartPotential)) { groundedPotential = std::move(start_states); }
    }
    if (evaluationMode) {
        scoreWindow = std::make_unique<SlicingWindow>(numEpisodes);
        episodeLength = std::make_unique<SlicingWindow>(numEpisodes);
    }

    // compute upper and lower bounds
    std::vector<TORCH_IN_PLAJA::Floating> lower_bounds;
    std::vector<TORCH_IN_PLAJA::Floating> upper_bounds;
    {
        const auto& model_info = model->get_model_information();
        const auto l = jani2NNet->get_num_input_features();
        lower_bounds.reserve(l);
        upper_bounds.reserve(l);
        for (std::size_t i = 0; i < l; ++i) {
            const VariableIndex_type state_index = jani2NNet->get_input_feature_index(i);
            lower_bounds.push_back(PLAJA_UTILS::cast_numeric<TORCH_IN_PLAJA::Floating>(model_info.get_lower_bound(state_index)));
            upper_bounds.push_back(PLAJA_UTILS::cast_numeric<TORCH_IN_PLAJA::Floating>(model_info.get_upper_bound(state_index)));
        }
    }

    // torch agent internal
    agent = std::make_unique<QlAgentTorch>(jani2NNet->get_layer_sizes(), lower_bounds, upper_bounds);

    // load existing model ...
    if (config.has_value_option(PLAJA_OPTION::load_nn)) {
        const auto& nn_to_load = config.get_value_option_string(PLAJA_OPTION::load_nn);
        if (PLAJA_UTILS::file_exists(nn_to_load)) { load_nn(nn_to_load, nn_to_load); }
        else {
            std::cout << "Warning: specified invalid path to load existing NN ... trying to load NN as specified by loaded interface ..." << std::endl;
            load_nn(propertyInfo->get_nn_interface()->get_torch_file(), propertyInfo->get_nn_interface()->get_nn_file()); // property_info to not use "save_external" path
        }
    } else if (evaluationMode) { load_nn(propertyInfo->get_nn_interface()->get_torch_file(), propertyInfo->get_nn_interface()->get_nn_file()); }

    // if "load_nn" (from different position) or "save_nnet" is set, save also in eval mode (if file non-existing)
    if (evaluationMode) {
        if (config.has_value_option(PLAJA_OPTION::load_nn) and not PLAJA_UTILS::file_exists(jani2NNet->get_torch_file())) { agent->save(&jani2NNet->get_torch_file(), nullptr, saveBackup.get()); }
        if (saveNNet) { agent->save(nullptr, &jani2NNet->get_nn_file(), saveBackup.get()); }
    }

    if (evaluationMode) { agent->eval(); } // set torch models to eval mode; however should have no impact

    /* Init teacher. */
    if (not evaluationMode) {
        PUSH_LAP(searchStatistics, PLAJA::StatsDouble::TimeForTeacher)
        policyTeacher = PolicyTeacher::construct_policy_teacher(config);
        if (policyTeacher) { teacherGreediness = std::min(1.0, config.get_double_option(PLAJA_OPTION::teacherGreediness)); }
        POP_LAP(searchStatistics, PLAJA::StatsDouble::TimeForTeacher)
    }

}

QlAgent::~QlAgent() = default;

// auxiliaries:

void QlAgent::load_nn(const std::string& torch_file, const std::string& nnet_file) {
    try { agent->load(torch_file); } // try to load PyTorch model ...
    catch (PlajaException& e) {
        PLAJA_LOG(e.what())
        agent->load_from_nnet(nnet_file); // try to load NNet ...
    }
}

void QlAgent::save_copy(const std::string& prefix) {
    auto torch_file = PLAJA_UTILS::join_path(PLAJA_UTILS::extract_directory_path(jani2NNet->get_torch_file()), { prefix + PLAJA_UTILS::extract_filename(jani2NNet->get_torch_file(), true) });

    if (saveNNet) {

        auto nnet_file = PLAJA_UTILS::join_path(PLAJA_UTILS::extract_directory_path(jani2NNet->get_nn_file()), { prefix + PLAJA_UTILS::extract_filename(jani2NNet->get_nn_file(), true) });
        agent->save(&torch_file, &nnet_file, saveBackup.get());

    } else {
        agent->save(&torch_file, nullptr, saveBackup.get());
    }

}

//

inline bool QlAgent::check_goal(const State& state) const {
    auto* goal = objective->get_goal();
    return goal and goal->evaluateInteger(&state);
}

inline bool QlAgent::check_avoid(const State& state) const {
    PLAJA_ASSERT(avoid)
    return avoid->evaluateInteger(&state);
}

inline USING_POLICY_LEARNING::Reward_type QlAgent::compute_step_reward(bool done) const {
    const bool is_terminal_or_stalling = not(nextState); // reward for stalling?

    USING_POLICY_LEARNING::Reward_type reward_sum = 0;

    // reward for step
    if (not is_terminal_or_stalling) { // there is no step reward when *stalling*
        reward_sum += static_cast<USING_POLICY_LEARNING::Reward_type>(objective->compute_step_reward(currentState.get(), nextState.get()));
    }

    const auto* goal_potential = evaluationMode ? nullptr : objective->get_goal_potential(); // no goal potential in eval mode
    if (goal_potential) {
        const bool ground_current_potential = groundedPotential.count(currentState->get_id_value());
        const bool ground_next_potential = is_terminal_or_stalling or (done and groundDonePotential) or groundedPotential.count(nextState->get_id_value());
        const auto potential_current_state = ground_current_potential ? 0 : goal_potential->evaluateFloating(currentState.get());
        const auto potential_next_state = ground_next_potential ? 0 : goal_potential->evaluateFloating(nextState.get());
        reward_sum += static_cast<USING_POLICY_LEARNING::Reward_type>(potential_next_state - potential_current_state);
    }

    return reward_sum;
}

// auxiliaries to interact in torch agent:

inline void QlAgent::set_current_state(const State& state) {
    currentState = state.to_ptr();
    USING_POLICY_LEARNING::NNState& current_nn_state = agent->currentState;
    const auto l = current_nn_state.size();
    PLAJA_ASSERT(l == jani2NNet->get_num_input_features())
    for (std::size_t i = 0; i < l; ++i) {
        current_nn_state[i] = PLAJA_UTILS::cast_numeric<TORCH_IN_PLAJA::Floating>(state[jani2NNet->get_input_feature_index(i)]);
    }
    agent->set_current_state_tensor();
}

inline void QlAgent::set_next_state(const State& state) {
    nextState = state.to_ptr();
    USING_POLICY_LEARNING::NNState& next_nn_state = agent->nextState;
    const auto l = next_nn_state.size();
    PLAJA_ASSERT(l == jani2NNet->get_num_input_features())
    for (std::size_t i = 0; i < l; ++i) {
        next_nn_state[i] = PLAJA_UTILS::cast_numeric<TORCH_IN_PLAJA::Floating>(state[jani2NNet->get_input_feature_index(i)]);
    }
}

inline void QlAgent::set_next_state_applicability() {
    // next state applicability
    if (jani2NNet->_do_applicability_filtering()) {
        std::list<USING_POLICY_LEARNING::NNAction_type>& next_state_app = agent->nextStateApplicability;
        next_state_app.clear();
        for (const ActionLabel_type action_label: cachedApplicableActions) {
            PLAJA_ASSERT(jani2NNet->is_learned(action_label))
            next_state_app.push_back(jani2NNet->get_output_index(action_label));
        }
    }
}

inline void QlAgent::set_next_to_current_state() {
    currentState = std::move(nextState);
    std::swap(agent->currentState, agent->nextState);
    agent->set_current_state_tensor();
}

inline void QlAgent::set_not_done(const State& state) {
    set_next_state(state);
    set_next_state_applicability();
    agent->set_not_done(compute_step_reward(false));
}

inline void QlAgent::set_stalling(bool terminal) {
    nextState = nullptr;
    agent->set_stalling(compute_step_reward(true) + (terminal ? terminalReward : stallingReward));
}

inline void QlAgent::set_goal(const State& state) {
    set_next_state(state);
    agent->set_goal(compute_step_reward(true) + goalReward);
}

inline void QlAgent::set_avoid(const State& state) {
    set_next_state(state);
    agent->set_avoid(compute_step_reward(true) + avoidReward);
}

//

ActionLabel_type QlAgent::act_epsilon_greedy() {
    PLAJA_ASSERT(not cachedApplicableActions.empty()) // stalling due to terminal state is expected to be handled in preceding simulation step

    searchStatistics->inc_attr_unsigned(PLAJA::StatsUnsigned::POLICY_ACTS);

    /* TEACHER GREEDY */
    if (teacherEpisode) {

        searchStatistics->inc_attr_unsigned(PLAJA::StatsUnsigned::TeacherActs);
        PUSH_LAP(searchStatistics, PLAJA::StatsDouble::TimeForTeacher)
        const auto teacher_action = policyTeacher->get_action(*currentState);
        POP_LAP(searchStatistics, PLAJA::StatsDouble::TimeForTeacher)

        if (teacher_action != ACTION::noneLabel) {
            PLAJA_ASSERT(jani2NNet->is_learned(teacher_action))
            agent->set_action(jani2NNet->get_output_index(teacher_action)); // Update action within agent.
            return teacher_action;
        }

        teacherEpisode = false; // Teacher failed once, so abort teacher usage for this episode.
        searchStatistics->inc_attr_unsigned(PLAJA::StatsUnsigned::TeacherActsUnsuccessful);

    }

    /* EPSILON GREEDY */
    if (rng->prob() < epsilon) {
        searchStatistics->inc_attr_unsigned(PLAJA::StatsUnsigned::EPSILON_GREEDY_ACTS);
        const auto& applicable_actions = cachedApplicableActions;
        if (applicable_actions.empty()) { return jani2NNet->get_output_label(agent->act()); } // (selecting inapplicable action)
        const ActionLabel_type random_action = applicable_actions[rng->index(applicable_actions.size())];
        PLAJA_ASSERT(jani2NNet->is_learned(random_action))
        agent->set_action(jani2NNet->get_output_index(random_action)); // update action within agent
        return random_action;
    }

    /* Applicability Filtering. */
    if (jani2NNet->_do_applicability_filtering()) {

        USING_POLICY_LEARNING::NNAction_type current_nn_action = agent->rank_actions(); // rank actions
        ActionLabel_type current_action_label = jani2NNet->get_output_label(current_nn_action);
        if (simulatorEnv->is_cached_applicable(current_action_label)) { return current_action_label; }

        searchStatistics->inc_attr_unsigned(PLAJA::StatsUnsigned::SUB_RANK_ACTS);

        // iterate DEC
        const auto num_ranks = jani2NNet->get_num_output_features();
        for (long int rank = 1; rank < num_ranks; ++rank) {
            current_nn_action = agent->retrieve_ranked_action(rank); // also updates current action on agent side
            current_action_label = jani2NNet->get_output_label(current_nn_action);
            if (simulatorEnv->is_cached_applicable(current_action_label)) { return current_action_label; }
        }

        // No applicable action, fall back to no-filtering (selecting inapplicable action) ...
        PLAJA_ASSERT(cachedApplicableActions.empty())
        PLAJA_ABORT // Currently not expected.

    }

    /* No Filtering. */
    return jani2NNet->get_output_label(agent->act());

}

std::unique_ptr<State> QlAgent::simulate(const State& state, ActionLabel_type action_label) {
    auto successor_state = simulatorEnv->compute_successor_if_applicable(state, action_label);
    learningStats->inc_epsilon_greedy_selection(model->get_action_name(action_label));
    searchStatistics->inc_attr_unsigned(PLAJA::StatsUnsigned::ACTION_LABELS);
    searchStatistics->inc_attr_unsigned(PLAJA::StatsUnsigned::EXPANDED_STATES);

    // termination checks
    if (not successor_state) {
        searchStatistics->inc_attr_unsigned(PLAJA::StatsUnsigned::DONE_STALLING);
        set_stalling(false);
        return successor_state;
    } else {
        searchStatistics->inc_attr_unsigned(PLAJA::StatsUnsigned::GENERATED_STATES);
        searchStatistics->inc_attr_unsigned(PLAJA::StatsUnsigned::ACTION_OPS);
        searchStatistics->inc_attr_unsigned(PLAJA::StatsUnsigned::TRANSITIONS);
    }
    if (check_goal(*successor_state)) {
        searchStatistics->inc_attr_unsigned(PLAJA::StatsUnsigned::GOAL_STATES);
        searchStatistics->inc_attr_unsigned(PLAJA::StatsUnsigned::DONE_GOAL);
        set_goal(*successor_state);
        return successor_state;
    }
    if (check_avoid(*successor_state)) {
        searchStatistics->inc_attr_unsigned(PLAJA::StatsUnsigned::DONE_AVOID);
        set_avoid(*successor_state);
        return successor_state;
    }

    // simulate until next choice point
    cachedApplicableActions = simulatorEnv->extract_applicable_actions(*successor_state, true);
    while (cachedApplicableActions.size() <= 1) {

        if (cachedApplicableActions.empty()) {
            searchStatistics->inc_attr_unsigned(PLAJA::StatsUnsigned::TERMINAL_STATES);
            searchStatistics->inc_attr_unsigned(PLAJA::StatsUnsigned::DONE_STALLING);
            set_stalling(true);
            return successor_state;
        }

        // step:
        const ActionLabel_type next_action = cachedApplicableActions[0];
        if (jani2NNet->is_learned(next_action)) { break; } // may happen on ipc benchmarks
        PLAJA_ASSERT(next_action == ACTION::silentAction)

        successor_state = simulatorEnv->compute_successor(*successor_state, next_action);
        searchStatistics->inc_attr_unsigned(PLAJA::StatsUnsigned::GENERATED_STATES);
        searchStatistics->inc_attr_unsigned(PLAJA::StatsUnsigned::ACTION_OPS);
        searchStatistics->inc_attr_unsigned(PLAJA::StatsUnsigned::TRANSITIONS);

        if (check_goal(*successor_state)) {
            searchStatistics->inc_attr_unsigned(PLAJA::StatsUnsigned::GOAL_STATES);
            searchStatistics->inc_attr_unsigned(PLAJA::StatsUnsigned::DONE_GOAL);
            set_goal(*successor_state);
            return successor_state;
        }

        if (check_avoid(*successor_state)) {
            searchStatistics->inc_attr_unsigned(PLAJA::StatsUnsigned::DONE_AVOID);
            set_avoid(*successor_state);
            return successor_state;
        }

        // applicable actions for next iteration
        cachedApplicableActions = simulatorEnv->extract_applicable_actions(*successor_state, true);
    }

    set_not_done(*successor_state);
    return successor_state;
}

std::unique_ptr<State> QlAgent::simulate_until_choice_point(const State& state) {
    std::unique_ptr<State> successor_state = state.to_ptr();

    // simulate until next choice point
    cachedApplicableActions = simulatorEnv->extract_applicable_actions(*successor_state, true);
    while (cachedApplicableActions.size() <= 1) {

        if (cachedApplicableActions.empty()) {
            searchStatistics->inc_attr_unsigned(PLAJA::StatsUnsigned::TERMINAL_STATES);
            return nullptr;
        }

        // step:
        const ActionLabel_type next_action = cachedApplicableActions[0];
        if (jani2NNet->is_learned(next_action)) { break; } // may happen on ipc benchmarks
        PLAJA_ASSERT(next_action == ACTION::silentAction)

        successor_state = simulatorEnv->compute_successor(*successor_state, next_action);
        searchStatistics->inc_attr_unsigned(PLAJA::StatsUnsigned::GENERATED_STATES);
        searchStatistics->inc_attr_unsigned(PLAJA::StatsUnsigned::ACTION_OPS);
        searchStatistics->inc_attr_unsigned(PLAJA::StatsUnsigned::TRANSITIONS);

        // termination checks
        if (check_goal(*successor_state)) {
            searchStatistics->inc_attr_unsigned(PLAJA::StatsUnsigned::GOAL_STATES);
            return nullptr;
        }
        if (check_avoid(*successor_state)) { return nullptr; }

        // applicable actions for next iteration
        cachedApplicableActions = simulatorEnv->extract_applicable_actions(*successor_state, true);
    }

    return successor_state;
}

State QlAgent::sample_start_state() {
    const StateID_type start_id = evaluationMode ? startStates[episodesCounter % startStates.size()] : rng->index<StateID_type>(startStates.size());
    //
    searchStatistics->inc_attr_unsigned(PLAJA::StatsUnsigned::GENERATED_STATES);
    return simulatorEnv->get_state(start_id);
}

//

SearchEngine::SearchStatus QlAgent::initialize() { return SearchStatus::IN_PROGRESS; }

SearchEngine::SearchStatus QlAgent::step() {
    if (episodesCounter >= numEpisodes) { return SearchStatus::SOLVED; }

    // sample start
    std::unique_ptr<State> current_state = simulate_until_choice_point(sample_start_state());
    // analyze result of simulate:
    if (not current_state) { // drawn start state leads to "done" without any decision
        static bool print_warning(true);
        PLAJA_LOG_IF(print_warning, PLAJA_UTILS::to_red_string("Terminal start state."))
        print_warning = false;
        return SearchStatus::IN_PROGRESS;
    } else { // found choice point -> count step as episode
        ++episodesCounter;
        searchStatistics->inc_attr_unsigned(PLAJA::StatsUnsigned::EPISODES);
    }

    // set current state in torch agent
    teacherEpisode = rng->prob() < teacherGreediness;
    cycleDetected = false;
    cachedTransitionsPerEpisode.clear();
    agent->set_not_done(0);
    set_current_state(*current_state);

    // run
    double score = 0;
    auto t = 0;
    for (; t < maxLengthEpisode and not(terminateCycles and cycleDetected) and not agent->_done(); ++t) {
        PLAJA_ASSERT(current_state)
        const ActionLabel_type action_label = act_epsilon_greedy();
        current_state = simulate(*current_state, action_label);
        if (cachedTransitionsPerEpisode.insert({ currentState->get_id_value(), action_label, nextState ? nextState->get_id_value() : STATE::none }).second) {
            if (not evaluationMode) { agent->step_learning(jani2NNet->_do_applicability_filtering()); }   // Only add new samples to buffer.
        } else {
            cycleDetected = true;
        }
        set_next_to_current_state();
        //
        score += agent->_current_reward() * std::pow(USING_POLICY_LEARNING::gamma, t);
    }
    // update stats:
    episodeLength->add(static_cast<double>(t));
    if (not agent->_done()) { searchStatistics->inc_attr_unsigned(PLAJA::StatsUnsigned::UNDONE); }
    if (cycleDetected) { searchStatistics->inc_attr_unsigned(PLAJA::StatsUnsigned::DETECTED_CYCLES); }

    // update epsilon
    scoreWindow->add(score * (1 - epsilon)); // adjust score wrt. epsilon, i.e., the larger epsilon the less "valuable" is the score
    if (not teacherEpisode) {
        epsilon = std::max(epsilonEnd, epsilonDecay * epsilon);
        searchStatistics->set_attr_double(PLAJA::StatsDouble::EPSILON, epsilon);
    }

    if (episodesCounter % scoreWindow->slicing_size() == 0 or episodesCounter >= numEpisodes) {
        PLAJA_LOG("Updating stats and backups...");

        // mean score
        auto mean_score = scoreWindow->average();
        auto average_length = episodeLength->average();
        searchStatistics->set_attr_double(PLAJA::StatsDouble::LAST_SCORE, mean_score);
        searchStatistics->set_attr_double(PLAJA::StatsDouble::LAST_AVERAGE_LENGTH, average_length);

        if (mean_score >= bestScore) {

            searchStatistics->set_attr_double(PLAJA::StatsDouble::BEST_SCORE, bestScore = mean_score); // update best score
            searchStatistics->set_attr_double(PLAJA::StatsDouble::AVERAGE_LENGTH_BEST_SCORE, average_length);
            if (not evaluationMode) { agent->save(&jani2NNet->get_torch_file(), saveNNet, saveBackup.get()); } // save NN

        } else if (not evaluationMode and (episodesCounter >= numEpisodes or is_almost_expired())) { // Final backup.
            save_copy("final_");
        } else if (not evaluationMode and episodesCounter % (scoreWindow->slicing_size() * 10) == 0) { // Tmp backup.
            save_copy("tmp_");
        }

        PLAJA_LOG(PLAJA_UTILS::string_f("Episode: %u, average length: %f, mean score: %f, best score: %f, epsilon: %f.", episodesCounter, average_length, mean_score, bestScore, epsilon))
        if (learningStatsFile) { generate_csv_during_search(*learningStatsFile); } // for learning curve
    }

    return SearchStatus::IN_PROGRESS;
}

// override:

void QlAgent::print_statistics() const {
    searchStatistics->print_statistics();
    learningStats->print_statistics();
}

void QlAgent::stats_to_csv(std::ofstream& file) const {
    searchStatistics->set_attr_unsigned(PLAJA::StatsUnsigned::EPISODES, episodesCounter); // work-around to order stats properly
    searchStatistics->stats_to_csv(file);
    learningStats->stats_to_csv(file);
}

void QlAgent::stat_names_to_csv(std::ofstream& file) const {
    searchStatistics->stat_names_to_csv(file);
    learningStats->stat_names_to_csv(file);
}
