//
// This file is part of the PlaJA code base (2019 - 2021).
//

#ifndef PLAJA_QL_AGENT_H
#define PLAJA_QL_AGENT_H

#include <vector>
#include <unordered_set>
#include "../parser/ast/expression/forward_expression.h"
#include "../search/fd_adaptions/search_engine.h"
#include "../search/information/jani2nnet/forward_jani_2_nnet.h"
#include "../search/information/jani2nnet/using_jani2nnet.h"
#include "../search/states/forward_states.h"
#include "../search/successor_generation/forward_successor_generation.h"
#include "../search/using_search.h"
#include "forward_policy_learning.h"
#include "transition_set.h"
#include "using_policy_learning.h"

class QlAgent: public SearchEngine {

private:
    /* Variable Hyperparameters. */
    const bool evaluationMode; // Run simulation without learning the NN, to evaluate an already learned NN.  // NOLINT(*-avoid-const-or-ref-data-members)
    unsigned int numEpisodes; // Non-cost due to evaluation mode.
    const unsigned int maxLengthEpisode; // NOLINT(*-avoid-const-or-ref-data-members)
    const double epsilonDecay; // NOLINT(*-avoid-const-or-ref-data-members)
    const double epsilonEnd; // NOLINT(*-avoid-const-or-ref-data-members)
    const USING_POLICY_LEARNING::Reward_type goalReward; // NOLINT(*-avoid-const-or-ref-data-members)
    const USING_POLICY_LEARNING::Reward_type avoidReward; // NOLINT(*-avoid-const-or-ref-data-members)
    const USING_POLICY_LEARNING::Reward_type stallingReward; // NOLINT(*-avoid-const-or-ref-data-members)
    const USING_POLICY_LEARNING::Reward_type terminalReward; // NOLINT(*-avoid-const-or-ref-data-members)
    const bool terminateCycles; // NOLINT(*-avoid-const-or-ref-data-members)
    //
    const std::string* saveNNet;
    std::unique_ptr<std::string> saveBackup; // Where/whether to *also* save NN structure.
    std::unique_ptr<std::string> learningStatsFile;

    /* Dynamic learning parameters & stats. */
    unsigned int episodesCounter;
    double epsilon;
    double bestScore;
    std::unique_ptr<LearningStatistics> learningStats; // Additional stats.
    std::unique_ptr<SlicingWindow> scoreWindow;
    std::unique_ptr<SlicingWindow> episodeLength;

    /* Learning task specification. */
    std::vector<StateID_type> startStates;
    const Jani2NNet* jani2NNet;
    const ObjectiveExpression* objective;
    const Expression* avoid;
    std::unordered_set<StateID_type> groundedPotential; // To check membership.
    bool groundDonePotential;

    /* Policy teacher. */
    std::unique_ptr<PolicyTeacher> policyTeacher;
    double teacherGreediness;

    /* Heart of the learning agent. */
    std::unique_ptr<SimulationEnvironment> simulatorEnv; //
    class RandomNumberGenerator* rng; //
    std::unique_ptr<QlAgentTorch> agent;

    /* Cache. */
    std::unique_ptr<State> currentState;
    std::unique_ptr<State> nextState;
    std::vector<ActionLabel_type> cachedApplicableActions;
    QL_AGENT::TransitionSet cachedTransitionsPerEpisode;
    bool teacherEpisode;
    bool cycleDetected;

    /* Auxiliaries. */
    [[nodiscard]] bool check_goal(const State& state) const;
    [[nodiscard]] bool check_avoid(const State& state) const;
    [[nodiscard]] USING_POLICY_LEARNING::Reward_type compute_step_reward(bool done) const;
    /* Auxiliaries to interact in torch agent. */
    void set_current_state(const State& state);
    void set_next_state(const State& state);
    void set_next_state_applicability();
    void set_next_to_current_state();
    void set_not_done(const State& state);
    void set_goal(const State& state);
    void set_avoid(const State& state);
    void set_stalling(bool terminal);

    ActionLabel_type act_epsilon_greedy();
    std::unique_ptr<State> simulate(const State& state, ActionLabel_type action_label); // True if reached terminal state.
    std::unique_ptr<State> simulate_until_choice_point(const State& state); // To reach first choice point starting from start state.
    State sample_start_state(); // To reach first choice point starting from start state.

    void load_nn(const std::string& torch_file, const std::string& nnet_file);
    void save_copy(const std::string& prefix);
    SearchStatus initialize() override;
    SearchStatus step() override;

    // override:
    void print_statistics() const override;
    void stats_to_csv(std::ofstream& file) const override;
    void stat_names_to_csv(std::ofstream& file) const override;

public:
    explicit QlAgent(const PLAJA::Configuration& config);
    ~QlAgent() override;
    DELETE_CONSTRUCTOR(QlAgent)

};

#endif //PLAJA_QL_AGENT_H
