//
// This file is part of the PlaJA code base (2019 - 2021).
//

#include <torch/optim/adam.h>
#include <torch/serialize.h>
#include "ql_agent_torch.h"
#include "../exception/constructor_exception.h"
#include "../utils/utils.h"
#include "neural_networks/neural_network.h"
#include "replay_buffer.h"

QlAgentTorch::QlAgentTorch(const std::vector<unsigned int>& layers, const std::vector<TORCH_IN_PLAJA::Floating>& lower_bounds, const std::vector<TORCH_IN_PLAJA::Floating>& upper_bounds):
    updateCounter(0)
    , currentActionRanking(new torch::Tensor())
    , currentAction(USING_POLICY_LEARNING::noneAction)
    , nextStateApplicability()
    , currentReward(0)
    , currentDone(false) {
    PLAJA_ASSERT(layers.size() >= 3) // input, output, at least one hidden layer
    const unsigned int input_size = layers[0];
    const unsigned int output_size = layers[layers.size() - 1];

    policyNN = std::make_shared<TORCH_IN_PLAJA::NeuralNetwork>(layers, lower_bounds, upper_bounds);
    targetNN = std::make_shared<TORCH_IN_PLAJA::NeuralNetwork>(layers, lower_bounds, upper_bounds);
    optimizer = std::make_unique<torch::optim::Adam>(policyNN->parameters(), USING_POLICY_LEARNING::learningRate); // TODO? non-constant LR, e.g., exponential decrease (see adjust_learning_rate)
    replayBuffer = std::make_unique<ReplayBuffer>(USING_POLICY_LEARNING::batchSize, USING_POLICY_LEARNING::bufferSize, input_size, output_size);

    currentStateTensor = std::make_unique<torch::Tensor>(torch::zeros(input_size));
    currentState = USING_POLICY_LEARNING::NNState(input_size);
    nextState = USING_POLICY_LEARNING::NNState(input_size);
}

QlAgentTorch::~QlAgentTorch() = default;

//

void QlAgentTorch::set_current_state_tensor() {
    const std::size_t state_size = currentState.size();
    PLAJA_ASSERT(currentStateTensor->size(0) == state_size)
    for (long int state_index = 0; state_index < state_size; ++state_index) {
        (*currentStateTensor)[state_index] = currentState[state_index];
    }
}

USING_POLICY_LEARNING::NNAction_type QlAgentTorch::act() {
    return currentAction = policyNN->forward(*currentStateTensor).argmax().item<USING_POLICY_LEARNING::NNAction_type>();
}

USING_POLICY_LEARNING::NNAction_type QlAgentTorch::rank_actions() {
    *currentActionRanking = std::get<1>(policyNN->forward(*currentStateTensor).sort(0, true));
    PLAJA_ASSERT(currentStateTensor->size(0) > 0)
    return currentAction = (*currentActionRanking)[0].item<USING_POLICY_LEARNING::NNAction_type>();
}

USING_POLICY_LEARNING::NNAction_type QlAgentTorch::retrieve_ranked_action(long int rank) {
    PLAJA_ASSERT(0 <= rank)
    PLAJA_ASSERT(rank < currentActionRanking->size(0))
    return currentAction = (*currentActionRanking)[rank].item<USING_POLICY_LEARNING::NNAction_type>();
}

void QlAgentTorch::step_learning(bool applicability_filtering) {
    replayBuffer->add_sample(currentState, nextState, currentAction, nextStateApplicability, currentReward, currentDone); // add the sample to the buffer
    ++updateCounter;
    updateCounter %= USING_POLICY_LEARNING::updateEvery; // set update counter

    // learn if requirements fulfilled
    if (updateCounter == 0 && replayBuffer->size() > USING_POLICY_LEARNING::batchSize) {
        replayBuffer->sample();
        learn(applicability_filtering);
    }
}

void QlAgentTorch::learn(bool applicability_filtering) {
    // Implementation of dqn algorithm
    torch::Tensor q_values_next_states;
    if (applicability_filtering) {
        auto values_next_states = targetNN->forward(replayBuffer->next_states_sample());
        auto min_value = std::get<0>(values_next_states.max(1)).min(); // suppress gradient required
        q_values_next_states = std::get<0>((values_next_states + ((min_value - 1 - values_next_states) * (1 - replayBuffer->next_state_applicability_sample()))).max(1)).detach(); // suppress gradient required;
    } else {
        q_values_next_states = std::get<0>(targetNN->forward(replayBuffer->next_states_sample()).max(1)).detach(); // suppress gradient required
    }

    auto targets = replayBuffer->rewards_sample() + (USING_POLICY_LEARNING::gamma * (q_values_next_states) * (1 - replayBuffer->done_sample()));
    auto q_values = policyNN->forward(replayBuffer->states_sample());

    const auto& actions_sample = replayBuffer->actions_sample();
    auto actions_view = actions_sample.view({ actions_sample.sizes()[0], 1 });
    auto predictions = torch::gather(q_values, 1, actions_view).view(targets.sizes()[0]);

    // loss calculation
    auto loss = torch::mse_loss(predictions, targets);

    // make backwards step (optimize)
    optimizer->zero_grad();
    loss.backward();
    // for (auto& param: policyNN->parameters()) { param.grad().data().clamp(-1, 1); } // TODO might be useful?
    optimizer->step();

    // perform a soft-update to the target network
    auto parameters = policyNN->parameters();
    auto parameters_target = targetNN->parameters();
    const std::size_t l = parameters.size();
    PLAJA_ASSERT(l == parameters_target.size())
    for (std::size_t i = 0; i < l; ++i) {
        parameters_target[i].data().copy_(USING_POLICY_LEARNING::tau * parameters[i].data() + (1.0 - USING_POLICY_LEARNING::tau) * parameters_target[i].data());
    }
}

// load and save

void QlAgentTorch::save(const std::string* pytorch_filename, const std::string* nnet_filename, const std::string* backup_path) const {
    if (pytorch_filename) { torch::save(policyNN, *pytorch_filename); }
    if (nnet_filename) { policyNN->to_nnet(*nnet_filename); }
    if (backup_path) {
        if (pytorch_filename) { torch::save(policyNN, PLAJA_UTILS::join_path(*backup_path, { PLAJA_UTILS::extract_filename(*pytorch_filename, true) })); }
        if (nnet_filename) { policyNN->to_nnet(PLAJA_UTILS::join_path(*backup_path, { PLAJA_UTILS::extract_filename(*nnet_filename, true) })); }
    }
}

void QlAgentTorch::load(const std::string& filename) {
    try {
        torch::load(policyNN, filename);
        torch::load(targetNN, filename);
    } catch (c10::Error& e) {
        throw ConstructorException("Parsing PyTorch model " + filename + " failed: " + e.what_without_backtrace());
    }
}

void QlAgentTorch::load_from_nnet(const std::string& filename) {
    std::cout << "Trying to load " << filename << " as NNet ..." << std::endl;
    policyNN->load_from_nnet(filename);
    targetNN->load_from_nnet(filename);
}

void QlAgentTorch::eval() {
    policyNN->eval(); // set to train mode, however using torch::nn::linear this should have no impact (?)
    targetNN->eval();
}

/***********************************************************************************************************************/

// TODO? * Maybe for decaying learning rate (cf. https://github.com/gandroz/rl-taxi/blob/main/pytorch/agent.py [September 2022])
/*
void QlAgentTorch::adjust_learning_rate(unsigned int episode) {
    auto lr_min = 0; // dummy
    auto delta = USING_POLICY_LEARNING::learningRate - lr_min;
    auto lr_decay_rate = 0; // dummy
    auto lr = lr_min + delta * std::exp(-episode / lr_decay_rate);
    for (auto& param: optimizer->param_groups()) {
        auto optimizer_opt = std::make_unique<torch::optim::OptimizerOptions>();
        optimizer_opt->set_lr(lr);
        param.set_options(std::move(optimizer_opt));
    }
}
 */
