//
// This file is part of the PlaJA code base (2019 - 2021).
//

#include "bmc.h"
#include "../../utils/utils.h"
#include "../../stats/stats_base.h"
#include "../factories/bmc/bmc_options.h"
#include "../factories/configuration.h"
#include "../fd_adaptions/timer.h"

namespace BOUNDED_MC {
    const std::unordered_map<std::string, BoundedMC::BMCMode> stringToBMCMode{ { "unsafe",           BoundedMC::UNSAFE }, // NOLINT(cert-err58-cpp)
                                                                               { "loop-free",        BoundedMC::LOOP_FREE },
                                                                               { "unsafe-loop-free", BoundedMC::UNSAFE_LOOP_FREE },
                                                                               { "loop-free-unsafe", BoundedMC::LOOP_FREE_UNSAFE } };

    std::string print_supported_bmc_modes() { return PLAJA::OptionParser::print_supported_options(stringToBMCMode); }
}

BoundedMC::BoundedMC(const PLAJA::Configuration& config):
    SearchEngine(config)
    , bmcMode(config.get_option(BOUNDED_MC::stringToBMCMode, PLAJA_OPTION::bmc_mode))
    , maxSteps(config.get_int_option(PLAJA_OPTION::max_steps))
    , findShortestPathOnly(config.get_bool_option(PLAJA_OPTION::find_shortest_path_only))
    , constrainNoStart(config.is_flag_set(PLAJA_OPTION::constrain_no_start))
    , constrainLoopFree(config.is_flag_set(PLAJA_OPTION::constrain_loop_free))
    , constrainNoReach(config.is_flag_set(PLAJA_OPTION::constrain_no_reach))
    , steps(0)
    , shortestUnsafePath(-1)
    , longestLoopFreePath(-1) {

    // stats
    searchStatistics->add_int_stats(PLAJA::StatsInt::BMC_STEPS, "Steps", steps);
    searchStatistics->add_int_stats(PLAJA::StatsInt::BMC_SHORTEST_UNSAFE_PATH, "ShortestUnsafePath", shortestUnsafePath);
    searchStatistics->add_double_stats(PLAJA::StatsDouble::BMC_TIME_SHORTEST_UNSAFE_PATH, "TimeShortestUnsafePath", 0.0);
    searchStatistics->add_int_stats(PLAJA::StatsInt::BMC_LONGEST_LOOP_FREE_PATH, "LongestLoopFreePath", longestLoopFreePath);
    searchStatistics->add_double_stats(PLAJA::StatsDouble::BMC_TIME_LONGEST_LOOP_FREE_PATH, "TimeLongestLoopFreePath", 0.0);

    // disable csv generation in main engine
    if (intermediateStatsFile) { bmcStatsFile = std::move(intermediateStatsFile); }

}

BoundedMC::~BoundedMC() = default;

//


SearchEngine::SearchStatus BoundedMC::initialize() {
    PLAJA_LOG(PLAJA_UTILS::string_f("Step: %i", steps))

    if (bmcMode != LOOP_FREE and check_start_smt()) {
        PLAJA_LOG("0-step reachability!");
        shortestUnsafePath = 0;
        searchStatistics->set_attr_int(PLAJA::StatsInt::BMC_SHORTEST_UNSAFE_PATH, shortestUnsafePath);
        searchStatistics->set_attr_double(PLAJA::StatsDouble::BMC_TIME_SHORTEST_UNSAFE_PATH, ((*timer)() - timeOffset));
        return findShortestPathOnly ? SearchEngine::SOLVED : SearchEngine::IN_PROGRESS;
    } else { return SearchEngine::IN_PROGRESS; }

}

SearchEngine::SearchStatus BoundedMC::step() {

    // increment step counter
    ++steps;
    searchStatistics->inc_attr_int(PLAJA::StatsInt::BMC_STEPS);

    // termination criterion
    if (steps > maxSteps) { return SearchEngine::SOLVED; }

    PLAJA_LOG(PLAJA_UTILS::string_f("Current search time: %f", ((*timer)() - timeOffset)))
    PLAJA_LOG(PLAJA_UTILS::string_f("Step: %i", steps))

    auto status = SearchStatus::IN_PROGRESS;

    switch (bmcMode) {
        case UNSAFE: {
            status = check_unsafe();
            break;
        }
        case LOOP_FREE: {
            status = check_loop_free();
            break;
        }
        case UNSAFE_LOOP_FREE: {
            status = check_unsafe();
            if (status == SearchStatus::IN_PROGRESS) { status = check_loop_free(); }
            break;
        }
        case LOOP_FREE_UNSAFE: {
            status = check_loop_free();
            if (status == SearchStatus::IN_PROGRESS) { status = check_unsafe(); } // if there is no loop-free path, there is no unsafe path (due to incremental checking)
            break;
        }
    }

    if (bmcStatsFile) { generate_csv_during_search(*bmcStatsFile); }

    return status;
}

SearchEngine::SearchStatus BoundedMC::check_loop_free() {

    if (not check_loop_free_smt(steps)) {
        PLAJA_LOG("No loop-free path exists!")
        if (longestLoopFreePath < 0) {
            longestLoopFreePath = steps - 1;
            searchStatistics->set_attr_int(PLAJA::StatsInt::BMC_LONGEST_LOOP_FREE_PATH, longestLoopFreePath);
            searchStatistics->set_attr_double(PLAJA::StatsDouble::BMC_TIME_LONGEST_LOOP_FREE_PATH, ((*timer)() - timeOffset));
            return findShortestPathOnly ? SearchEngine::SOLVED : SearchEngine::IN_PROGRESS;
        }
    } else { PLAJA_LOG("Loop free path exists!"); }

    return SearchStatus::IN_PROGRESS;

}

SearchEngine::SearchStatus BoundedMC::check_unsafe() {

    if (check_unsafe_smt(steps)) {
        PLAJA_LOG("Found unsafe path");
        if (shortestUnsafePath < 0) {
            shortestUnsafePath = steps;
            searchStatistics->set_attr_int(PLAJA::StatsInt::BMC_SHORTEST_UNSAFE_PATH, shortestUnsafePath);
            searchStatistics->set_attr_double(PLAJA::StatsDouble::BMC_TIME_SHORTEST_UNSAFE_PATH, ((*timer)() - timeOffset));
            return findShortestPathOnly ? SearchEngine::SOLVED : SearchEngine::IN_PROGRESS;
        }
    }

    return SearchStatus::IN_PROGRESS;

}
