//
// This file is part of the PlaJA code base (2019 - 2023).
//

#include "bmc_marabou.h"
#include "../factories/configuration.h"
#include "../smt_nn/model/model_marabou.h"
#include "../smt_nn/solver/smt_solver_marabou.h"
#include "../smt_nn/solver/statistics_smt_solver_marabou.h"

// TODO: Solution checker (path extraction) || order of add_to_query || incrementality (learned conflicts)?
// Maybe external per-action (operator) enumeration baseline.

BMCMarabou::BMCMarabou(const PLAJA::Configuration& config):
    BoundedMC(config)
    , solverMarabou(nullptr) {

    // prepare solver:
    MARABOU_IN_PLAJA::add_solver_stats(*searchStatistics);

    PLAJA::Configuration config_loc(config);
    config_loc.set_sharable(PLAJA::SharableKey::STATS, searchStatistics);
    config_loc.set_sharable_ptr(PLAJA::SharableKey::STATS, searchStatistics.get());

    modelMarabou = std::make_unique<ModelMarabou>(config_loc);
    solverMarabou = MARABOU_IN_PLAJA::SMT_SOLVER::construct(config_loc, *modelMarabou->make_query(false, false, 0));

}

BMCMarabou::~BMCMarabou() = default;

bool BMCMarabou::check_start_smt() const {

    /* Hack: For Marabou we check path for model's initial state individually, because rewriting the start disjunction may blow up. */
    /* Model init. */
    if (not modelMarabou->initial_state_is_subsumed_by_start()) {
        solverMarabou->push();
        modelMarabou->add_init(*solverMarabou);
        modelMarabou->add_reach(*solverMarabou, 0);
        const bool rlt_init = solverMarabou->check();
        solverMarabou->pop();
        if (rlt_init) { return true; }
    }

    /* Start. */
    solverMarabou->push();
    modelMarabou->add_start(*solverMarabou);
    modelMarabou->add_reach(*solverMarabou, 0);
    const bool rlt = solverMarabou->check();
    solverMarabou->pop();

    return rlt;

}

void BMCMarabou::prepare_solver(MARABOU_IN_PLAJA::SMTSolver& solver, unsigned int steps_) const {

    modelMarabou->generate_steps(steps_); // Prepare model.

    solver.set_query(modelMarabou->make_query(false, true, steps_ - 1));

    /* NN. */
    if (modelMarabou->has_nn()) {
        for (auto it = modelMarabou->iterate_path(); !it.end(); ++it) {
            modelMarabou->add_nn_to_query(solver._query(), it.step());
        }
    }

    /* Transition structure. */
    for (auto it = modelMarabou->iterate_path(); !it.end(); ++it) {
        modelMarabou->add_choice_point(solver, modelMarabou->has_nn(), it.step());
    }

    /* Optimizations possible due to incremental checks. */

    if (constrainNoStart) {
        for (auto it = modelMarabou->iterate_steps_from_to(1, steps_); !it.end(); ++it) {
            modelMarabou->exclude_start(solver, steps_);
        }
    }

    if (constrainNoReach) {
        for (auto it = modelMarabou->iterate_steps_from_to(0, steps_ - 1); !it.end(); ++it) {
            modelMarabou->exclude_reach(solver, it.step());
        }
    }
}

bool BMCMarabou::check_loop_free_smt(unsigned int steps_) const {
    prepare_solver(*solverMarabou, steps_);

    if (not modelMarabou->initial_state_is_subsumed_by_start()) {
        solverMarabou->push();
        modelMarabou->add_init(*solverMarabou);
        modelMarabou->add_loop_exclusion(*solverMarabou, steps_);
        const bool rlt_init = solverMarabou->check();
        solverMarabou->pop();
        if (rlt_init) { return true; }
    }

    /* Start. */
    solverMarabou->push();
    modelMarabou->add_start(*solverMarabou);
    modelMarabou->add_loop_exclusion(*solverMarabou, steps_);
    const bool rlt = solverMarabou->check();
    solverMarabou->pop();

    return rlt;
}

bool BMCMarabou::check_unsafe_smt(unsigned int steps_) const {
    prepare_solver(*solverMarabou, steps_);

    if (not modelMarabou->initial_state_is_subsumed_by_start()) {
        solverMarabou->push();
        modelMarabou->add_init(*solverMarabou);
        modelMarabou->add_reach(*solverMarabou, steps_);
        if (constrainLoopFree) { modelMarabou->add_loop_exclusion(*solverMarabou, steps_); }
        const bool rlt_init = solverMarabou->check();
        solverMarabou->pop();
        if (rlt_init) { return true; }
    }

    /* Start. */
    solverMarabou->push();
    modelMarabou->add_start(*solverMarabou);
    modelMarabou->add_reach(*solverMarabou, steps_);
    if (constrainLoopFree) { modelMarabou->add_loop_exclusion(*solverMarabou, steps_); }
    // String s("step_" + std::to_string(steps_) + "_new.lp");
    //  GlobalConfiguration::PRINT_BASE_QUERY_IN_GUROBI = &s;
    const bool rlt = solverMarabou->check();
    //  GlobalConfiguration::PRINT_BASE_QUERY_IN_GUROBI = nullptr;
    solverMarabou->pop();

    return rlt;
}
