//
// This file is part of the PlaJA code base (2019 - 2023).
//

#include "bmc_z3.h"
#include "../factories/configuration.h"
#include "../smt/model/model_z3.h"
#include "../smt/solver/smt_solver_z3.h"
#include "../smt/solver/statistics_smt_solver_z3.h"

BMC_Z3::BMC_Z3(const PLAJA::Configuration& config):
    BoundedMC(config)
    , modelZ3(nullptr) {

    PLAJA::Configuration config_loc(config);
    config_loc.set_flag(PLAJA_OPTION::nn_multi_step_support_z3, true);

    modelZ3 = std::make_unique<ModelZ3>(config_loc);
    modelZ3->generate_steps(maxSteps);

    Z3_IN_PLAJA::add_solver_stats(*searchStatistics);
}

BMC_Z3::~BMC_Z3() = default;

//

bool BMC_Z3::check_start_smt() const {
    Z3_IN_PLAJA::SMTSolver solver(modelZ3->share_context(), searchStatistics.get());
    solver.push(); // to suppress scoped_timer thread in z3, which is started when using tactic2solver

    modelZ3->add_bounds(solver, 0);
    modelZ3->add_start(solver, true);
    modelZ3->add_reach(solver, 0);

    return solver.check();
}

void BMC_Z3::prepare_solver(Z3_IN_PLAJA::SMTSolver& solver, unsigned int steps_) const {
    solver.push(); // to suppress scoped_timer thread in z3, which is started when using tactic2solver

    // state variable bounds
    for (auto it = modelZ3->iterate_steps_from_to(0, steps_); !it.end(); ++it) { modelZ3->add_bounds(solver, it.step()); }

    // start
    modelZ3->add_start(solver, true);

    // NN
    if (modelZ3->has_nn()) { for (auto it = modelZ3->iterate_steps_from_to(0, steps_ - 1); !it.end(); ++it) { modelZ3->add_nn(solver, it.step()); } }

    // transition structure
    for (auto it = modelZ3->iterate_steps_from_to(0, steps_ - 1); !it.end(); ++it) { modelZ3->add_choice_point(solver, it.step(), modelZ3->has_nn()); }

    // optimizations possible due to incremental checks
    if (constrainNoStart) {
        for (auto it = modelZ3->iterate_steps_from_to(1, steps_); !it.end(); ++it) { modelZ3->exclude_start(solver, it.step()); }
    }

    if (constrainNoReach) {
        for (auto it = modelZ3->iterate_steps_from_to(0, steps_ - 1); !it.end(); ++it) { modelZ3->exclude_reach(solver, it.step()); }
    }

}

bool BMC_Z3::check_loop_free_smt(unsigned int steps_) const {
    Z3_IN_PLAJA::SMTSolver solver(modelZ3->share_context(), searchStatistics.get());
    prepare_solver(solver, steps_);

    modelZ3->add_loop_exclusion(solver, steps_);

    return solver.check();
}

bool BMC_Z3::check_unsafe_smt(unsigned int steps_) const {
    Z3_IN_PLAJA::SMTSolver solver(modelZ3->share_context(), searchStatistics.get());
    prepare_solver(solver, steps_);

    if (constrainLoopFree) { modelZ3->add_loop_exclusion(solver, steps_); }
    modelZ3->add_reach(solver, steps_);

    return solver.check();
}