//
// This file is part of the PlaJA code base (2019 - 2021).
//

#include "bmc_factory.h"
#include "../../../include/factory_config_const.h"
#include "../../../option_parser/option_parser_aux.h"
#include "../../information/property_information.h"
#include "../configuration.h"
#include "bmc_options.h"

#ifdef USE_Z3

#include "../../bmc/bmc_z3.h"

#endif

#ifdef USE_MARABOU

#include "../../bmc/bmc_marabou.h"

#endif

namespace BMC_FACTORY {

    enum SolverType {
        Z3,
        MARABOU
    };

    const std::unordered_map<std::string, BMC_FACTORY::SolverType> stringToSolverType { // NOLINT(cert-err58-cpp)
#ifdef USE_Z3
        { "z3", SolverType::Z3 },
#endif
#ifdef USE_MARABOU
        { "marabou", SolverType::MARABOU },
#endif
    };

    std::string print_supported_solver_types() { return PLAJA::OptionParser::print_supported_options(stringToSolverType); }

}

BMCFactory::BMCFactory():
    SMTBaseFactory(SearchEngineFactory::BmcType) {}

BMCFactory::~BMCFactory() = default;

void BMCFactory::add_options(PLAJA::OptionParser& option_parser) const {
    SMTBaseFactory::add_options(option_parser);
    //
    OPTION_PARSER::add_option(option_parser, PLAJA_OPTION::bmc_mode, PLAJA_OPTION_DEFAULT::bmc_mode);
    OPTION_PARSER::add_int_option(option_parser, PLAJA_OPTION::max_steps, PLAJA_OPTION_DEFAULT::max_steps);
    OPTION_PARSER::add_bool_option(option_parser, PLAJA_OPTION::find_shortest_path_only, PLAJA_OPTION_DEFAULT::find_shortest_path_only);
    OPTION_PARSER::add_flag(option_parser, PLAJA_OPTION::constrain_no_start);
    OPTION_PARSER::add_flag(option_parser, PLAJA_OPTION::constrain_loop_free);
    OPTION_PARSER::add_flag(option_parser, PLAJA_OPTION::constrain_no_reach);
    OPTION_PARSER::add_option(option_parser, PLAJA_OPTION::bmc_solver, PLAJA_OPTION_DEFAULT::bmc_solver);
}

void BMCFactory::print_options() const {
    OPTION_PARSER::print_options_headline("BMC");
    OPTION_PARSER::print_option(PLAJA_OPTION::bmc_mode, OPTION_PARSER::valueStr, PLAJA_OPTION_DEFAULT::bmc_mode, "BMC mode to be used:", true);
    OPTION_PARSER::print_additional_specification(BOUNDED_MC::print_supported_bmc_modes());
    OPTION_PARSER::print_int_option(PLAJA_OPTION::max_steps, PLAJA_OPTION_DEFAULT::max_steps, "Maximal number of steps to be unrolled.");
    OPTION_PARSER::print_bool_option(PLAJA_OPTION::find_shortest_path_only, PLAJA_OPTION_DEFAULT::find_shortest_path_only, "Terminate after finding the shortest unsafe path.");
    OPTION_PARSER::print_flag(PLAJA_OPTION::constrain_no_start, "Constrain non-start states to be non-start.");
    OPTION_PARSER::print_flag(PLAJA_OPTION::constrain_loop_free, "Constrain (unsafe) path to be loop free.");
    OPTION_PARSER::print_flag(PLAJA_OPTION::constrain_no_reach, "Constrain non-terminal states to be non-reach.");
    OPTION_PARSER::print_option(PLAJA_OPTION::bmc_solver, OPTION_PARSER::valueStr, PLAJA_OPTION_DEFAULT::bmc_solver, "SMT-solver used for BMC.", true);
    OPTION_PARSER::print_additional_specification(BMC_FACTORY::print_supported_solver_types());
    SMTBaseFactory::print_options_base();
    OPTION_PARSER::print_options_footline();
}

std::unique_ptr<SearchEngine> BMCFactory::construct(const PLAJA::Configuration& config) const {

    switch (config.get_option(BMC_FACTORY::stringToSolverType, PLAJA_OPTION::bmc_solver)) {

#ifdef USE_Z3
        case BMC_FACTORY::Z3: { return std::make_unique<BMC_Z3>(config); }
#endif

#ifdef USE_MARABOU
        case BMC_FACTORY::MARABOU: { return std::make_unique<BMCMarabou>(config); }
#endif

        default: PLAJA_ABORT

    }

    PLAJA_ABORT
}

bool BMCFactory::supports_property_sub(const PropertyInformation& prop_info) const { return prop_info.get_reach(); }
