//
// This file is part of the PlaJA code base (2019 - 2021).
//

#include "ql_factory.h"
#include "../../../include/factory_config_const.h"
#include "../../../option_parser/option_parser_aux.h"
#include "../../fd_adaptions/search_engine.h"
#include "../../information/property_information.h"
#include "../../non_prob_search/restrictions_checker_explicit.h"
#include "../configuration.h"
#include "ql_options.h"

#ifdef USE_Z3
namespace PLAJA_OPTION::SMT_SOLVER_Z3 {
    extern void add_options(PLAJA::OptionParser& option_parser);
    extern void print_options();
}
#endif

// #include "torch/utils.h"
namespace at {
    extern void set_num_threads(int);
    extern void set_num_interop_threads(int);
}

namespace torch {
    using at::set_num_threads;
    using at::set_num_interop_threads;
}

namespace QL_AGENT { extern std::unique_ptr<SearchEngine> construct(const PLAJA::Configuration& config); }

/**********************************************************************************************************************/

QlFactory::QlFactory():
    SearchEngineFactory(SearchEngineType::QlAgentType) {}

QlFactory::~QlFactory() = default;

void QlFactory::add_options(PLAJA::OptionParser& option_parser) const {
    SearchEngineFactory::add_options(option_parser);
    //
    OPTION_PARSER::add_flag(option_parser, PLAJA_OPTION::evaluation_mode);
    OPTION_PARSER::add_value_option(option_parser, PLAJA_OPTION::load_nn);
    OPTION_PARSER::add_value_option(option_parser, PLAJA_OPTION::save_backup);
    OPTION_PARSER::add_flag(option_parser, PLAJA_OPTION::save_nnet);
    OPTION_PARSER::add_int_option(option_parser, PLAJA_OPTION::num_episodes, PLAJA_OPTION_DEFAULT::numEpisodes);
    OPTION_PARSER::add_int_option(option_parser, PLAJA_OPTION::max_length_episode, PLAJA_OPTION_DEFAULT::maxLengthEpisode);
    OPTION_PARSER::add_flag(option_parser, PLAJA_OPTION::terminate_cycles);
    OPTION_PARSER::add_double_option(option_parser, PLAJA_OPTION::epsilon_start, PLAJA_OPTION_DEFAULT::epsilonStart);
    OPTION_PARSER::add_double_option(option_parser, PLAJA_OPTION::epsilon_decay, PLAJA_OPTION_DEFAULT::epsilonDecay);
    OPTION_PARSER::add_double_option(option_parser, PLAJA_OPTION::epsilon_end, PLAJA_OPTION_DEFAULT::epsilonEnd);
    OPTION_PARSER::add_double_option(option_parser, PLAJA_OPTION::goal_reward, PLAJA_OPTION_DEFAULT::goalReward);
    OPTION_PARSER::add_double_option(option_parser, PLAJA_OPTION::avoid_reward, PLAJA_OPTION_DEFAULT::avoidReward);
    OPTION_PARSER::add_double_option(option_parser, PLAJA_OPTION::stalling_reward, PLAJA_OPTION_DEFAULT::stallingReward);
    OPTION_PARSER::add_double_option(option_parser, PLAJA_OPTION::terminal_reward, PLAJA_OPTION_DEFAULT::terminalReward);
    OPTION_PARSER::add_bool_option(option_parser, PLAJA_OPTION::groundStartPotential, PLAJA_OPTION_DEFAULT::groundStartPotential);
    OPTION_PARSER::add_bool_option(option_parser, PLAJA_OPTION::groundDonePotential, PLAJA_OPTION_DEFAULT::groundDonePotential);
    OPTION_PARSER::add_double_option(option_parser, PLAJA_OPTION::minimal_score, PLAJA_OPTION_DEFAULT::minimalScore);
    OPTION_PARSER::add_int_option(option_parser, PLAJA_OPTION::slicingSize, PLAJA_OPTION_DEFAULT::slicingSize);

    /* Policy teacher. */
    OPTION_PARSER::add_value_option(option_parser, PLAJA_OPTION::teacherModel);
    OPTION_PARSER::add_double_option(option_parser, PLAJA_OPTION::teacherGreediness, PLAJA_OPTION_DEFAULT::teacherGreediness);
    OPTION_PARSER::add_bool_option(option_parser, PLAJA_OPTION::teacherExplore, PLAJA_OPTION_DEFAULT::teacherExplore);

#ifdef USE_Z3
    PLAJA_OPTION::SMT_SOLVER_Z3::add_options(option_parser);
#endif
}

void QlFactory::print_options() const {
    OPTION_PARSER::print_options_headline("Policy Learning");

    OPTION_PARSER::print_flag(PLAJA_OPTION::evaluation_mode, "Perform evaluation for existing NN (without updating/learning).");

    OPTION_PARSER::print_option(PLAJA_OPTION::load_nn, OPTION_PARSER::fileStr, "Load existing PyTorch model.", true);
    OPTION_PARSER::print_additional_specification("If that fails, try to load as NNet model.", true);
    OPTION_PARSER::print_additional_specification("If " + OPTION_PARSER::fileStr + " is invalid, load models specified by the interface (if possible).");

    OPTION_PARSER::print_option(PLAJA_OPTION::save_backup, OPTION_PARSER::dirStr, "Save additional backup of learned network to external/customized directory.");

    OPTION_PARSER::print_flag(PLAJA_OPTION::save_nnet, "Save NN also in nnet format.");

    OPTION_PARSER::print_int_option(PLAJA_OPTION::num_episodes, PLAJA_OPTION_DEFAULT::numEpisodes, "Number of episodes to train.");

    OPTION_PARSER::print_int_option(PLAJA_OPTION::max_length_episode, PLAJA_OPTION_DEFAULT::maxLengthEpisode, "Maximal length of a training episode.");

    OPTION_PARSER::print_flag(PLAJA_OPTION::terminate_cycles, "Terminate cycle episodes.");

    const std::string hyper_parameters = PLAJA_OPTION::epsilon_start + "|" + PLAJA_OPTION::epsilon_decay + "|" + PLAJA_OPTION::epsilon_end;
    OPTION_PARSER::print_option(hyper_parameters, OPTION_PARSER::valueStr, "Hyperparameters of epsilon-greed.");

    const std::string reward_values = PLAJA_OPTION::goal_reward + "|" + PLAJA_OPTION::avoid_reward + "|" + PLAJA_OPTION::stalling_reward + "|" + PLAJA_OPTION::terminal_reward;
    OPTION_PARSER::print_option(reward_values, OPTION_PARSER::valueStr, "Reward values.");

    const std::string ground_potential_values = PLAJA_OPTION::groundStartPotential + "|" + PLAJA_OPTION::groundDonePotential;
    OPTION_PARSER::print_option(ground_potential_values, OPTION_PARSER::valueStr, "Ground reward-shaping potential of respective states");

    OPTION_PARSER::print_double_option(PLAJA_OPTION::minimal_score, PLAJA_OPTION_DEFAULT::minimalScore, "Minimal learning score for saving.");

    OPTION_PARSER::print_int_option(PLAJA_OPTION::slicingSize, PLAJA_OPTION_DEFAULT::slicingSize, "Size of the sliding window to compute average.");

    OPTION_PARSER::print_option(PLAJA_OPTION::teacherModel, OPTION_PARSER::fileStr, "Load model used by a teacher guiding the trained policy.");

    OPTION_PARSER::print_double_option(PLAJA_OPTION::teacherGreediness, PLAJA_OPTION_DEFAULT::teacherGreediness, "Fraction of *non*-epsilon-greediness used for policy.");

    OPTION_PARSER::print_bool_option(PLAJA_OPTION::teacherExplore, PLAJA_OPTION_DEFAULT::teacherExplore, "Teacher explores state space of teacher model to compute goal path.");

#ifdef USE_Z3
    PLAJA_OPTION::SMT_SOLVER_Z3::print_options();
#endif

    OPTION_PARSER::print_options_footline();
}

std::unique_ptr<SearchEngine> QlFactory::construct(const PLAJA::Configuration& config) const {
    torch::set_num_threads(1);
    torch::set_num_interop_threads(1);
    return QL_AGENT::construct(config);
}

bool QlFactory::supports_property_sub(const PropertyInformation& prop_info) const { return prop_info.get_reach() and prop_info.get_learning_objective(); }

/**********************************************************************************************************************/

namespace SEARCH_ENGINE_FACTORY {

    std::unique_ptr<SearchEngineFactory> construct_ql_agent_factory() { return std::make_unique<QlFactory>(); }

}