//
// This file is part of the PlaJA code base (2019 - 2020).
//

#include "lrtdp_factory.h"
#include "../../../option_parser/option_parser.h"
#include "../../information/property_information.h"
#include "../../prob_search/lrtdp.h"
#include "../configuration.h"

// extern:
namespace PLAJA_OPTION { extern const std::string upper_and_lower_bound; }

LRTDP_Factory::LRTDP_Factory():
    ProbSearchFactory(SearchEngineType::LrtdpType) {}

LRTDP_Factory::~LRTDP_Factory() = default;

std::unique_ptr<SearchEngine> LRTDP_Factory::construct(const PLAJA::Configuration& config) const {
    bool upperAndLowerBound = PLAJA_GLOBAL::optionParser->is_flag_set(PLAJA_OPTION::upper_and_lower_bound);
    const auto* property_info = config.get_sharable_const<PropertyInformation>(PLAJA::SharableKey::PROP_INFO);
    PLAJA_ASSERT(property_info)
    switch (property_info->_property_type()) {
        case PropertyInformation::MINCOST: {
            if (upperAndLowerBound) {
                return (std::make_unique<LRTDP_DUAL<UpperAndLowerPNodeInfo, UpperAndLowerPNode, ProbSearchSpaceDualCMin, UpperAndLowerValueInitializer<OptimizationCriterion::COST, InitializationCriterion::MIN>>>(config, true));
            } else {
                return std::make_unique<LRTDP<ProbSearchNodeInfo, ProbSearchNode, ProbSearchSpaceCMin, ValueInitializer<OptimizationCriterion::COST, InitializationCriterion::MIN>>>(config, true);
            }
        }
        case PropertyInformation::PROBMIN: {
            if (upperAndLowerBound) {
                return std::make_unique<LRTDP_DUAL<UpperAndLowerPNodeInfo, UpperAndLowerPNode, ProbSearchSpaceDualPMin, UpperAndLowerValueInitializer<OptimizationCriterion::PROB, InitializationCriterion::MIN>>>(config, true);
            } else {
                return std::make_unique<LRTDP<ProbSearchNodeInfo, ProbSearchNode, ProbSearchSpacePMin, ValueInitializer<OptimizationCriterion::PROB, InitializationCriterion::MIN>>>(config, true);
            }
        }
        case PropertyInformation::PROBMAX: {
            if (upperAndLowerBound) {
                return std::make_unique<LRTDP_DUAL<UpperAndLowerPNodeInfo, UpperAndLowerPNode, ProbSearchSpaceDualPMax, UpperAndLowerValueInitializer<OptimizationCriterion::PROB, InitializationCriterion::MAX>>>(config, false);
            } else {
                return std::make_unique<LRTDP<ProbSearchNodeInfo, ProbSearchNode, ProbSearchSpacePMax, ValueInitializer<OptimizationCriterion::PROB, InitializationCriterion::MAX>>>(config, false);
            }
        }
        default: PLAJA_ABORT
    }
}

std::unique_ptr<SearchEngine> LRTDP_Factory::construct_LRTDP(const PLAJA::Configuration& configuration) { return LRTDP_Factory().construct(configuration); }

std::unique_ptr<SearchEngine> LRTDP_Factory::construct_LRTDP(const PLAJA::Configuration& config, SearchEngine* parent) {
    bool upperAndLowerBound = PLAJA_GLOBAL::optionParser->is_flag_set(PLAJA_OPTION::upper_and_lower_bound);
    const auto* property_info = config.get_sharable_const<PropertyInformation>(PLAJA::SharableKey::PROP_INFO);
    PLAJA_ASSERT(property_info)
    switch (property_info->_property_type()) {
        case PropertyInformation::MINCOST: {
            if (upperAndLowerBound) {
                auto* parent_ = dynamic_cast<ProbSearchEngine<UpperAndLowerPNodeInfo, UpperAndLowerPNode, ProbSearchSpaceDualCMin, UpperAndLowerValueInitializer<OptimizationCriterion::COST, InitializationCriterion::MIN>>*>(parent);
                PLAJA_ASSERT(parent_)
                return (std::make_unique<LRTDP_DUAL<UpperAndLowerPNodeInfo, UpperAndLowerPNode, ProbSearchSpaceDualCMin, UpperAndLowerValueInitializer<OptimizationCriterion::COST, InitializationCriterion::MIN>>>(config, true, parent_));
            } else {
                auto* parent_ = dynamic_cast<ProbSearchEngine<ProbSearchNodeInfo, ProbSearchNode, ProbSearchSpaceCMin, ValueInitializer<OptimizationCriterion::COST, InitializationCriterion::MIN>>*>(parent);
                PLAJA_ASSERT(parent_)
                return std::make_unique<LRTDP<ProbSearchNodeInfo, ProbSearchNode, ProbSearchSpaceCMin, ValueInitializer<OptimizationCriterion::COST, InitializationCriterion::MIN>>>(config, true, parent_);
            }
        }
        case PropertyInformation::PROBMIN: {
            if (upperAndLowerBound) {
                auto* parent_ = dynamic_cast<ProbSearchEngine<UpperAndLowerPNodeInfo, UpperAndLowerPNode, ProbSearchSpaceDualPMin, UpperAndLowerValueInitializer<OptimizationCriterion::PROB, InitializationCriterion::MIN>>*>(parent);
                PLAJA_ASSERT(parent_)
                return std::make_unique<LRTDP_DUAL<UpperAndLowerPNodeInfo, UpperAndLowerPNode, ProbSearchSpaceDualPMin, UpperAndLowerValueInitializer<OptimizationCriterion::PROB, InitializationCriterion::MIN>>>(config, true, parent_);
            } else {
                auto* parent_ = dynamic_cast<ProbSearchEngine<ProbSearchNodeInfo, ProbSearchNode, ProbSearchSpacePMin, ValueInitializer<OptimizationCriterion::PROB, InitializationCriterion::MIN>>*>(parent);
                PLAJA_ASSERT(parent_)
                return std::make_unique<LRTDP<ProbSearchNodeInfo, ProbSearchNode, ProbSearchSpacePMin, ValueInitializer<OptimizationCriterion::PROB, InitializationCriterion::MIN>>>(config, true, parent_);
            }
        }
        case PropertyInformation::PROBMAX: {
            if (upperAndLowerBound) {
                auto* parent_ = dynamic_cast<ProbSearchEngine<UpperAndLowerPNodeInfo, UpperAndLowerPNode, ProbSearchSpaceDualPMax, UpperAndLowerValueInitializer<OptimizationCriterion::PROB, InitializationCriterion::MAX>>*>(parent);
                PLAJA_ASSERT(parent_)
                return std::make_unique<LRTDP_DUAL<UpperAndLowerPNodeInfo, UpperAndLowerPNode, ProbSearchSpaceDualPMax, UpperAndLowerValueInitializer<OptimizationCriterion::PROB, InitializationCriterion::MAX>>>(config, false, parent_);
            } else {
                auto* parent_ = dynamic_cast<ProbSearchEngine<ProbSearchNodeInfo, ProbSearchNode, ProbSearchSpacePMax, ValueInitializer<OptimizationCriterion::PROB, InitializationCriterion::MAX>>*>(parent);
                PLAJA_ASSERT(parent_)
                return std::make_unique<LRTDP<ProbSearchNodeInfo, ProbSearchNode, ProbSearchSpacePMax, ValueInitializer<OptimizationCriterion::PROB, InitializationCriterion::MAX>>>(config, false, parent_);
            }
        }
        default: PLAJA_ABORT
    }
}

