//
// This file is part of the PlaJA code base (2019 - 2020).
//

#ifndef PLAJA_JANI_2_NNET_H
#define PLAJA_JANI_2_NNET_H

#include <memory>
#include <string>
#include <unordered_map>
#include <vector>
#include "../../../parser/ast/expression/forward_expression.h"
#include "../../../parser/ast/forward_ast.h"
#include "../../../assertions.h"
#include "../../factories/forward_factories.h"
#include "../../non_prob_search/policy/forward_policy.h"
#include "../../states/forward_states.h"
#include "../forward_information.h"
#include "forward_jani_2_nnet.h"
#include "using_jani2nnet.h"

namespace PLAJA { class NeuralNetwork; }

/**
 * This class documents the interface between a JANI model and an underlying neural network stored in .nnet format.
 * The interface is dependent on the current implementation of JANI models, \ie, the used indexes and ids.
 */
class Jani2NNet {
    friend class Jani2NNetParser;

protected:
    const Model* model; // underlying model
    std::string interfaceFile;
    std::string nnFile;
    std::string torchFile;
    bool doApplicabilityFiltering;
    std::vector<unsigned int> layerSizes;
    std::vector<unsigned int> hiddenLayerSizes;
    std::vector<std::unique_ptr<InputFeatureToJANI>> inputFeatureMapping; // For applicability filtering any variable occurring in a guard is to be considered an input to the policy (inputFeatureMapping.size() >= layerSizes[0]).
    std::vector<ActionLabel_type> outputsToLabels;
    std::unordered_map<ActionLabel_type, OutputIndex_type> labelsToOutputs;
    std::vector<ActionLabel_type> unlearnedLabels;

private:
    // loaded structures
    mutable std::unique_ptr<PLAJA::NeuralNetwork> nn;
    mutable std::unique_ptr<NNPolicy> nnPolicy;

protected:
    void compute_labels_to_outputs();
    void compute_layer_sizes();
    void compute_extended_inputs();

    explicit Jani2NNet(const Model& model_);
    Jani2NNet(const Jani2NNet& jani_2_nnet);
public:
    virtual ~Jani2NNet();

    Jani2NNet(Jani2NNet&& jani_2_nnet) = delete;
    DELETE_ASSIGNMENT(Jani2NNet)

    static std::unique_ptr<Jani2NNet> load_interface(const Model& model);
    static std::unique_ptr<Jani2NNet> load_interface(const Model& model, const std::string& jani_2_nnet_file);
    std::unique_ptr<Jani2NNet> copy_interface(const std::string& path_to_directory, const std::string* filename_prefix = nullptr, bool dump = true) const;

    [[nodiscard]] const PLAJA::NeuralNetwork& load_network() const;
    [[nodiscard]] const NNPolicy& load_policy(const PLAJA::Configuration& config) const;
    // For debugging only, load policy directly for runtime usage.
    FCT_IF_DEBUG([[nodiscard]] ActionLabel_type eval_policy(const StateBase& state) const;)
    FCT_IF_DEBUG([[nodiscard]] bool is_chosen(const StateBase& state, ActionLabel_type action_label) const;)

    [[nodiscard]] inline const Model& get_model() const { return *model; }

    [[nodiscard]] const ModelInformation& get_model_info() const;

    [[nodiscard]] inline const std::string& get_interface_file() const { return interfaceFile; }

    [[nodiscard]] inline const std::string& get_nn_file() const { return nnFile; }

    [[nodiscard]] inline const std::string& get_torch_file() const { return torchFile; }

    [[nodiscard]] inline bool _do_applicability_filtering() const { return doApplicabilityFiltering; }

    /* Layers: */

    [[nodiscard]] inline std::size_t get_num_layers() const { return layerSizes.size(); }

    [[nodiscard]] inline std::size_t get_num_hidden_layers() const { return layerSizes.size() - 2; }

    [[nodiscard]] inline const std::vector<unsigned int>& get_layer_sizes() const { return layerSizes; }

    [[nodiscard]] inline const std::vector<unsigned int>& get_hidden_layer_sizes() const { return hiddenLayerSizes; }

    /* Input: */

    [[nodiscard]] inline std::size_t get_num_input_features() const { return layerSizes[0]; };

    [[nodiscard]] inline const InputFeatureToJANI* get_input_feature(InputIndex_type index) const {
        PLAJA_ASSERT(index < inputFeatureMapping.size())
        return inputFeatureMapping[index].get();
    }

    [[nodiscard]] VariableIndex_type get_input_feature_index(InputIndex_type index) const;

    [[nodiscard]] const LValueExpression* get_input_feature_expression(InputIndex_type index) const;

    /* Output: */

    [[nodiscard]] inline std::size_t get_num_output_features() const { return outputsToLabels.size(); };

    [[nodiscard]] inline ActionLabel_type get_output_label(OutputIndex_type index) const {
        PLAJA_ASSERT(index < outputsToLabels.size())
        return outputsToLabels[index];
    }

    [[nodiscard]] inline const std::vector<ActionLabel_type>& get_unlearned_labels() const { return unlearnedLabels; }

    [[nodiscard]] inline bool is_learned(ActionLabel_type action_label) const { return labelsToOutputs.count(action_label); } // silent-labeled operators should not be learned

    [[nodiscard]] inline OutputIndex_type get_output_index(ActionLabel_type action_label) const {
        PLAJA_ASSERT(is_learned(action_label))
        return labelsToOutputs.at(action_label);
    }

    /* Auxiliary counting: */

    [[nodiscard]] inline std::size_t count_hidden_neurons() const {
        std::size_t rlt = 0;
        for (const auto num_neurons: get_hidden_layer_sizes()) { rlt += num_neurons; }
        return rlt;
    }

    [[nodiscard]] inline std::size_t count_neurons() const { return get_num_input_features() + count_hidden_neurons() + get_num_output_features(); }

    [[nodiscard]] inline std::size_t count_weighted_sums() const { return count_hidden_neurons() + get_num_output_features(); } // for each hidden/output neuron a weighted input sum

    [[nodiscard]] inline std::size_t count_activations() const { return count_hidden_neurons(); } // for each hidden neuron an activation

    [[nodiscard]] inline std::size_t count_output_constraints() const { return get_num_output_features() - 1; } // #outputs - 1 argmax constraints

    [[nodiscard]] inline std::size_t count_linear_constraints() const { return count_weighted_sums() + count_output_constraints(); }

    [[nodiscard]] inline std::size_t number_of_marabou_encoding_vars() const { return count_neurons() + count_hidden_neurons(); } // number of vars in the standard Marabou encoding: one variable per neuron + one auxiliary variable for each hidden neuron

/**********************************************************************************************************************/

    class InputFeatureIterator {
        friend Jani2NNet;
    private:
        const Jani2NNet* nnInterface;
        InputIndex_type currentIndex;
        std::size_t numInputs;

        explicit InputFeatureIterator(const Jani2NNet& nn_interface, InputIndex_type start = 0, bool extended = false):
            nnInterface(&nn_interface)
            , currentIndex(start)
            , numInputs(extended ? nnInterface->inputFeatureMapping.size() : nnInterface->layerSizes[0]) {}

    public:
        ~InputFeatureIterator() = default;
        DELETE_CONSTRUCTOR(InputFeatureIterator)

        inline void operator++() { ++currentIndex; }

        [[nodiscard]] inline bool end() const { return currentIndex >= numInputs; }

        [[nodiscard]] inline InputIndex_type _input_index() const { return currentIndex; }

        [[nodiscard]] inline const InputFeatureToJANI* get_input_feature() const { return nnInterface->get_input_feature(currentIndex); }

        [[nodiscard]] inline VariableIndex_type get_input_feature_index() const { return nnInterface->get_input_feature_index(currentIndex); }

        [[nodiscard]] inline const LValueExpression* get_input_feature_expression() const { return nnInterface->get_input_feature_expression(currentIndex); }
    };

    [[nodiscard]] inline InputFeatureIterator init_input_feature_iterator(bool extended = false) const { return InputFeatureIterator(*this, 0, extended); }

    [[nodiscard]] inline InputFeatureIterator init_input_feature_iterator_extended() const { return InputFeatureIterator(*this, get_num_input_features(), true); }

    class OutputFeatureIterator {
        friend Jani2NNet;
    private:
        const Jani2NNet* nnInterface;
        OutputIndex_type currentIndex;

        explicit OutputFeatureIterator(const Jani2NNet& nn_interface):
            nnInterface(&nn_interface)
            , currentIndex(0) {}

    public:
        ~OutputFeatureIterator() = default;
        DELETE_CONSTRUCTOR(OutputFeatureIterator)

        inline void operator++() { ++currentIndex; }

        [[nodiscard]] inline bool end() const { return currentIndex >= nnInterface->get_num_output_features(); }

        [[nodiscard]] inline OutputIndex_type _output_index() const { return currentIndex; }

        [[nodiscard]] inline ActionLabel_type get_output_label() const { return nnInterface->get_output_label(currentIndex); }

        [[nodiscard]] inline bool is_learned() const { return nnInterface->is_learned(get_output_label()); }
    };

    [[nodiscard]] inline OutputFeatureIterator init_output_feature_iterator() const { return OutputFeatureIterator(*this); }

};

#endif //PLAJA_JANI_2_NNET_H
