//
// This file is part of the PlaJA code base (2019 - 2021).
//

#include "invariant_strengthening.h"
#include <memory>
#include "../../include/marabou_include/input_query.h"
#include "../../option_parser/option_parser.h"
#include "../information/jani2nnet/jani_2_nnet.h"
#include "../information/property_information.h"
#include "../predicate_abstraction/successor_generation/action_op/action_op_abstract.h"
#include "../predicate_abstraction/successor_generation/successor_generator_abstract.h"
#include "../predicate_abstraction/search_statistics_pa.h"
#include "../smt/model/model_z3.h"
#include "../smt/solver/smt_solver_z3.h"
#include "../smt/visitor/extern/to_z3_visitor.h"
#include "../smt_nn/model/model_marabou.h"
#include "../smt_nn/to_z3/marabou_to_z3.h"
#include "../states/state_values.h"

namespace PLAJA_OPTION {
    extern const std::string nn_sat;
    extern const std::string per_action_query;
    extern const std::string preprocess_query;
}
namespace MARABOU_IN_PLAJA { [[nodiscard]] extern std::unique_ptr<InputQuery> load_nn(const Jani2NNet& jani_2_nnet); }

InvariantStrengthening::InvariantStrengthening(const PLAJA::Configuration& config):
    SearchEngine(config)
    , searchStatistics(new SearchStatisticsPA())
    , modelZ3(new ModelZ3(*model))
    , solver(new PLAJA::SMTSolverM(*modelZ3, &searchStatistics->share_smt_stats_z3()))
    , modelMarabou(nullptr)
    , query(nullptr)
    , forwardingStructure(new z3::expr_vector(modelZ3->get_context()))
    , perActionQuery(PLAJA_GLOBAL::optionParser->is_flag_set(PLAJA_OPTION::per_action_query)) {

    // set predicates and compute state info of resulting abstract model

    // add src variable (bound)s
    solver->add_and_register_variables(modelZ3->get_src_bounds());
    solver->add(modelZ3->get_target_bounds().to_vec());
    // finalize successor generation
    successorGenerator = SuccessorGeneratorAbstract::construct(*solver, nullptr, searchStatistics.get());

    // Reachability as start invariant
    const auto* start_exp = propertyInfo->get_start();
    const auto* reach_exp = propertyInfo->get_reach();
    PLAJA_ASSERT(start_exp)
    PLAJA_ASSERT(reach_exp)
    invariant = std::make_unique<z3::expr>(modelZ3->to_z3(*start_exp, false));
    non_invariant = std::make_unique<z3::expr>(modelZ3->to_z3(*reach_exp, true));

    // load NN
    if (propertyInfo->get_nn_interface()) {
        modelMarabou = std::make_unique<ModelMarabou>(*propertyInfo->get_nn_interface());

        // generate base query
        query = modelMarabou->make_query(true);

        auto nn_vars = MARABOU_TO_Z3::generate_nn_variables(*query, modelZ3->get_context(), Z3_IN_PLAJA::z3NNVarPrefix);
        MARABOU_TO_Z3::match_state_vars_to_nn(*nn_vars, modelMarabou->get_nn_in_marabou(), modelZ3->get_src_vars(), modelMarabou->get_nn_interface());
        std::unique_ptr<MARABOU_TO_Z3::EncodingInformation> encoding_information = MARABOU_TO_Z3::EncodingInformation::construct_encoding_information(*PLAJA_GLOBAL::optionParser);
        if (encoding_information->_mip_encoding() || PLAJA_GLOBAL::optionParser->is_flag_set(PLAJA_OPTION::preprocess_query)) { // forwarding structures
            auto preprocessed_bounds = MARABOU_IN_PLAJA::PreprocessedBounds::preprocess_bounds(query->_query());
            encoding_information->set_preprocessed_bounds(preprocessed_bounds.get());
            forwardingStructure = std::make_unique<z3::expr_vector>(MARABOU_TO_Z3::generate_nn_forwarding_structures(*query, *nn_vars, *encoding_information));
        } else {
            forwardingStructure = std::make_unique<z3::expr_vector>(MARABOU_TO_Z3::generate_nn_forwarding_structures(*query, *nn_vars, *encoding_information));
        }
        outputInterfacePerAction = MARABOU_TO_Z3::generate_output_interfaces(modelMarabou->get_nn_interface(), modelMarabou->get_nn_in_marabou(), *nn_vars);

    }
}

InvariantStrengthening::~InvariantStrengthening() = default;

SearchEngine::SearchStatus InvariantStrengthening::initialize() { return SearchEngine::IN_PROGRESS; }

SearchEngine::SearchStatus InvariantStrengthening::finalize() { /*TODO*/ return SearchStatus::FINISHED; }

SearchEngine::SearchStatus InvariantStrengthening::step() {
    // find a s not in S_inv
    // for each action op check whether there exists a transition leaving the invariant.
    bool violation_found = false;
    solver->push();
    solver->add(*invariant);
    solver->add(*non_invariant);

    successorGenerator->explore(*solver);
    // ADD NN
    solver->add(*forwardingStructure); // empty iff non-policy-restricted case

    if (perActionQuery) {
        for (auto actions_it = successorGenerator->actionsIterator(); !actions_it.end() && !violation_found; ++actions_it) {
            solver->push(); // PUSH ACTION
            if (modelMarabou) { solver->add(outputInterfacePerAction[actions_it.action_label()]); }
            solver->add(successorGenerator->extract_action_in_z3(actions_it.action_label()));
            violation_found = check_incremental();
            solver->pop(); // POP ACTION
        }
    } else {
        solver->add(extract_expansion_in_z3());
        violation_found = check_incremental();
    }

    solver->pop();

    // adapt invariant
    if (violation_found) {
        // std::cout << *invariant << std::endl << std::endl;
        // std::cout << *non_invariant << std::endl << std::endl << "--------" << std::endl << std::endl;
        solver->reset();
        return SearchEngine::IN_PROGRESS;
    } else {
        return SearchEngine::SOLVED;
    }
}

// auxiliaries

bool InvariantStrengthening::check_incremental() {
    bool violation_found = false;
    bool rlt = true;
    while (rlt) {
        rlt = solver->check(); // TODO for now we do not count NN-SAT queries
        // analyse result
        if (rlt) {
            violation_found = true;
            searchStatistics->inc_attr_unsigned(PLAJA::StatsUnsigned::ITERATIONS);
            const auto m = solver->get_model();
            StateValues violation_state(modelZ3->get_number_automata_instances() + modelZ3->get_number_vars());
            auto inv_ext = !solver->encode_and_extract_solution(modelZ3->get_src_var_indexes(), m, violation_state);
            invariant = std::make_unique<z3::expr>(*invariant && inv_ext);
            { // violation state in target vars:
                std::vector<z3::expr> violation_state_in_z3;
                violation_state_in_z3.reserve(violation_state.size());
                for (const auto& [state_index, state_expr]: modelZ3->get_target_var_indexes()) {
                    violation_state_in_z3.push_back(state_expr == state_expr.ctx().int_val(violation_state.get_int(state_index)));
                }
                non_invariant = std::make_unique<z3::expr>(*non_invariant || Z3_IN_PLAJA::to_conjunction(violation_state_in_z3));
            }
            solver->add(inv_ext); // extend invariant and check violation from new invariant incrementally
        }
    }
    //
    return violation_found;
}

z3::expr InvariantStrengthening::extract_expansion_in_z3() const {
    std::vector<z3::expr> disjunction;
    for (auto actions_it = successorGenerator->actionsIterator(); !actions_it.end(); ++actions_it) {
        const auto action_label = actions_it.action_label();
        if (modelMarabou && modelMarabou->get_nn_interface().is_learned(action_label)) {
            disjunction.push_back(this->outputInterfacePerAction[action_label] && successorGenerator->extract_action_in_z3(action_label));
        } else { disjunction.push_back(successorGenerator->extract_action_in_z3(action_label)); }
    }
    return Z3_IN_PLAJA::to_disjunction(disjunction);
}

// override stats:

SearchStatistics& InvariantStrengthening::get_search_statistics() { return *searchStatistics; }

void InvariantStrengthening::print_statistics() const { searchStatistics->print_statistics(); }

void InvariantStrengthening::stats_to_csv(std::ofstream& file) const { searchStatistics->stats_to_csv(file); }

void InvariantStrengthening::stat_names_to_csv(std::ofstream& file) const { searchStatistics->stat_names_to_csv(file); }
