//
// This file is part of the PlaJA code base (2019 - 2021).
//

#include "search_engine_base.h"
#include "../../../parser/ast/model.h"
#include "../../../stats/stats_base.h"
#include "../../fd_adaptions/search_statistics.h"
#include "../../information/model_information.h"
#include "../../information/property_information.h"
#include "../../prob_search/state_space.h"
#include "../../states/state_values.h"
#include "../../successor_generation/successor_generator.h"
#include "../initial_states_enumerator.h"
#include "../policy/policy_restriction.h"
#include "search_node_info_base.h"
#include "search_space_base.h"

SearchEngineBase::SearchEngineBase(const PLAJA::Configuration& config):
    SearchEngine(config)
    , searchSpace(nullptr) {
}

SearchEngineBase::~SearchEngineBase() = default;

void SearchEngineBase::add_additional_initial_states() {
    PLAJA_ASSERT(propertyInfo)
    if (propertyInfo->get_start()) {
        for (const auto& initial_state_values: InitialStatesEnumerator::compute_initial_states(*model, *propertyInfo->get_start())) {
            searchSpace->add_initial_state(initial_state_values);
        }
    }
}

SearchEngine::SearchStatus SearchEngineBase::initialize() {
    PLAJA_LOG("Initializing ...")

    /* Compute additional initial states. */
    add_additional_initial_states();

    for (const auto initial_id: searchSpace->_initial_states()) {
        if (initialize(initial_id) == SearchEngine::SOLVED) { return SearchEngine::SOLVED; }
    }

    return SearchEngine::IN_PROGRESS;

}

/* Override stats. */

void SearchEngineBase::print_statistics() const {
    searchStatistics->print_statistics();
    searchSpace->print_statistics();
}

void SearchEngineBase::stats_to_csv(std::ofstream& file) const {
    searchStatistics->stats_to_csv(file);
    searchSpace->stats_to_csv(file);
}

void SearchEngineBase::stat_names_to_csv(std::ofstream& file) const {
    searchStatistics->stat_names_to_csv(file);
    SearchSpaceBase::stat_names_to_csv(file);
}

/**********************************************************************************************************************/

SearchEngineNonProbBase::SearchEngineNonProbBase(const PLAJA::Configuration& config):
    SearchEngineBase(config)
    , successorGenerator(new SuccessorGenerator(config, *model))
    , frontier()
    , goal(propertyInfo->get_reach())
    , policyRestriction(PolicyRestriction::construct_policy_restriction(config)) {

    SEARCH_STATISTICS::add_basic_stats(*searchStatistics);
}

SearchEngineNonProbBase::~SearchEngineNonProbBase() = default;

/**********************************************************************************************************************/

SearchEngineExplorerBase::SearchEngineExplorerBase(const PLAJA::Configuration& config):
    SearchEngineNonProbBase(config)
    , stateSpace(new StateSpaceProb())
    , currentStateInfo(nullptr) {

    if (searchStatistics) { searchStatistics->add_unsigned_stats(PLAJA::StatsUnsigned::GoalPathStates, "GoalPathStates", 0); }

}

SearchEngineExplorerBase::~SearchEngineExplorerBase() = default;

/***********************************************************************************************************************/

void SearchEngineExplorerBase::compute_goal_path_states() {
    PLAJA_ASSERT(get_status() == SearchEngine::SOLVED or get_status() == SearchEngine::FINISHED)
    if (searchStatistics) { searchStatistics->set_attr_unsigned(PLAJA::StatsUnsigned::GoalPathStates, searchSpace->compute_goal_path_states()); }
    else { searchSpace->compute_goal_path_states(); }
}

/* */

StateID_type SearchEngineExplorerBase::search_goal_path(const StateValues& state) {
    const auto id = searchSpace->add_state(state);

    auto node = searchSpace->add_node(id);

    if (node().known_goal_distance()) { return id; }

    if (node.has_been_reached()) { // Has been reached, but no goal path found. Hence, dead-end.
        node().set_dead_end();
        return id;
    }

    /* Search. */

    frontier.clear();
    searchSpace->reset();

    auto rlt = initialize(id);
    PLAJA_ASSERT_EXPECT(rlt == SearchStatus::IN_PROGRESS)

    while (SearchEngine::SearchStatus::IN_PROGRESS == rlt) { rlt = step(); }

    // Set goal information.
    if (searchSpace->goal_path_frontier_empty()) { searchSpace->get_node_info(id).set_dead_end(); } // IMPORTANT: Reload node since information ptr may have been invalidated due to reallocation.
    else {
        searchSpace->compute_goal_path_states();
        PLAJA_ASSERT(searchSpace->get_node_info(id).has_goal_path())
    }

    return id;
}

HeuristicValue_type SearchEngineExplorerBase::get_goal_distance(StateID_type state) const {
    return searchSpace->get_node_info(state).get_goal_distance();
}

UpdateOpID_type SearchEngineExplorerBase::get_goal_op(StateID_type state) const {
    return searchSpace->get_node_info(state).get_goal_path_op();
}

/* */

struct UpdateOpStructure SearchEngineExplorerBase::get_update_op(UpdateOpID_type update_op) const { return searchSpace->compute_update_op(update_op); }

ActionLabel_type SearchEngineExplorerBase::get_label(ActionOpID_type op_id) const {
    const auto& model_info = model->get_model_information();
    const auto op_structure = model_info.compute_action_op_id_structure(op_id);
    return ModelInformation::action_id_to_label(model_info.sync_id_to_action(op_structure.syncID));
}

FCT_IF_DEBUG(std::string SearchEngineExplorerBase::get_label_name(ActionLabel_type label) const { return model->get_action_name(ModelInformation::action_label_to_id(label)); })