//
// This file is part of the PlaJA code base (2019 - 2021).
//

#ifndef PLAJA_SEARCH_ENGINE_BASE_H
#define PLAJA_SEARCH_ENGINE_BASE_H

#include <list>
#include "../../../parser/ast/expression/forward_expression.h"
#include "../../../assertions.h"
#include "../../fd_adaptions/search_engine.h"
#include "../../successor_generation/forward_successor_generation.h"
#include "../../states/forward_states.h"
#include "../../using_search.h"
#include "../policy/forward_policy.h"

class SearchSpaceBase;

class StateSpaceProb;

namespace STATE_SPACE { struct StateInformationProb; }

class SearchEngineBase: public SearchEngine {
    friend struct InitializeBase;
    friend struct StepBase;

protected:
    SearchSpaceBase* searchSpace;

    inline void set_search_space(SearchSpaceBase* search_space) { searchSpace = search_space; }

    void add_additional_initial_states();

    [[nodiscard]] SearchEngine::SearchStatus initialize() final;
    [[nodiscard]] virtual SearchEngine::SearchStatus initialize(StateID_type initial_id) = 0;

public:
    explicit SearchEngineBase(const PLAJA::Configuration& config);
    ~SearchEngineBase() override = 0;
    DELETE_CONSTRUCTOR(SearchEngineBase)

    // Override stats. */
    void print_statistics() const override;
    void stats_to_csv(std::ofstream& file) const override;
    void stat_names_to_csv(std::ofstream& file) const override;

};

class SearchEngineNonProbBase: public SearchEngineBase {
    friend struct InitializeBase;
    friend struct StepBase;

protected:

    enum SearchEngineType { // Used by templates to share code maintenance.
        BFS,
        EXPLORER,
        NN_EXPLORER
    };

    std::unique_ptr<SuccessorGenerator> successorGenerator;
    std::list<StateID_type> frontier;
    const Expression* goal;
    std::unique_ptr<PolicyRestriction> policyRestriction;

public:
    explicit SearchEngineNonProbBase(const PLAJA::Configuration& config);
    ~SearchEngineNonProbBase() override = 0;
    DELETE_CONSTRUCTOR(SearchEngineNonProbBase)

};

class SearchEngineExplorerBase: public SearchEngineNonProbBase {
    friend struct InitializeBase;
    friend struct StepBase;

protected:
    std::unique_ptr<StateSpaceProb> stateSpace;
    STATE_SPACE::StateInformationProb* currentStateInfo; // Cache for state space maintenance.

public:
    explicit SearchEngineExplorerBase(const PLAJA::Configuration& config);
    ~SearchEngineExplorerBase() override = 0;
    DELETE_CONSTRUCTOR(SearchEngineExplorerBase)

    /** Compute goal paths -- after search() is completed. */
    void compute_goal_path_states();

    /* Policy interface. */

    /**
     * Search goal path for provided state.
     * @return the id of the provided state.
     */
    StateID_type search_goal_path(const StateValues& state);

    [[nodiscard]] HeuristicValue_type get_goal_distance(StateID_type state) const;

    [[nodiscard]] UpdateOpID_type get_goal_op(StateID_type state) const;

    /* Aux. */

    [[nodiscard]] struct UpdateOpStructure get_update_op(UpdateOpID_type update_op) const;

    [[nodiscard]] ActionLabel_type get_label(ActionOpID_type op_id) const;

    FCT_IF_DEBUG([[nodiscard]] std::string get_label_name(ActionLabel_type label) const;)

};

#endif //PLAJA_SEARCH_ENGINE_BASE_H
