//
// This file is part of the PlaJA code base (2019 - 2021).
//

#ifndef PLAJA_STEP_BASE_H
#define PLAJA_STEP_BASE_H

struct StepBase {

    template<SearchEngineNonProbBase::SearchEngineType engine_type, typename SearchEngine_t, typename SearchNode_t>
    inline static SearchEngine::SearchStatus step(SearchEngineNonProbBase* const engine, SearchEngine_t* const derived_engine) { // i.e., the "this" pointer
        if (engine->frontier.empty()) {
            if constexpr (engine_type == SearchEngineNonProbBase::BFS) { std::cout << "Goal unreachable!" << std::endl; }
            return SearchEngine::SOLVED;
        }

        // Expand state
        StateID_type current_id = engine->frontier.front();
        engine->frontier.pop_front();
        engine->searchStatistics->inc_attr_unsigned(PLAJA::StatsUnsigned::EXPANDED_STATES);

        // Explore
        State current_state = engine->searchSpace->get_state(current_id);
        const auto& current_node_info = derived_engine->searchSpace->get_node_info(current_id);
        engine->successorGenerator->explore(current_state);

        // State space maintenance
        if constexpr (engine_type != SearchEngineNonProbBase::BFS) {
            derived_engine->currentStateInfo = &derived_engine->stateSpace->get_state_information(current_id);
            derived_engine->currentStateInfo->reserve(engine->successorGenerator->_number_of_enabled_ops());
        }

        // Terminal check
        bool is_terminal = true;
        int is_policy_terminal = -1;

        // Iterate
        for (auto action_it = engine->successorGenerator->actionIterator(); !action_it.end(); ++action_it) { // for each action ...
            ActionLabel_type action_label = action_it.action_label();

            bool is_chosen_by_policy(false);

            // policy restriction
            if (engine->policyRestriction and engine->policyRestriction->is_pruned(current_state, action_label)) {
                is_policy_terminal = is_policy_terminal == -1 ? 1 : 0;
                if constexpr (engine_type != SearchEngineNonProbBase::NN_EXPLORER) { continue; } // only explore for chosen actions ...
            } else {
                is_policy_terminal = 0;
                is_chosen_by_policy = true;
            }

            engine->searchStatistics->inc_attr_unsigned(PLAJA::StatsUnsigned::ACTION_LABELS); // applicable action, thus at least one transition
            is_terminal = false;

            for (auto op_it = action_it.iterator(); !op_it.end(); ++op_it) {
                const ActionOpInit& action_op = *op_it;

                PLAJA_ASSERT(action_op.size() > 0) // at least one transition
                engine->searchStatistics->inc_attr_unsigned(PLAJA::StatsUnsigned::ACTION_OPS);
                CHECK_NARROWING_CONVERSION(action_op.size(), int)
                engine->searchStatistics->inc_attr_unsigned(PLAJA::StatsUnsigned::TRANSITIONS, action_op.size()); // NOLINT(cppcoreguidelines-narrowing-conversions)

                if constexpr (engine_type != SearchEngineNonProbBase::BFS) { derived_engine->currentStateInfo->add_operator(action_op._op_id(), action_op.size()); } // maintain state space

                for (auto transition_it = action_op.transitionIterator(); !transition_it.end(); ++transition_it) { // for each update of the operator
                    std::unique_ptr<State> successor_state = engine->successorGenerator->compute_successor(transition_it.transition(), current_state);
                    engine->searchStatistics->inc_attr_unsigned(PLAJA::StatsUnsigned::GENERATED_STATES);

                    const StateID_type successor_id = successor_state->get_id_value();
                    SearchNode successor_node = derived_engine->searchSpace->add_node(successor_id);

                    if constexpr (engine_type != SearchEngineNonProbBase::BFS) {
                        derived_engine->currentStateInfo->add_successor_to_current_op(successor_id, transition_it.prob());

                        // policy
                        if constexpr (engine_type == SearchEngineNonProbBase::NN_EXPLORER) {
                            if (is_chosen_by_policy) { successor_node.cast_info<SearchNode_t>().add_parent_under_policy(current_id); }
                        }

                        if (not derived_engine->stateSpace->add_parent(successor_id, current_id)) { continue; } // If parent already added, then state already reached via current state.
                    }

                    /* Have seen state in previous iterations, can reuse information? */
                    if (successor_node.has_been_reached() and not successor_node.is_reached()) {
                        PLAJA_ABORT // TODO Unexpected on current benchmark set, since operations are invertible.

                        successor_node().add_parent(current_id, engine->searchSpace->compute_update_op(action_op._op_id(), transition_it.update_index()));

                        if (successor_node().has_goal_path()) { engine->searchSpace->add_goal_path_frontier(successor_id); }
                        else if (not successor_node().known_goal_distance()) { successor_node().set_dead_end(); } // Has been reached, but no goal state found -> so dead end.
                        else { PLAJA_ASSERT(successor_node().is_dead_end()) }

                        continue;
                    }

                    // reach:
                    if (not successor_node.is_reached()) {
                        successor_node.set_reached(current_id, engine->searchSpace->compute_update_op(action_op._op_id(), transition_it.update_index()), current_node_info.get_start_distance() + 1);
                        engine->searchStatistics->inc_attr_unsigned(PLAJA::StatsUnsigned::REACHABLE_STATES);

                        // Goal check:
                        if (engine->goal and engine->goal->evaluateInteger(successor_state.get())) {
                            engine->searchStatistics->inc_attr_unsigned(PLAJA::StatsUnsigned::GOAL_STATES);
                            engine->searchSpace->add_goal_path_frontier(successor_id);
                            successor_node().set_goal();
                            if constexpr (engine_type == SearchEngineNonProbBase::BFS) { return SearchEngine::SOLVED; }
                        }

                        engine->frontier.push_back(successor_id);

                    } else {
                        successor_node().add_parent(current_id, engine->searchSpace->compute_update_op(action_op._op_id(), transition_it.update_index()));
                    }

                }

            }

        }

        // Terminal check
        if (is_terminal) {
            engine->searchStatistics->inc_attr_unsigned(PLAJA::StatsUnsigned::TERMINAL_STATES);
            if constexpr (engine_type == SearchEngineNonProbBase::BFS) { engine->searchStatistics->inc_attr_unsigned(PLAJA::StatsUnsigned::DEAD_END_STATES); } // also a dead-end as goal states will not be expanded
        }
        if (is_policy_terminal == 1) { engine->searchStatistics->inc_attr_unsigned(PLAJA::StatsUnsigned::POLICY_TERMINAL_STATES); }

        return SearchEngine::IN_PROGRESS;
    }

};

#endif //PLAJA_STEP_BASE_H
