//
// This file is part of the PlaJA code base (2019 - 2021).
//

#include <algorithm>
#include "initial_states_enumerator.h"
#include "../../include/factory_config_const.h"
#include "../../option_parser/option_parser.h"
#include "../../parser/ast/expression/non_standard/location_value_expression.h"
#include "../../parser/ast/expression/non_standard/state_condition_expression.h"
#include "../../parser/ast/expression/non_standard/state_values_expression.h"
#include "../../parser/ast/expression/non_standard/states_values_expression.h"
#include "../../parser/ast/model.h"
#include "../factories/configuration.h"
#include "../information/model_information.h"
#include "../states/state_values.h"

#ifdef USE_Z3

#include "../smt/model/model_z3.h"
#include "../smt/solver/smt_solver_z3.h"
#include "../smt/solution_enumerator/solution_enumerator_z3.h"
#include "../smt/solution_enumerator/solution_enumerator_bs.h"

#else
#include "../smt/solution_enumerator/solution_enumerator_naive.h"
#endif

// extern:
namespace PLAJA_OPTION {
    extern const std::string naive_initial_state_enum;
    extern const std::string additional_initial_states;
}

std::list<StateValues> InitialStatesEnumerator::compute_initial_states(const Model& model, const StatesValuesExpression& initial_states_values) {
    std::list<StateValues> initial_states;
    for (auto it = initial_states_values.init_state_values_iterator(); !it.end(); ++it) {
        initial_states.push_back(it->construct_state_values(model.get_model_information().get_initial_values()));
    }
    return initial_states;
}

namespace SOLUTION_ENUMERATOR_BS { STMT_IF_DEBUG(extern const Expression* groundTruth;) }

std::list<StateValues> InitialStatesEnumerator::compute_initial_states(const Model& model, const StateConditionExpression& initial_state_condition) {

#ifdef USE_Z3
    // prepare z3 encoding
    PLAJA::Configuration config;
    config.set_sharable_const(PLAJA::SharableKey::MODEL, &model);
    //
    const ModelZ3 model_z3(config); // this loads all the model in z3, which is relatively quit an overhead, but usually negligible in absolut terms
    Z3_IN_PLAJA::SMTSolver solver(model_z3.share_context());
    model_z3.add_src_bounds(solver);
    const auto* start_constraint = initial_state_condition.get_constraint();
    if (start_constraint) { model_z3.add_expression(solver, *start_constraint, 0); }

    // prepare constructor
    StateValues constructor(model.get_model_information().get_initial_values());
    for (auto loc_it = initial_state_condition.init_loc_value_iterator(); !loc_it.end(); ++loc_it) {
        constructor.assign_int<true>(loc_it->get_location_index(), loc_it->get_location_value());
    }

    std::list<StateValues> initial_states;

    if (PLAJA_GLOBAL::optionParser->is_flag_set(PLAJA_OPTION::naive_initial_state_enum)) {
        // naive solution enumeration (but still SMT-based)
        solver.push(); // to suppress scoped_timer thread in z3, which is started when using tactic2solver:
        initial_states = SolutionEnumeratorZ3::compute_solutions(solver, model_z3, constructor);
        solver.pop();
    } else {
        STMT_IF_DEBUG(SOLUTION_ENUMERATOR_BS::groundTruth = &initial_state_condition;)
        initial_states = SolutionEnumeratorBS::compute_solutions(solver, model_z3, std::move(constructor));
        STMT_IF_DEBUG(SOLUTION_ENUMERATOR_BS::groundTruth = nullptr;)
    }

    return initial_states;
#else
    return SolutionEnumeratorNaive::compute_solutions(initial_state_condition, model.get_model_information());
#endif

}

std::list<StateValues> InitialStatesEnumerator::compute_initial_states(const Model& model, const Expression& condition) {
    std::list<StateValues> initial_states;
    const auto* initial_states_values = dynamic_cast<const StatesValuesExpression*>(&condition);
    if (initial_states_values) {
        initial_states = compute_initial_states(model, *initial_states_values);
    } else {
        const auto* initial_states_condition = dynamic_cast<const StateConditionExpression*>(&condition);
        if (initial_states_condition) {
            initial_states = compute_initial_states(model, *initial_states_condition);
        } else {
            PLAJA_LOG(PLAJA_UTILS::to_red_string("Warning: Unsupported initial states expression!"))
            return initial_states;
        }
    }

    // special case: additional initial states restricted (mostly for debugging)
    if (PLAJA_GLOBAL::optionParser->has_value_option(PLAJA_OPTION::additional_initial_states)) {

        // retrieve intervals:
        const auto intervals = PLAJA_GLOBAL::optionParser->get_option_intervals(PLAJA_OPTION::additional_initial_states);

        // add initial state if within interval:
        std::size_t index = 0;
        std::list<StateValues> restricted_initial_states;
        for (auto& initial_state: initial_states) {
            if (std::any_of(intervals.cbegin(), intervals.cend(), [index](const std::pair<int, int>& interval) { return interval.first <= index && index <= interval.second; })) {
                restricted_initial_states.push_back(std::move(initial_state));
            }
            ++index;
        }

        return restricted_initial_states;

    } else {
        return initial_states;
    }

}
