//
// This file is part of the PlaJA code base (2019 - 2021).
//

#include <forward_list>
#include "nn_explorer.h"
#include "../../exception/constructor_exception.h"
#include "../../parser/ast/model.h"
#include "../../parser/ast/expression/expression.h"
#include "../../stats/stats_base.h"
#include "../fd_adaptions/state.h"
#include "../prob_search/state_space.h"
#include "../successor_generation/successor_generator.h"
#include "base/search_node_info_base.h"
#include "base/search_space_base.h"
#include "policy/policy_restriction.h"

namespace NN_EXPLORER {

    struct SearchNodeInfo final: public SearchNodeInfoBase {

    private:
        bool inevitableGoalPathState;
        std::unordered_set<StateID_type> parentsUnderPolicy;

    public:
        SearchNodeInfo():
            SearchNodeInfoBase()
            , inevitableGoalPathState(false)
            , parentsUnderPolicy() {
        }

        ~SearchNodeInfo() override = default;

        DEFAULT_CONSTRUCTOR(SearchNodeInfo)

        inline void set_inevitable_goal_path_state() { inevitableGoalPathState = true; }

        inline void add_parent_under_policy(StateID_type state_id) { parentsUnderPolicy.insert(state_id); }

        [[nodiscard]] inline bool is_inevitable_goal_path_state() const { return inevitableGoalPathState; }

        inline const std::unordered_set<StateID_type>& parents_under_policy() { return parentsUnderPolicy; }

    };

    class SearchSpace final: public SearchSpace_<SearchNodeInfo> {
    public:
        explicit SearchSpace(const ModelInformation& model_info):
            SearchSpace_<SearchNodeInfo>(model_info) {}

        ~SearchSpace() override = default;
        DELETE_CONSTRUCTOR(SearchSpace)
    };

}

/**********************************************************************************************************************/

NN_Explorer::NN_Explorer(const PLAJA::Configuration& config):
    SearchEngineExplorerBase(config)
    , searchSpace(new NN_EXPLORER::SearchSpace(model->get_model_information())) {

    if (not policyRestriction) { throw ConstructorException("NN EXPLORER needs policy restriction."); }

    SearchEngineBase::set_search_space(searchSpace.get());

    if (searchStatistics) {
        searchStatistics->add_unsigned_stats(PLAJA::StatsUnsigned::InevitableGoalPathStates, "InevitableGoalPathStates", 0);
        searchStatistics->add_unsigned_stats(PLAJA::StatsUnsigned::InevitableGoalPathStatesStart, "InevitableGoalPathStatesStart", 0);
        searchStatistics->add_unsigned_stats(PLAJA::StatsUnsigned::PolicyGoalPathStates, "PolicyGoalPathStates", 0);
        searchStatistics->add_unsigned_stats(PLAJA::StatsUnsigned::PolicyGoalPathStatesStart, "PolicyGoalPathStatesStart", 0);
        searchStatistics->add_unsigned_stats(PLAJA::StatsUnsigned::EvitablePolicyGoalPathStates, "EvitablePolicyGoalPathStates", 0);
    }
}

NN_Explorer::~NN_Explorer() = default;

#include "base/initialize_base.h"

SearchEngine::SearchStatus NN_Explorer::initialize(StateID_type initial_id) {
    return InitializeBase::initialize<SearchEngineType::NN_EXPLORER, NN_Explorer, NN_EXPLORER::SearchNodeInfo>(this, this, initial_id);
}

#include "base/step_base.h"

SearchEngine::SearchStatus NN_Explorer::step() {
    return StepBase::step<SearchEngineType::NN_EXPLORER, NN_Explorer, NN_EXPLORER::SearchNodeInfo>(this, this);
}

SearchEngine::SearchStatus NN_Explorer::finalize() {
    compute_goal_path_states();
    compute_inevitable_goal_path_states();
    compute_policy_goal_path_states();
    return SearchStatus::FINISHED;
}

/***********************************************************************************************************************/

namespace NN_EXPLORER {

    inline void compute_inevitable(StateID_type id,
                                   std::forward_list<StateID_type>& states_stack,
                                   std::unordered_set<StateID_type>& known_states,
                                   segmented_vector::SegmentedVector<unsigned int>& number_successors,
                                   SearchSpace& search_space,
                                   const StateSpaceProb& state_space) {

        for (const StateID_type parent_id: state_space.get_state_information(id)._parents()) {

            PLAJA_ASSERT(not known_states.count(parent_id) or number_successors[parent_id] > 0)

            if (--number_successors[parent_id] == 0) {
                if (known_states.insert(parent_id).second) {
                    search_space.get_node_info_cast(parent_id).set_inevitable_goal_path_state();
                    states_stack.push_front(parent_id);
                }
            }

        }
    }

}

void NN_Explorer::compute_inevitable_goal_path_states() {

    const auto& goal_states = searchSpace->get_goal_path_frontier();

    segmented_vector::SegmentedVector<unsigned int> number_successors; // Array of states by id (including unreachable states).
    const std::size_t state_space_size = stateSpace->information_size();
    for (StateID_type state_id = 0; state_id < state_space_size; ++state_id) {
        number_successors.resize(state_id + 1, stateSpace->get_state_information(state_id).number_of_successors());
    }

    std::unordered_set<StateID_type> inevitable_goal_path_states_set(goal_states.cbegin(), goal_states.cend()); // immediately add goal states
    std::forward_list<StateID_type> inevitable_goal_path_states;

    /* Goal states. */
    for (const StateID_type goal_state: goal_states) {
        NN_EXPLORER::compute_inevitable(goal_state, inevitable_goal_path_states, inevitable_goal_path_states_set, number_successors, *searchSpace, *stateSpace);
    }

    /* Goal path states. */
    while (not inevitable_goal_path_states.empty()) {
        const StateID_type goal_path_state = inevitable_goal_path_states.front();
        inevitable_goal_path_states.pop_front();
        NN_EXPLORER::compute_inevitable(goal_path_state, inevitable_goal_path_states, inevitable_goal_path_states_set, number_successors, *searchSpace, *stateSpace);
    }

    /* Inevitable goal path initial states. */
    unsigned int num_inevitable_goal_path_initial_states = 0;
    for (const StateID_type state_id: searchSpace->_initial_states()) {
        if (inevitable_goal_path_states_set.count(state_id)) { ++num_inevitable_goal_path_initial_states; }
    }

    if (searchStatistics) {
        searchStatistics->set_attr_unsigned(PLAJA::StatsUnsigned::InevitableGoalPathStates, inevitable_goal_path_states_set.size());
        searchStatistics->set_attr_unsigned(PLAJA::StatsUnsigned::InevitableGoalPathStatesStart, num_inevitable_goal_path_initial_states);
    }

}

/***********************************************************************************************************************/

namespace NN_EXPLORER {

    inline unsigned int compute_policy_goal_path_states(StateID_type id,
                                                        std::forward_list<StateID_type>& states_stack,
                                                        std::unordered_set<StateID_type>& known_states,
                                                        SearchSpace& search_space) {

        unsigned num_evitable_policy_goal_path_states(0);

        for (const StateID_type parent_id: search_space.get_node_info_cast(id).parents_under_policy()) {
            if (known_states.insert(parent_id).second) {
                auto& parent_node_info = search_space.get_node_info_cast(parent_id);
                if (not parent_node_info.is_inevitable_goal_path_state()) { ++num_evitable_policy_goal_path_states; }
                states_stack.push_front(parent_id);
            }
        }

        return num_evitable_policy_goal_path_states;

    }

}

void NN_Explorer::compute_policy_goal_path_states() {

    const auto& goal_states = searchSpace->get_goal_path_frontier();

    std::unordered_set<StateID_type> policy_goal_path_states_set(goal_states.cbegin(), goal_states.cend()); // Immediately add goal states.
    std::forward_list<StateID_type> policy_goal_path_states;
    unsigned int num_evitable_policy_goal_path_states = 0;

    /* Goal states. */
    for (const StateID_type goal_state: goal_states) {
        num_evitable_policy_goal_path_states += NN_EXPLORER::compute_policy_goal_path_states(goal_state, policy_goal_path_states, policy_goal_path_states_set, *searchSpace);
    }

    /* Goal path states. */
    while (not policy_goal_path_states.empty()) {
        const StateID_type goal_path_state = policy_goal_path_states.front();
        policy_goal_path_states.pop_front();
        num_evitable_policy_goal_path_states += NN_EXPLORER::compute_policy_goal_path_states(goal_path_state, policy_goal_path_states, policy_goal_path_states_set, *searchSpace);
    }

    /* Policy goal path initial states. */
    unsigned int num_policy_goal_path_initial_states = 0;
    for (const StateID_type state_id: searchSpace->_initial_states()) {
        if (policy_goal_path_states_set.count(state_id)) { ++num_policy_goal_path_initial_states; }
    }

    if (searchStatistics) {
        searchStatistics->set_attr_unsigned(PLAJA::StatsUnsigned::PolicyGoalPathStates, policy_goal_path_states_set.size());
        searchStatistics->set_attr_unsigned(PLAJA::StatsUnsigned::PolicyGoalPathStatesStart, num_policy_goal_path_initial_states);
        searchStatistics->set_attr_unsigned(PLAJA::StatsUnsigned::EvitablePolicyGoalPathStates, num_evitable_policy_goal_path_states); // User service.
    }
}
