//
// This file is part of the PlaJA code base (2019 - 2022).
//

#include "expansion_pa_base.h"
#include "../../../option_parser/option_parser.h"
#include "../../../parser/ast/expression/non_standard/predicate_abstraction_expression.h"
#include "../../../globals.h"
#include "../../factories/predicate_abstraction/search_engine_config_pa.h"
#include "../../information/jani2nnet/jani_2_nnet.h"
#include "../../fd_adaptions/search_statistics.h"
#include "../../smt/solver/smt_solver_z3.h"
#include "../../smt/successor_generation/op_applicability.h"
#include "../../successor_generation/successor_generator_c.h"
#include "../nn/nn_sat_checker/nn_sat_checker.h"
#include "../optimization/predicate_relations.h"
#include "../search_space/search_space_pa_base.h"
#include "../smt/model_z3_pa.h"
#include "../successor_generation/update/update_pa.h"
#include "../successor_generation/successor_generator_pa.h"
#include "../solution_checker.h"
#include "../solution_checker_instance_pa.h"

/**********************************************************************************************************************/

ExpansionPABase::ExpansionPABase(const SearchEngineConfigPA& config, SearchSpacePABase* search_space, const ModelZ3PA& model_z3):
    sharedStats(config.get_sharable_ptr<PLAJA::StatsBase>(PLAJA::SharableKey::STATS))
    , searchSpace(search_space)
    , solutionChecker(nullptr)
    , modelZ3(&model_z3)
    , solverOwn(new Z3_IN_PLAJA::SMTSolver(modelZ3->share_context(), sharedStats))
    , solver(solverOwn.get()) // solver((z3::tactic(modelZ3->get_context(), "solve-eqs") & z3::tactic(modelZ3->get_context(), "smt")).mk_solver(), *modelZ3),
    , reachChecker(new StateConditionCheckerPa(config, *modelZ3, not PLAJA_GLOBAL::enableTerminalStateSupport or PLAJA_GLOBAL::reachMayBeTerminal or not modelZ3->get_terminal() ? modelZ3->get_reach() : modelZ3->get_reach_non_terminal()))
    , terminalChecker(PLAJA_GLOBAL::enableTerminalStateSupport and config.get_bool_option(PLAJA_OPTION::check_for_pa_terminal_states) and modelZ3->get_terminal() ? new StateConditionCheckerPa(config, *modelZ3, modelZ3->get_terminal()) : nullptr)
    , predicateRelations(nullptr)
    , successorGenerator(nullptr)
    , opApp(PLAJA_GLOBAL::enableApplicabilityCache ? OpApplicability::construct(config) : nullptr)
    , incOpAppTest(config.get_bool_option(PLAJA_OPTION::inc_op_app_test) and not(PLAJA_GLOBAL::enableApplicabilityFiltering and opApp))
    , lazyNonTerminalPush(PLAJA_GLOBAL::enableTerminalStateSupport and not PLAJA_GLOBAL::doSelectionTest and config.get_bool_option(PLAJA_OPTION::lazyNonTerminalPush))
    , lazyInterfacePush(PLAJA_GLOBAL::enableApplicabilityFiltering and not PLAJA_GLOBAL::doSelectionTest and config.get_bool_option(PLAJA_OPTION::lazyPolicyInterfacePush))
    , mayBeTerminal(true)
    CONSTRUCT_IF_DEBUG(expectExactTerminalCheck(PLAJA_GLOBAL::enableTerminalStateSupport and config.has_bool_option(PLAJA_OPTION::terminal_as_preds) and config.get_bool_option(PLAJA_OPTION::terminal_as_preds)))
    , actionIsLearned(false)
    , currentSuccessors() {
    PLAJA_ASSERT(sharedStats)

    const bool do_app_filtering = modelZ3->get_nn_interface() and modelZ3->get_nn_interface()->_do_applicability_filtering();
    PLAJA_LOG_IF(not PLAJA_GLOBAL::enableApplicabilityFiltering and do_app_filtering, "Error: applicability filtering currently not supported!")
    PLAJA_ABORT_IF(do_app_filtering and not PLAJA_GLOBAL::enableApplicabilityFiltering)

    PLAJA_LOG_IF(not do_app_filtering and opApp, PLAJA_UTILS::to_red_string(PLAJA_UTILS::string_f("Using operator applicability cache for policy verification without applicability-filtering.")))

    // Add src variable (bound)s in predicates.
    const auto start = modelZ3->get_start();
    if (start) { modelZ3->register_src_bounds(*solver, *start); }
    const auto reach = modelZ3->get_reach();
    if (reach) { modelZ3->register_src_bounds(*solver, *reach); }
    for (PredicateIndex_type index = 0; index < modelZ3->get_number_predicates(); ++index) { modelZ3->register_src_bounds(*solver, *modelZ3->get_predicate(index)); }
    // solver->add_and_register_variables(modelZ3->get_src_vars());

    // Construct successor generation (predicate dependent).
    successorGenerator = SuccessorGeneratorPA::construct(*modelZ3, config);
    PLAJA_ASSERT(config.has_sharable(PLAJA::SharableKey::MODEL_Z3))
    PLAJA_ASSERT(config.get_sharable<ModelZ3>(PLAJA::SharableKey::MODEL_Z3).get() == modelZ3)
    config.get_sharable<ModelZ3>(PLAJA::SharableKey::MODEL_Z3)->share_action_ops(*successorGenerator); // TODO quick fix: access instance as shared object

    // Share op applicability cache:
    if constexpr (PLAJA_GLOBAL::enableApplicabilityCache) { if (opApp) { modelZ3->set_op_applicability(*opApp); } }

    // Optimizations:
    if (config.get_bool_option(PLAJA_OPTION::predicate_relations)) {

        if (config.has_sharable(PLAJA::SharableKey::PREDICATE_RELATIONS)) {
            predicateRelations = config.get_sharable<PredicateRelations>(PLAJA::SharableKey::PREDICATE_RELATIONS);
        } else {
            predicateRelations = config.set_sharable(PLAJA::SharableKey::PREDICATE_RELATIONS, std::make_shared<PredicateRelations>(config, *modelZ3, nullptr)); // predicates relation // OLD_INC: solver
        }

        PLAJA_ASSERT(predicateRelations->has_empty_ground_truth())

    }

    if (modelZ3->get_nn_interface()) {
        nnSatChecker = NN_SAT_CHECKER::construct_checker(*modelZ3->get_nn_interface(), config); // load nn-sat checker
        if constexpr (PLAJA_GLOBAL::enableApplicabilityCache) { if (opApp) { nnSatChecker->set_op_applicability_cache(*opApp); } }
    }

    // Solution checker.
    solutionChecker = std::make_unique<SolutionChecker>(config, modelZ3->get_model(), modelZ3->get_nn_interface());
    PLAJA_ASSERT(not modelZ3->get_nn_interface() or solutionChecker->has_policy())
    if (nnSatChecker) { nnSatChecker->set_solution_checker(solutionChecker.get()); }

    reset();
}

ExpansionPABase::~ExpansionPABase() = default;

/* */

void ExpansionPABase::increment() {
    /* Reset solver. */
    solver->clear();
    const auto start = modelZ3->get_start();
    if (start) { modelZ3->register_src_bounds(*solver, *start); }
    const auto reach = modelZ3->get_reach();
    if (reach) { modelZ3->register_src_bounds(*solver, *reach); }
    for (PredicateIndex_type index = 0; index < modelZ3->get_number_predicates(); ++index) { modelZ3->register_src_bounds(*solver, *modelZ3->get_predicate(index)); }

    /* Increment exploration structures. */

    if (reachChecker) { reachChecker->increment(); }

    if (terminalChecker) { terminalChecker->increment(); }

    successorGenerator->increment();

    if (predicateRelations) { predicateRelations->increment(); }

    if (nnSatChecker) { nnSatChecker->increment(); }

}

void ExpansionPABase::reset() {
    transitionStruct = PATransitionStructure();
    if constexpr (PLAJA_GLOBAL::enableTerminalStateSupport) { mayBeTerminal = true; }
    actionIsLearned = false;
    currentSuccessors.clear();
    solutionsStack.clear();
    push_solution(std::make_unique<StateValues>(solutionChecker->get_solution_constructor())); // dummy solution
    solutionStackSize.clear();
    newlyReached.clear();
}


//

const SuccessorGeneratorC& ExpansionPABase::_concrete_succ_gen() { return successorGenerator->get_concrete(); }

void ExpansionPABase::print_stats() const { if (nnSatChecker) { nnSatChecker->print_statistics(); } }

/** transition cache **************************************************************************************************/




/** solution cache ****************************************************************************************************/

void ExpansionPABase::push_solution(std::unique_ptr<StateValues>&& solution, std::unique_ptr<StateValues>&& target_solution) {
    PLAJA_ASSERT(solutionChecker->check_state(*solution))
    // PLAJA_ASSERT(not target_solution or solutionChecker->check_state(*target_solution)) // -> target may actually be invalid
    solutionsStack.emplace_back(std::move(solution), std::move(target_solution));
}

void ExpansionPABase::push_solution(std::unique_ptr<StateValues>&& solution) { push_solution(std::move(solution), nullptr); }

void ExpansionPABase::push_solution(StateValues&& solution) { push_solution(std::make_unique<StateValues>(std::move(solution))); }

void ExpansionPABase::pop_down(std::size_t stack_size) {
    PLAJA_ASSERT(solutionsStack.size() >= stack_size)
    while (solutionsStack.size() > stack_size) { solutionsStack.pop_back(); }
}

const StateValues& ExpansionPABase::get_last_solution() const {
    PLAJA_ASSERT(not solutionsStack.empty())
    PLAJA_ASSERT(solutionsStack.back().first)
    return *solutionsStack.back().first;
}

const StateValues* ExpansionPABase::get_last_target_solution() const {
    PLAJA_ASSERT(not solutionsStack.empty())
    return solutionsStack.back().second.get();
}

/** incremental *******************************************************************************************************/

void ExpansionPABase::push_source(std::unique_ptr<AbstractState>&& source) {
    PUSH_LAP(sharedStats, PLAJA::StatsDouble::PA_PUSH_POP_SRC_TIME)

    PLAJA_ASSERT(not PLAJA_GLOBAL::enableApplicabilityCache or not opApp or opApp->empty())

    transitionStruct.set_source(std::move(source));
    solver->push();
    modelZ3->add_to(*transitionStruct._source(), *solver);
    if constexpr (PLAJA_GLOBAL::enableTerminalStateSupport) {
        PLAJA_ASSERT(not expectExactTerminalCheck or not mayBeTerminal)
        if (mayBeTerminal) { modelZ3->exclude_terminal(*solver, 0); }
    }

    if (nnSatChecker) {
        nnSatChecker->push();
        nnSatChecker->add_source_state(*transitionStruct._source());
        if constexpr (PLAJA_GLOBAL::enableTerminalStateSupport) { if (not lazyNonTerminalPush and mayBeTerminal) { modelZ3->exclude_terminal(*solver, 0); } } // TODO optimize?
    }

    POP_LAP(sharedStats, PLAJA::StatsDouble::PA_PUSH_POP_SRC_TIME)
}

void ExpansionPABase::pop_source() {
    PUSH_LAP(sharedStats, PLAJA::StatsDouble::PA_PUSH_POP_SRC_TIME)

    if (nnSatChecker) {
        nnSatChecker->unset_source_state();
        nnSatChecker->pop();
    }

    solver->pop();
    transitionStruct.unset_source();

    if constexpr (PLAJA_GLOBAL::enableApplicabilityCache) { if (opApp) { opApp->reset(); } }

    POP_LAP(sharedStats, PLAJA::StatsDouble::PA_PUSH_POP_SRC_TIME)
}

inline bool ExpansionPABase::push_output_interface() {
    if constexpr (PLAJA_GLOBAL::enableApplicabilityCache and not PLAJA_GLOBAL::doSelectionTest) { if (opApp) { opApp->set_self_applicability(true); } }
    const bool rlt = nnSatChecker->add_output_interface();
    if constexpr (PLAJA_GLOBAL::enableApplicabilityCache and not PLAJA_GLOBAL::doSelectionTest) { if (opApp) { opApp->unset_self_applicability(); } }
    return rlt;
}

bool ExpansionPABase::push_action(ActionLabel_type action_label) {
    PUSH_LAP(sharedStats, PLAJA::StatsDouble::PA_PUSH_POP_ACTION_LABEL_TIME)
    transitionStruct.set_action_label(action_label);

    bool rlt = true;

    if (nnSatChecker) {
        actionIsLearned = nnSatChecker->is_action_selectable(action_label);
        if (actionIsLearned) {
            push_solution(); // to properly POP selection test solution
            nnSatChecker->push();
            nnSatChecker->set_label(action_label);
            if (not lazyInterfacePush) { rlt = push_output_interface(); }
        }
    }

    POP_LAP(sharedStats, PLAJA::StatsDouble::PA_PUSH_POP_ACTION_LABEL_TIME)

    return rlt;
}

void ExpansionPABase::pop_action() {
    PUSH_LAP(sharedStats, PLAJA::StatsDouble::PA_PUSH_POP_ACTION_LABEL_TIME)
    if (nnSatChecker && actionIsLearned) {
        nnSatChecker->unset_label();
        nnSatChecker->pop();
        pop_solution(); // POP selection test solution.
    }
    actionIsLearned = false;
    transitionStruct.unset_action_label();
    POP_LAP(sharedStats, PLAJA::StatsDouble::PA_PUSH_POP_ACTION_LABEL_TIME)
}

void ExpansionPABase::set_action_op(const ActionOpPA& action_op) {
    PUSH_LAP(sharedStats, PLAJA::StatsDouble::PA_PUSH_POP_ACTION_OP_TIME)
    transitionStruct.set_action_op(action_op);
    push_solution(); // to properly POP applicability test solution
    POP_LAP(sharedStats, PLAJA::StatsDouble::PA_PUSH_POP_ACTION_OP_TIME)
}

void ExpansionPABase::unset_action_op() {
    PUSH_LAP(sharedStats, PLAJA::StatsDouble::PA_PUSH_POP_ACTION_OP_TIME)
    pop_solution();
    transitionStruct.unset_action_op();
    POP_LAP(sharedStats, PLAJA::StatsDouble::PA_PUSH_POP_ACTION_OP_TIME)
}

void ExpansionPABase::push_update(UpdateIndex_type update_index, const UpdatePA& update_pa) {
    PUSH_LAP(sharedStats, PLAJA::StatsDouble::PA_PUSH_POP_UPD_TIME)

    transitionStruct.set_update(update_index, update_pa);

    solver->push();
    update_pa.add_to_solver(*solver);

    if (nnSatChecker and actionIsLearned) {
        nnSatChecker->push();
        nnSatChecker->set_update(update_index);
        nnSatChecker->add_update();
    }

    if constexpr (PLAJA_GLOBAL::reuseSolutions) {
        PLAJA_ASSERT(not get_last_target_solution())
        PLAJA_ASSERT(transitionStruct.has_action_label())
        PLAJA_ASSERT(transitionStruct.has_action_op())
        PLAJA_ASSERT(transitionStruct.has_update())

        push_solution();

        // Sample target solution for update.
        /*
         * TODO
         *  Since we do not cache the solution of the app test in non-inc exploration,
         *  get_last_solution() returns a default state.
         *  The operator guard may not be applicable and hence out-of-bounds-updates are no modelling error.
         *  Moreover, the SMT solver may return an unreachable state as solution, where, again, an oob-successor is no modelling error.
         *  We just suppress the oob check here as a quick fix.
         *  This is dirty. There really is no point in trying to reuse the cached "non-solution".
         *  However, the long term long is direct solution-to-transition usage, which will make the solution stake obsolete anyway.
         */
        STMT_IF_DEBUG(StateBase::suppress_is_valid(true);)
        push_solution(std::make_unique<StateValues>(get_last_solution()), std::make_unique<StateValues>(solutionChecker->compute_successor_solution(get_last_solution(), solutionChecker->get_action_op(transitionStruct._action_op_id()), transitionStruct._update_index())));
        STMT_IF_DEBUG(StateBase::suppress_is_valid(false);)
    }

    POP_LAP(sharedStats, PLAJA::StatsDouble::PA_PUSH_POP_UPD_TIME)
}

void ExpansionPABase::pop_update() {
    PUSH_LAP(sharedStats, PLAJA::StatsDouble::PA_PUSH_POP_UPD_TIME)

    pop_solution();

    if (nnSatChecker and actionIsLearned) {
        nnSatChecker->unset_update();
        nnSatChecker->pop();
    }

    solver->pop();
    transitionStruct.unset_update();

    POP_LAP(sharedStats, PLAJA::StatsDouble::PA_PUSH_POP_UPD_TIME)
}

void ExpansionPABase::nn_sat_push_action_op() {
    PUSH_LAP(sharedStats, PLAJA::StatsDouble::PA_PUSH_POP_ACTION_OP_TIME)
    if (nnSatChecker and actionIsLearned) {
        push_solution(); // To properly POP applicability test solution.
        nnSatChecker->push();
        nnSatChecker->set_operator(transitionStruct._action_op_id());
        nnSatChecker->add_guard();
        if (lazyNonTerminalPush) { nnSatChecker->add_non_terminal_condition(); }
        if (lazyInterfacePush) { push_output_interface(); }
    }
    POP_LAP(sharedStats, PLAJA::StatsDouble::PA_PUSH_POP_ACTION_OP_TIME)
}

void ExpansionPABase::nn_sat_pop_action_op() {
    PUSH_LAP(sharedStats, PLAJA::StatsDouble::PA_PUSH_POP_ACTION_OP_TIME)
    if (nnSatChecker and actionIsLearned) {
        nnSatChecker->unset_operator();
        nnSatChecker->pop();
        pop_solution(); // POP applicability test solution.
    }
    POP_LAP(sharedStats, PLAJA::StatsDouble::PA_PUSH_POP_ACTION_OP_TIME)
}

/**********************************************************************************************************************/

bool ExpansionPABase::applicability_test() {
    sharedStats->inc_attr_unsigned(PLAJA::StatsUnsigned::APP_TESTS);

    PUSH_LAP(sharedStats, PLAJA::StatsDouble::TIME_APP_TEST)

    if constexpr (PLAJA_GLOBAL::reuseSolutions and PLAJA_GLOBAL::doSelectionTest) {
        if (solutionChecker->check_applicability(get_last_solution(), solutionChecker->get_action_op(transitionStruct._action_op_id()), transitionStruct._source())) {
            sharedStats->inc_attr_unsigned(PLAJA::StatsUnsigned::REUSED_SOLUTIONS);
            POP_LAP(sharedStats, PLAJA::StatsDouble::TIME_APP_TEST)
            return true; // TODO may be beneficial to still (incrementally) check solver.
        }
    }

    const bool rlt = solver->check();
    if (rlt) {
        if constexpr (PLAJA_GLOBAL::reuseSolutions) {
            auto instance = solver->extract_solution_via_checker(*modelZ3, std::make_unique<SolutionCheckerInstancePa>(*solutionChecker, transitionStruct));
            PLAJA_ASSERT(instance->get_solution_size() == 1) // No target expected.
            push_solution(instance->retrieve_solution(0));
            PLAJA_ASSERT(solutionChecker->check_applicability(get_last_solution(), solutionChecker->get_action_op(transitionStruct._action_op_id()), transitionStruct._source()))
        }
    } else {
        searchSpace->set_excluded_op(*transitionStruct._source(), transitionStruct._action_op_id()); // Mark excluded.
        sharedStats->inc_attr_unsigned(PLAJA::StatsUnsigned::APP_TESTS_UNSAT);
    }

    if constexpr (PLAJA_GLOBAL::enableApplicabilityCache) {
        if (opApp) { opApp->set_applicability(transitionStruct._action_op_id(), rlt); }
    }

    POP_LAP(sharedStats, PLAJA::StatsDouble::TIME_APP_TEST)
    return rlt;
}

bool ExpansionPABase::transition_test() {
    sharedStats->inc_attr_unsigned(PLAJA::StatsUnsigned::TRANSITION_TESTS);
    PUSH_LAP(sharedStats, PLAJA::StatsDouble::TIME_TRANSITION_TEST)

    if constexpr (PLAJA_GLOBAL::reuseSolutions) {
        // Note: not reusing solutions for non-NN SMT-tests seems to outperform,
        // i.e., reusing should/may be disabled when computing standard PA.
        // When computing PPA: Reusing significantly improves performance.
        // Presumably in this case reusing solutions for non-NN SMT-tests is ok
        // -- (more expensive) solution caching is done anyway, we only have to check the solution (which is cheap).
        // Moreover, if *not* reusing solutions for non-NN tests, we could end up pushing a non-policy solution to the stack, hiding a valid policy solution.

        PLAJA_ASSERT(get_last_target_solution())

        STMT_IF_DEBUG(StateBase::suppress_is_valid(true);)
        if (solutionChecker->check_transition(get_last_solution(), *transitionStruct._source(), solutionChecker->get_action_op(transitionStruct._action_op_id()), transitionStruct._update_index(), *transitionStruct._target(), get_last_target_solution())) {
            sharedStats->inc_attr_unsigned(PLAJA::StatsUnsigned::REUSED_SOLUTIONS);
            POP_LAP(sharedStats, PLAJA::StatsDouble::TIME_TRANSITION_TEST)
            return true;
        }
        STMT_IF_DEBUG(StateBase::suppress_is_valid(false);)

    }

    solver->push();

    transitionStruct._update()->add_predicates_to_solver(*solver, *transitionStruct._target());
    const bool rlt = solver->check();
    if (rlt) {
        if constexpr (PLAJA_GLOBAL::reuseSolutions) {
            auto instance = solver->extract_solution_via_checker(*modelZ3, std::make_unique<SolutionCheckerInstancePa>(*solutionChecker, transitionStruct));
            PLAJA_ASSERT(instance->get_solution_size() == 2)
            push_solution(instance->retrieve_solution(0), instance->retrieve_solution(1));
            PLAJA_ASSERT(solutionChecker->check_transition(get_last_solution(), *transitionStruct._source(), solutionChecker->get_action_op(transitionStruct._action_op_id()), transitionStruct._update_index(), *transitionStruct._target(), get_last_target_solution()))
        }
    } else { sharedStats->inc_attr_unsigned(PLAJA::StatsUnsigned::TRANSITION_TESTS_UNSAT); }

    solver->pop();

    POP_LAP(sharedStats, PLAJA::StatsDouble::TIME_TRANSITION_TEST)
    return rlt;
}

bool ExpansionPABase::nn_sat_selection_test() {
    PLAJA_ASSERT(PLAJA_GLOBAL::doSelectionTest)

    if (not(nnSatChecker and nnSatChecker->is_action_selectable(transitionStruct._action_label()))) { return true; }

    PUSH_LAP(sharedStats, PLAJA::StatsDouble::TIME_NN_SELECTION_TEST)
    sharedStats->inc_attr_unsigned(PLAJA::StatsUnsigned::NN_SELECTION_TESTS);

    bool rlt;
    if constexpr (PLAJA_GLOBAL::doRelaxedSelectionTestOnly) { rlt = nnSatChecker->check_relaxed(); }
    else { rlt = nnSatChecker->check(); }

    if (rlt) {

        if constexpr (PLAJA_GLOBAL::debug and not PLAJA_GLOBAL::doRelaxedSelectionTestOnly) {
            PLAJA_ASSERT(not nnSatChecker->has_solution() or solutionChecker->check_policy(nnSatChecker->get_last_solution(), transitionStruct._action_label(), transitionStruct._source()))
        }

        if constexpr (PLAJA_GLOBAL::reuseSolutions) {
            if (nnSatChecker->has_solution()) { push_solution(nnSatChecker->release_solution()); }
        }

    } else {
        searchSpace->set_excluded_label(*transitionStruct._source(), transitionStruct._action_label()); // mark excluded
        sharedStats->inc_attr_unsigned(PLAJA::StatsUnsigned::NN_SELECTION_TESTS_UNSAT);
    }

    POP_LAP(sharedStats, PLAJA::StatsDouble::TIME_NN_SELECTION_TEST)
    return rlt;
}

bool ExpansionPABase::nn_sat_applicability_test() {
    PLAJA_ASSERT(PLAJA_GLOBAL::doNNApplicabilityTest)

    if (not(nnSatChecker and nnSatChecker->is_action_selectable(transitionStruct._action_label()))) { return true; }

    PUSH_LAP(sharedStats, PLAJA::StatsDouble::TIME_NN_APP_TEST)
    sharedStats->inc_attr_unsigned(PLAJA::StatsUnsigned::NN_APP_TESTS);

    if constexpr (PLAJA_GLOBAL::reuseSolutions) {
        if (solutionChecker->check_policy_applicability(get_last_solution(), solutionChecker->get_action_op(transitionStruct._action_op_id()), transitionStruct._source())) {
            sharedStats->inc_attr_unsigned(PLAJA::StatsUnsigned::REUSED_SOLUTIONS);
            POP_LAP(sharedStats, PLAJA::StatsDouble::TIME_NN_APP_TEST)
            return true;
        }
    }

    bool rlt;
    if constexpr (PLAJA_GLOBAL::doRelaxedNNApplicabilityTestOnly) { rlt = nnSatChecker->check_relaxed(); }
    else { rlt = nnSatChecker->check(); }

    if (rlt) {

        if constexpr (PLAJA_GLOBAL::debug and not PLAJA_GLOBAL::doRelaxedNNApplicabilityTestOnly) {
            PLAJA_ASSERT(not nnSatChecker->has_solution() or solutionChecker->check_policy_applicability(nnSatChecker->get_last_solution(), successorGenerator->get_action_op_concrete(transitionStruct._action_op_id()), transitionStruct._source()))
        }

        if constexpr (PLAJA_GLOBAL::reuseSolutions) {
            if (nnSatChecker->has_solution()) { push_solution(nnSatChecker->release_solution()); }
        }

    } else {
        sharedStats->inc_attr_unsigned(PLAJA::StatsUnsigned::NN_APP_TESTS_UNSAT);
    }

    POP_LAP(sharedStats, PLAJA::StatsDouble::TIME_NN_APP_TEST)
    return rlt;
}

bool ExpansionPABase::nn_sat_transition_test() {

    if (not(nnSatChecker and nnSatChecker->is_action_selectable(transitionStruct._action_label()))) { return true; } // No nn transition test necessary.

    PUSH_LAP(sharedStats, PLAJA::StatsDouble::TIME_NN_TRANSITION_TEST)
    sharedStats->inc_attr_unsigned(PLAJA::StatsUnsigned::NN_TRANSITION_TESTS);

    if constexpr (PLAJA_GLOBAL::reuseSolutions) {

        const auto& last_solution = get_last_solution();
        const auto& last_target_solution = get_last_target_solution();
        PLAJA_ASSERT(last_target_solution)

        if (solutionChecker->check_policy_transition(last_solution, *transitionStruct._source(), solutionChecker->get_action_op(transitionStruct._action_op_id()), transitionStruct._update_index(), *transitionStruct._target(), last_target_solution)) {
            sharedStats->inc_attr_unsigned(PLAJA::StatsUnsigned::REUSED_SOLUTIONS);
            POP_LAP(sharedStats, PLAJA::StatsDouble::TIME_NN_TRANSITION_TEST)
            searchSpace->add_transition(transitionStruct, last_solution, *last_target_solution); // TODO For now we only add transitions possible under policy-restriction (see also below).
            return true;
        }
    }

    nnSatChecker->push();
    nnSatChecker->add_target_state(*transitionStruct._target());
    const bool rlt = nnSatChecker->check();
    nnSatChecker->unset_target_state();
    nnSatChecker->pop();

    if (rlt) {

        if constexpr (PLAJA_GLOBAL::debug) {
            STMT_IF_DEBUG(PLAJA_GLOBAL::push_additional_outputs_frame(true);)
            PLAJA_ASSERT(not nnSatChecker->has_solution()
                         or (nnSatChecker->get_target_solution() and
                             solutionChecker->check_policy_transition(nnSatChecker->get_last_solution(), *transitionStruct._source(), solutionChecker->get_action_op(transitionStruct._action_op_id()), transitionStruct._update_index(), *transitionStruct._target(), nnSatChecker->get_target_solution())
                         ))
            STMT_IF_DEBUG(PLAJA_GLOBAL::pop_additional_outputs_frame();)
        }

        if (nnSatChecker->has_solution()) { // E.g., due to handling unknown as sat. TODO solution extraction not supported by MarabouInZ3Checker.
            PLAJA_ASSERT(nnSatChecker->get_target_solution()) // Redundant, see above. Though, why not ...
            searchSpace->add_transition(transitionStruct, *nnSatChecker->release_solution(), *nnSatChecker->release_target_solution());
            // Nothing to reuse it for ... // if constexpr (PLAJA_GLOBAL::reuseSolutions) { push_solution(nnSatChecker->release_solution(), nnSatChecker->release_target_solution()); }
        }

    } else {
        sharedStats->inc_attr_unsigned(PLAJA::StatsUnsigned::NN_TRANSITION_TESTS_UNSAT);
        searchSpace->exclude_transition(transitionStruct); // TODO currently we only cache transition excluded under policy-restriction.
    }

    POP_LAP(sharedStats, PLAJA::StatsDouble::TIME_NN_TRANSITION_TEST)
    return rlt;
}
