//
// This file is part of the PlaJA code base (2019 - 2022).
//

#ifndef PLAJA_EXPANSION_PA_BASE_H
#define PLAJA_EXPANSION_PA_BASE_H

#include <unordered_set>
#include <list>
#include "../../../stats/forward_stats.h"
#include "../../../assertions.h"
#include "../../factories/predicate_abstraction/forward_factories_pa.h"
#include "../../fd_adaptions/search_engine.h"
#include "../../information/forward_information.h"
#include "../../smt/base/forward_solution_check.h"
#include "../../smt/forward_smt_z3.h"
#include "../../successor_generation/forward_successor_generation.h"
#include "../optimization/forward_optimization_pa.h"
#include "../smt/forward_smt_pa.h"
#include "../successor_generation/forward_succ_gen_pa.h"
#include "../smt/state_condition_checker_pa.h"
#include "../pa_transition_structure.h"
#include "../using_predicate_abstraction.h"
#include "../search_space/forward_search_space.h"

class NNSatChecker;

class SearchPABase;

class ExpansionPABase {

protected:
    PLAJA::StatsBase* sharedStats;
    SearchSpacePABase* searchSpace;
    std::unique_ptr<const SolutionChecker> solutionChecker;

    const ModelZ3PA* modelZ3;
    std::unique_ptr<Z3_IN_PLAJA::SMTSolver> solverOwn;
    Z3_IN_PLAJA::SMTSolver* solver;
    std::unique_ptr<StateConditionCheckerPa> reachChecker; // To check the reach condition.
    std::unique_ptr<StateConditionCheckerPa> terminalChecker; // To check whether an abstract state fulfills terminal state condition.

    std::shared_ptr<PredicateRelations> predicateRelations; // Prune branches based on binary entailment relations of predicates
    std::unique_ptr<SuccessorGeneratorPA> successorGenerator;
    std::unique_ptr<OpApplicability> opApp; // Optimization for applicability filtering.
    std::unique_ptr<NNSatChecker> nnSatChecker;

    bool incOpAppTest;
    bool lazyNonTerminalPush; // Push non-terminal state condition lazily, i.e., after selection and applicability condition (only for policy solver).
    bool lazyInterfacePush;

    PATransitionStructure transitionStruct;
    bool mayBeTerminal;
    FIELD_IF_DEBUG(bool expectExactTerminalCheck;)
    bool actionIsLearned;
    std::unordered_set<StateID_type> currentSuccessors; // May be maintained per state, label, operator, ...

    // Solution cache.
    std::list<std::pair<std::unique_ptr<StateValues>, std::unique_ptr<StateValues>>> solutionsStack;
    std::list<std::size_t> solutionStackSize;

    // Transition cache.
    std::list<StateID_type> newlyReached;

/** solution cache ****************************************************************************************************/

    [[nodiscard]] inline std::size_t get_solution_stack_size() const { return solutionsStack.size(); }

    void push_solution(std::unique_ptr<StateValues>&& solution, std::unique_ptr<StateValues>&& target_solution);

    void push_solution(std::unique_ptr<StateValues>&& solution);

    void push_solution(StateValues&& solution);

    void pop_down(std::size_t stack_size);

    [[nodiscard]] const StateValues& get_last_solution() const;

    [[nodiscard]] const StateValues* get_last_target_solution() const;

    /* To properly POP applicability test solution. */
    inline void push_solution() {
        if constexpr (PLAJA_GLOBAL::reuseSolutions) { solutionStackSize.push_back(get_solution_stack_size()); }
    }

    inline void pop_solution() {
        if constexpr (PLAJA_GLOBAL::reuseSolutions) {
            pop_down(solutionStackSize.back());
            solutionStackSize.pop_back();
        }
    }

/** incremental *******************************************************************************************************/

    void push_source(std::unique_ptr<AbstractState>&& source);
    void pop_source();

    bool push_output_interface();

    [[nodiscard]] bool push_action(ActionLabel_type action_label);
    void pop_action();

    void set_action_op(const ActionOpPA& action_op);
    void unset_action_op();

    void push_update(UpdateIndex_type update_index, const UpdatePA& update_pa);
    void pop_update();

    void nn_sat_push_action_op();
    void nn_sat_pop_action_op();

/** smt transition tests **********************************************************************************************/

    bool applicability_test();
    bool transition_test();

    // void stalling_detection(const State& current_state);
    bool nn_sat_selection_test();
    bool nn_sat_applicability_test();
    bool nn_sat_transition_test();

/**********************************************************************************************************************/

public:
    explicit ExpansionPABase(const SearchEngineConfigPA& config, SearchSpacePABase* search_space, const ModelZ3PA& model_z3);
    virtual ~ExpansionPABase();
    DELETE_CONSTRUCTOR(ExpansionPABase)

    // auxiliaries
    inline bool check_reach(const AbstractState& state) { return reachChecker->check(state); }

    [[nodiscard]] const SuccessorGeneratorC& _concrete_succ_gen();
    void print_stats() const;

    FCT_IF_DEBUG(NNSatChecker* _nn_sat_checker() { return nnSatChecker.get(); })

    [[nodiscard]] const NNSatChecker* _nn_sat_checker() const { return nnSatChecker.get(); }

    void increment();
    void reset();
    virtual SearchEngine::SearchStatus step(StateID_type src_id) = 0;

    [[nodiscard]] const std::list<StateID_type>& _newly_reached() const { return newlyReached; }
};

#endif //PLAJA_EXPANSION_PA_BASE_H
