//
// This file is part of the PlaJA code base (2019 - 2020).
//

#pragma GCC diagnostic push
#pragma ide diagnostic ignored "Simplify"

#include "search_pa_base.h"
#include "../../../stats/stats_base.h"
#include "../../factories/predicate_abstraction/search_engine_config_pa.h"
#include "../../information/property_information.h"
#include "../heuristic/heuristic_state_queue.h"
#include "../pa_states/abstract_path.h"
#include "../search_space/search_space_pa_base.h"
#include "../smt/model_z3_pa.h"
#include "../search_statistics_pa.h"
#include "expansion_pa_base.h"

// extern:
namespace MODEL_MARABOU_PA { extern void make_sharable(const PLAJA::Configuration& config); }

SearchPABase::SearchPABase(const SearchEngineConfigPA& config):
    SearchEngine(config)
    , statsHeuristic(config.get_sharable_ptr<PLAJA::StatsBase>(PLAJA::SharableKey::STATS_HEURISTIC))
    , searchSpace(nullptr)
    , expansionPa(nullptr)
    , cachedStartStates(config.is_flag_set(PLAJA_OPTION::bfs_per_start_state) ? new std::list<StateID_type>() : nullptr)
    // solver(*modelZ3),
    // goalChecker(new StateConditionCheckerPA(*modelZ3)),
    // predicateRelations(nullptr),
    // jani2NNet(propertyInfo->get_nn_interface() ? propertyInfo->get_nn_interface() : nullptr)
    // stallingDetectionEnabled(optionParser.is_flag_set(PLAJA_OPTION::stalling_detection)),
    , pathLength(-1) {
    PLAJA_ASSERT(config.has_sharable_ptr(PLAJA::SharableKey::STATS) and config.get_sharable_ptr<PLAJA::StatsBase>(PLAJA::SharableKey::STATS) == searchStatistics.get());

    SEARCH_PA_BASE::add_stats(*searchStatistics);

    if (config.has_sharable(PLAJA::SharableKey::MODEL_Z3) and dynamic_cast<const ModelZ3PA*>(config.get_sharable<ModelZ3>(PLAJA::SharableKey::MODEL_Z3).get())) {
        auto model_z3 = config.get_sharable<ModelZ3>(PLAJA::SharableKey::MODEL_Z3);
        modelZ3 = std::static_pointer_cast<ModelZ3PA>(model_z3);
    } else {
        if (propertyInfo->get_nn_interface()) { MODEL_MARABOU_PA::make_sharable(config); } // TODO as long as ModelZ3 loads ModelMarabou for NN encoding
        modelZ3 = std::make_shared<ModelZ3PA>(config);
        config.set_sharable<ModelZ3>(PLAJA::SharableKey::MODEL_Z3, modelZ3);
    }

    // search space:
    searchSpace = std::make_unique<SearchSpacePABase>(config, *modelZ3); // construct search space (note: model Z3 needs predicates to be set and state info to be computed)

    // frontier:
    construct_heuristic(config);

    // set stats properly:
    searchStatistics->set_attr_unsigned(PLAJA::StatsUnsigned::PREDICATES, modelZ3->get_number_predicates());
    searchStatistics->set_attr_int(PLAJA::StatsInt::PATH_LENGTH, pathLength);

}

SearchPABase::~SearchPABase() = default;

/**********************************************************************************************************************/

void SearchPABase::construct_heuristic(const SearchEngineConfigPA& config) {
#ifdef ENABLE_HEURISTIC_SEARCH_PA
    config.set_sharable_ptr<PLAJA::StatsBase>(PLAJA::SharableKey::STATS_HEURISTIC, searchStatistics.get()); // set stats for heuristic
    frontier = HeuristicStateQueue::make_state_queue(config);
#else
    frontier = std::make_unique<FIFOStateQueue>();
#endif
}

/**********************************************************************************************************************/

bool SearchPABase::add_start_state() {
    if (not cachedStartStates) { return false; }

    while (not cachedStartStates->empty()) {
        // pop start state
        const StateID_type start_id = cachedStartStates->front();
        cachedStartStates->pop_front();

        // get search node of start state
        auto start_node = searchSpace->get_node(start_id);
        if (not start_node.is_reached()) { // not already reached (and expanded)
            // set reached
            start_node.set_reached_init();
            searchStatistics->inc_attr_unsigned(PLAJA::StatsUnsigned::REACHABLE_STATES);

            // add to frontier
            frontier->push(start_node);

            // state space
            // stateSpace->add_state_information(start_id);

            if (expansionPa->check_reach(*searchSpace->get_abstract_state(start_id))) {
                std::cout << "Start state is a goal!" << std::endl;
                start_node.set_goal();
                searchStatistics->inc_attr_unsigned(PLAJA::StatsUnsigned::GOAL_STATES);
                searchSpace->add_goal_path_frontier(start_id);
            }

            return true;
        } else {
            // still print goal state warning:
            if (searchSpace->has_goal_path(start_id)) { std::cout << "Start state is a goal!" << std::endl; }
        }
    }

    return false;
}

//

bool SearchPABase::has_goal_path() const {
    PLAJA_ASSERT(this->get_status() == SearchStatus::FINISHED)
    return not searchSpace->goal_path_frontier_empty() or pathLength != -1; // check if path exists
}

AbstractPath SearchPABase::extract_goal_path() const {
    PLAJA_ASSERT(not searchSpace->goal_path_frontier_empty())
    auto path = searchSpace->extract_goal_path(searchSpace->goal_frontier_state());
    CHECK_NARROWING_CONVERSION(path.get_size(), int)
    pathLength = static_cast<int>(path.get_size());
    searchStatistics->set_attr_int(PLAJA::StatsInt::PATH_LENGTH, pathLength);
    return path;
}

/** incremental *******************************************************************************************************/

void SearchPABase::increment() {
    searchStatistics->reset(); // reset stats
    modelZ3->increment(); // increment model *first*
    //
    // reset solver and successor generation:
    // solver.clear();
    // solver.add_unregistered_variables(&modelZ3->get_pa_expression());
    // successorGenerator->increment(solver);
    // solver.unregister_all_vars();
    //
    searchSpace->increment();
    //
    frontier->clear();
    frontier->increment();
    //
    // if (predicateRelations) { predicateRelations->increment(); }
    //
    // solution checker (if re-constructing succ gen)
    // solutionChecker = std::make_unique<SolutionChecker>(*model, modelZ3->get_pa_expression(), jani2NNet, &successorGenerator->get_concrete());
    //
    // if (nnSatChecker) { nnSatChecker->increment(); }
    //     nnSatChecker = NN_SAT_CHECKER::construct_checker(*jani2NNet, generate_information());
    //     nnSatChecker->set_statistics(&searchStatistics->share_nn_sat_stats());
    // }
    expansionPa->increment();

    if (cachedStartStates) { cachedStartStates->clear(); }
    pathLength = -1;

    // reset stats properly:
    searchStatistics->set_attr_unsigned(PLAJA::StatsUnsigned::PREDICATES, modelZ3->get_number_predicates());
    searchStatistics->set_attr_int(PLAJA::StatsInt::PATH_LENGTH, pathLength);
}

STMT_IF_DEBUG(NNSatChecker* SearchPABase::get_nn_sat_checker() { return expansionPa->_nn_sat_checker(); })

/** statistics ********************************************************************************************************/

void SearchPABase::update_statistics() const { searchSpace->update_statistics(); }

void SearchPABase::print_statistics() const {
    searchStatistics->print_statistics();
    searchSpace->print_statistics();
    // additional stats:
    expansionPa->print_stats();
    frontier->print_statistics();
}

void SearchPABase::stats_to_csv(std::ofstream& file) const {
    searchStatistics->stats_to_csv(file);
    searchSpace->stats_to_csv(file);
    frontier->stats_to_csv(file);
}

void SearchPABase::stat_names_to_csv(std::ofstream& file) const {
    searchStatistics->stat_names_to_csv(file);
    SearchSpacePABase::stat_names_to_csv(file);
    frontier->stat_names_to_csv(file);
}

#pragma GCC diagnostic pop