//
// This file is part of the PlaJA code base (2019 - 2023).
//

#include "inc_state_space_reg.h"
#include "../optimization/predicate_relations.h"
#include "../pa_states/abstract_state.h"
#include "../pa_states/predicate_state.h"

#ifdef USE_INC_REG
#ifdef USE_REF_REG

SEARCH_SPACE_PA::IncStateSpaceReg::IncStateSpaceReg(const PLAJA::Configuration& config, std::shared_ptr<StateRegistryPa> state_reg_pa):
    IncStateSpace(config, state_reg_pa)
    , stateRegistryPa(std::move(state_reg_pa)) {

}

#else

SEARCH_SPACE_PA::IncStateSpaceReg::IncStateSpaceReg(const PLAJA::Configuration& /*config*/, std::shared_ptr<StateRegistryPa> state_reg_pa):
    stateRegistryPa(std::move(state_reg_pa)) {}

#endif
#else

SEARCH_SPACE_PA::IncStateSpaceReg::IncStateSpaceReg(const PLAJA::Configuration& /*config*/, std::shared_ptr<StateRegistryPa> state_reg_pa):
    stateRegistryPa(std::move(state_reg_pa)) {}

#endif

SEARCH_SPACE_PA::IncStateSpaceReg::~IncStateSpaceReg() = default;

/**********************************************************************************************************************/

void SEARCH_SPACE_PA::IncStateSpaceReg::set_registry(PredicatesNumber_type num_preds) {

    if (not nodeRegistries.count(num_preds)) {

        nodeRegistries.emplace(num_preds, std::make_unique<SourceNodeRegistry>());

    }

}

void SEARCH_SPACE_PA::IncStateSpaceReg::update_stats() const { STMT_IF_DEBUG(if (incStateSpaceRef) { incStateSpaceRef->update_stats(); }) }

void SEARCH_SPACE_PA::IncStateSpaceReg::increment() {

    const auto num_preds = stateRegistryPa->get_predicates_size();

    PLAJA_ASSERT(not nodeRegistries.count(num_preds))

    nodeRegistries.emplace(num_preds, std::make_unique<SourceNodeRegistry>());

    STMT_IF_DEBUG(if (incStateSpaceRef) { incStateSpaceRef->increment(); })

}

void SEARCH_SPACE_PA::IncStateSpaceReg::clear_coarser() {
    const auto num_preds = stateRegistryPa->get_predicates_size();

    for (auto it = nodeRegistries.begin(); it != nodeRegistries.end();) {

        if (it->first != num_preds) { it = nodeRegistries.erase(it); }

        else { ++it; }

    }

    STMT_IF_DEBUG(if (incStateSpaceRef) { incStateSpaceRef->clear_coarser(); })

}

/**********************************************************************************************************************/

void SEARCH_SPACE_PA::IncStateSpaceReg::set_excluded_label(const AbstractState& src, ActionLabel_type action_label) {
    PLAJA_ASSERT(stateRegistryPa->has_pa_state(src.get_size_predicates(), src.get_id_value()))
    get_node_reg(src.get_size_predicates()).set_excluded_label(src.get_id_value(), action_label);

    STMT_IF_DEBUG(if (incStateSpaceRef) { incStateSpaceRef->set_excluded_label(src, action_label); })
}

void SEARCH_SPACE_PA::IncStateSpaceReg::set_excluded_op(const AbstractState& src, ActionOpID_type action_op) {
    PLAJA_ASSERT(stateRegistryPa->has_pa_state(src.get_size_predicates(), src.get_id_value()))
    get_node_reg(src.get_size_predicates()).set_excluded_op(src.get_id_value(), action_op);

    STMT_IF_DEBUG(if (incStateSpaceRef) { incStateSpaceRef->set_excluded_op(src, action_op); })
}

void SEARCH_SPACE_PA::IncStateSpaceReg::store_entailments(const AbstractState& src, UpdateOpID_type update_op_id, std::vector<std::pair<PredicateIndex_type, bool>>&& entailments_) {
    STMT_IF_DEBUG(if (incStateSpaceRef) { incStateSpaceRef->store_entailments(src, update_op_id, std::vector<std::pair<PredicateIndex_type, bool>>(entailments_)); })

    PLAJA_ASSERT(stateRegistryPa->has_pa_state(src.get_size_predicates(), src.get_id_value()))
    get_node_reg(src.get_size_predicates()).add_update_node(src.get_id_value(), update_op_id).store_entailments(std::move(entailments_));
}

void SEARCH_SPACE_PA::IncStateSpaceReg::add_transition_witness(const AbstractState& src, UpdateOpID_type update_op_id, StateID_type src_witness, StateID_type suc_witness) {
    PLAJA_ASSERT(stateRegistryPa->has_pa_state(src.get_size_predicates(), src.get_id_value()))
    get_node_reg(src.get_size_predicates()).add_transition_witness(src.get_id_value(), update_op_id, src_witness, suc_witness);

    STMT_IF_DEBUG(if (incStateSpaceRef) { incStateSpaceRef->add_transition_witness(src, update_op_id, src_witness, suc_witness); })
}

void SEARCH_SPACE_PA::IncStateSpaceReg::set_excluded(const AbstractState& src, UpdateOpID_type update_op_id, const AbstractState& excluded_successor) {
    PLAJA_ASSERT(stateRegistryPa->has_pa_state(src.get_size_predicates(), src.get_id_value()))
    PLAJA_ASSERT(stateRegistryPa->has_pa_state(excluded_successor.get_size_predicates(), excluded_successor.get_id_value()))
    get_node_reg(src.get_size_predicates()).set_excluded(src.get_id_value(), update_op_id, excluded_successor.get_id_value());

    STMT_IF_DEBUG(if (incStateSpaceRef) { incStateSpaceRef->set_excluded(src, update_op_id, excluded_successor); })
}

void SEARCH_SPACE_PA::IncStateSpaceReg::refine_for(const AbstractState& source) {

    const auto num_preds = source.get_size_predicates();
    auto& current_node_reg = get_node_reg(num_preds);
    auto& source_node = current_node_reg.add_source(source.get_id_value());

    for (auto it_ss = registryIterator(num_preds - 1); !it_ss.end(); ++it_ss) {
        PLAJA_ASSERT(it_ss.number_of_preds() < num_preds)
        auto& coarser_ss = it_ss.node_reg();
        const StateID_type coarser_id = stateRegistryPa->is_contained_in(source, it_ss.number_of_preds());
        if (coarser_id == STATE::none) { continue; }
        //
        auto* coarser_source_node = coarser_ss.get_source(coarser_id);
        if (not coarser_source_node) { continue; }
        for (auto& [update_op_id, update_op_node]: coarser_source_node->_transitions_per_op()) {
            auto& transition_witnesses = update_op_node._transition_witnesses();
            for (auto it = transition_witnesses.begin(); it != transition_witnesses.end();) {
                const State concrete_source = stateRegistryPa->lookup_witness(it->_source());
                if (source.is_same_predicate_state_sub(concrete_source, it_ss.number_of_preds())) {
                    PLAJA_ASSERT(source.is_abstraction(concrete_source))
                    source_node.add_transition_witness(update_op_id, it->_source(), it->_successor());
                    it = transition_witnesses.erase(it);
                } else { ++it; }
            }
        }
    }

    STMT_IF_DEBUG(if (incStateSpaceRef) { incStateSpaceRef->refine_for(source); })
}

bool SEARCH_SPACE_PA::IncStateSpaceReg::is_excluded_label(const AbstractState& source, ActionLabel_type action_label) const {
    for (auto& [num_preds, node_reg]: nodeRegistries) {
        if (num_preds >= source.get_size_predicates()) { break; } // considered all sub states
        const StateID_type coarser_id = stateRegistryPa->is_contained_in(source, num_preds);
        if (coarser_id == STATE::none) { continue; } // does not contain any information concerning source
        if (node_reg->is_excluded_label(coarser_id, action_label)) {
            STMT_IF_DEBUG(check_is_excluded_label(true, source, action_label))
            return true;
        }
    }

    STMT_IF_DEBUG(check_is_excluded_label(false, source, action_label))
    return false;
}

bool SEARCH_SPACE_PA::IncStateSpaceReg::is_excluded_op(const AbstractState& source, ActionOpID_type action_op) const {
    for (auto& [num_preds, node_reg]: nodeRegistries) {
        if (num_preds >= source.get_size_predicates()) { break; } // considered all sub states
        const StateID_type coarser_id = stateRegistryPa->is_contained_in(source, num_preds);
        if (coarser_id == STATE::none) { continue; } // does not contain any information concerning source
        if (node_reg->is_excluded_op(coarser_id, action_op)) {
            STMT_IF_DEBUG(check_is_excluded_op(true, source, action_op))
            return true;
        }
    }

    STMT_IF_DEBUG(check_is_excluded_op(false, source, action_op))
    return false;
}

PredicatesNumber_type SEARCH_SPACE_PA::IncStateSpaceReg::load_entailments(PredicateState& target, const AbstractState& source, UpdateOpID_type update_op, const PredicateRelations* predicate_relations) const {
    PLAJA_ASSERT(stateRegistryPa->has_pa_state(source.get_size_predicates(), source.get_id_value()))

    PredicatesNumber_type num_preds_checked = 0;

    for (auto it = registryIterator(source.get_size_predicates() - 1); !it.end(); ++it) {
        auto update_op_node = it.get_update_node(source, update_op);
        if (update_op_node) {
            PLAJA_ASSERT(num_preds_checked < it.number_of_preds())
            num_preds_checked = it.number_of_preds();
            for (const auto& [pred, value]: update_op_node->_entailments()) {
                target.set_entailed(pred, value);
                if (predicate_relations) { predicate_relations->fix_predicate_state(target, pred); }
            }
        }
    }

    return num_preds_checked;

}

inline std::unique_ptr<StateID_type> SEARCH_SPACE_PA::IncStateSpaceReg::transition_exists_internal(const AbstractState& source, UpdateOpID_type update_op_id, const AbstractState& successor) const {

    for (auto it = registryIterator(source.get_size_predicates() - 1); !it.end(); ++it) {
        const auto* update_op_node = it.get_update_node(source, update_op_id);
        if (not update_op_node) { continue; }
        // successor exclusion?
        const StateID_type successor_in_registry = stateRegistryPa->is_contained_in(successor, it.number_of_preds());
        if (successor_in_registry != STATE::none and update_op_node->is_excluded(successor_in_registry)) { return std::make_unique<StateID_type>(STATE::none); }
        //
        PLAJA_ASSERT(not std::any_of(update_op_node->_transition_witnesses().cbegin(), update_op_node->_transition_witnesses().cend(), [this, &source](const SEARCH_SPACE_PA::TransitionWitness& witness) { return source.is_abstraction(stateRegistryPa->lookup_witness(witness._source())); }))
# if 0
        for (const auto& transition_witness: update_op_node->_transition_witnesses()) {
                const auto concrete_source = stateRegistryPa->lookup_witness(transition_witness._source());
                if (not source.is_abstraction(concrete_source)) { continue; }
                const auto concrete_successor = stateRegistryPa->lookup_witness(transition_witness._successor());
                if (successor.is_abstraction(concrete_successor)) { return std::make_unique<StateID_type>(transition_witness._source()); }
            }
#endif
    }

    {
        const auto& current_node_reg = get_node_reg(source.get_size_predicates());
        const auto* update_op_node = current_node_reg.get_update_node(source.get_id_value(), update_op_id);
        if (not update_op_node) { return nullptr; }
        if (update_op_node->is_excluded(successor.get_id_value())) { return std::make_unique<StateID_type>(STATE::none); }
        for (const auto& transition_witness: update_op_node->_transition_witnesses()) { // works only if abstract state is up-to-date
            PLAJA_ASSERT(source.is_abstraction(stateRegistryPa->lookup_witness(transition_witness._source())))
            const auto concrete_successor = stateRegistryPa->lookup_witness(transition_witness._successor());
            if (successor.is_abstraction(concrete_successor)) { return std::make_unique<StateID_type>(transition_witness._source()); }
        }
    }

    return nullptr;

}

std::unique_ptr<StateID_type> SEARCH_SPACE_PA::IncStateSpaceReg::transition_exists(const AbstractState& source, UpdateOpID_type update_op_id, const AbstractState& successor) const {
    auto rlt = transition_exists_internal(source, update_op_id, successor);
    STMT_IF_DEBUG(check_transition_exists(rlt, source, update_op_id, successor))
    return rlt;
}
