//
// This file is part of the PlaJA code base (2019 - 2023).
//

#ifndef PLAJA_INC_STATE_SPACE_REG_H
#define PLAJA_INC_STATE_SPACE_REG_H

#include <map>
#include "../pa_states/state_registry_pa.h"
#include "incremental_state_space.h"
#include "source_node.h"

namespace SEARCH_SPACE_PA {

    struct SourceNodeRegistry final {
    private:
        segmented_vector::SegmentedVector<SourceNode<StateID_type>> sources;

    public:
        SourceNodeRegistry() = default;
        ~SourceNodeRegistry() = default;
        DELETE_CONSTRUCTOR(SourceNodeRegistry)

        [[nodiscard]] inline const SourceNode<StateID_type>* get_source(StateID_type src_id) const { if (src_id < sources.size()) { return &sources[src_id]; } else { return nullptr; } }

        [[nodiscard]] inline SourceNode<StateID_type>* get_source(StateID_type src_id) { if (src_id < sources.size()) { return &sources[src_id]; } else { return nullptr; } }

        [[nodiscard]] inline SourceNode<StateID_type>& add_source(StateID_type src_id) {
            if (src_id >= sources.size()) { sources.resize(src_id + 1); }
            return sources[src_id];
        }

        /* label */

        inline void set_excluded_label(StateID_type src_id, ActionLabel_type action_label) { add_source(src_id).set_excluded_label(action_label); }

        [[nodiscard]] inline bool is_excluded_label(StateID_type src_id, ActionLabel_type action_label) const {
            auto* src_node = get_source(src_id);
            return src_node ? src_node->is_excluded_label(action_label) : false;
        }

        /* op */

        inline void set_excluded_op(StateID_type src_id, ActionOpID_type action_op) { add_source(src_id).set_excluded_op(action_op); }

        [[nodiscard]] inline bool is_excluded_op(StateID_type src_id, ActionOpID_type action_op) const {
            auto* src_node = get_source(src_id);
            return src_node ? src_node->is_excluded_op(action_op) : false;
        }

        /* update node / transition */

        [[nodiscard]] inline const UpdateOpNode<StateID_type>* get_update_node(StateID_type src_id, UpdateOpID_type update_op_id) const {
            const auto* src_node = get_source(src_id);
            return src_node ? src_node->get_update_op(update_op_id) : nullptr;
        }

        [[nodiscard]] inline UpdateOpNode<StateID_type>* get_update_node(StateID_type src_id, UpdateOpID_type update_op_id) {
            auto* src_node = get_source(src_id);
            return src_node ? src_node->get_update_op(update_op_id) : nullptr;
        }

        inline UpdateOpNode<StateID_type>& add_update_node(StateID_type src_id, UpdateOpID_type update_op_id) { return add_source(src_id).add_update_op(update_op_id); }

        inline void add_transition_witness(StateID_type src_id, UpdateOpID_type update_op_id, StateID_type src_witness, StateID_type suc_witness) { add_source(src_id).add_transition_witness(update_op_id, src_witness, suc_witness); }

        inline void set_excluded(StateID_type src_id, UpdateOpID_type update_op_id, StateID_type excluded_successor) { add_source(src_id).set_excluded(update_op_id, excluded_successor); }
    };

/**********************************************************************************************************************/

    struct IncStateSpaceReg final: public IncStateSpace {
    private:
        std::shared_ptr<StateRegistryPa> stateRegistryPa;
        std::map<PredicatesNumber_type, std::unique_ptr<SourceNodeRegistry>> nodeRegistries;

        [[nodiscard]] inline SourceNodeRegistry& get_node_reg(PredicatesNumber_type num_preds) {
            PLAJA_ASSERT(nodeRegistries.count(num_preds))
            return *nodeRegistries.at(num_preds);
        };

        [[nodiscard]] inline const SourceNodeRegistry& get_node_reg(PredicatesNumber_type num_preds) const {
            PLAJA_ASSERT(nodeRegistries.count(num_preds))
            return *nodeRegistries.at(num_preds);
        };

        [[nodiscard]] inline StateRegistry& get_state_reg(PredicatesNumber_type num_preds) { return stateRegistryPa->get_registry(num_preds); };

        [[nodiscard]] inline const StateRegistry& get_state_reg(PredicatesNumber_type num_preds) const { return stateRegistryPa->get_registry(num_preds); };

        void set_registry(PredicatesNumber_type num_preds);

        [[nodiscard]] inline std::size_t number_of_registries() const { return nodeRegistries.size(); };

        [[nodiscard]] std::unique_ptr<StateID_type> transition_exists_internal(const AbstractState& source, UpdateOpID_type update_op_id, const AbstractState& successor) const;

    public:
        explicit IncStateSpaceReg(const PLAJA::Configuration& config, std::shared_ptr<StateRegistryPa> state_reg_pa);
        ~IncStateSpaceReg() final;
        DELETE_CONSTRUCTOR(IncStateSpaceReg)

        void update_stats() const override;

        void increment() override;

        void clear_coarser() override;

        /**************************************************************************************************************/

        class RegistryIteratorBase {
        protected:
            IncStateSpaceReg& stateSpace;
            std::map<PredicatesNumber_type, std::unique_ptr<SourceNodeRegistry>>::const_iterator it;
            std::map<PredicatesNumber_type, std::unique_ptr<SourceNodeRegistry>>::const_iterator endIt;
            PredicatesNumber_type maxIndex;

            RegistryIteratorBase(IncStateSpaceReg& state_space, PredicatesNumber_type max_num_preds):
                stateSpace(state_space)
                , it(state_space.nodeRegistries.cbegin())
                , endIt(state_space.nodeRegistries.cend())
                , maxIndex(max_num_preds) {}

        public:
            virtual ~RegistryIteratorBase() = default;
            DELETE_CONSTRUCTOR(RegistryIteratorBase)

            inline void operator++() { ++it; }

            [[nodiscard]] inline bool end() const { return it == endIt or it->first > maxIndex; }

            [[nodiscard]] inline const SourceNodeRegistry& node_reg() const { return *it->second; }

            [[nodiscard]] inline PredicatesNumber_type number_of_preds() const { return it->first; }

            [[nodiscard]] inline const UpdateOpNode<StateID_type>* get_update_node(const AbstractState& source, UpdateOpID_type update_op) const {
                auto coarser_id = stateSpace.stateRegistryPa->is_contained_in(source, number_of_preds());
                if (coarser_id == STATE::none) { return nullptr; } // does not contain any information concerning source
                return node_reg().get_update_node(coarser_id, update_op);
            }

        };

        class RegistryIterator final: public RegistryIteratorBase {
            friend IncStateSpaceReg;
        private:
            RegistryIterator(IncStateSpaceReg& state_space, PredicatesNumber_type max_num_preds):
                RegistryIteratorBase(state_space, max_num_preds) {}

        public:
            ~RegistryIterator() final = default;
            DELETE_CONSTRUCTOR(RegistryIterator)

            [[nodiscard]] inline SourceNodeRegistry& node_reg() { return *it->second; }

            [[nodiscard]] inline UpdateOpNode<StateID_type>* get_update_node(const AbstractState& source, UpdateOpID_type update_op) {
                auto coarser_id = stateSpace.stateRegistryPa->is_contained_in(source, number_of_preds());
                if (coarser_id == STATE::none) { return nullptr; } // does not contain any information concerning source
                return node_reg().get_update_node(coarser_id, update_op);
            }
        };

        class RegistryConstIterator final: public RegistryIteratorBase {
            friend IncStateSpaceReg;
        private:
            RegistryConstIterator(const IncStateSpaceReg& state_space, PredicatesNumber_type max_num_preds):
                RegistryIteratorBase(const_cast<IncStateSpaceReg&>(state_space), max_num_preds) {}

        public:
            ~RegistryConstIterator() final = default;
            DELETE_CONSTRUCTOR(RegistryConstIterator)
        };

        [[nodiscard]] inline RegistryConstIterator registryIterator() const { return { *this, (--nodeRegistries.end())->first }; }

        [[nodiscard]] inline RegistryConstIterator registryIterator(PredicatesNumber_type max_num_preds) const { return { *this, max_num_preds }; }

        [[nodiscard]] inline RegistryIterator registryIterator() { return { *this, (--nodeRegistries.end())->first }; }

        [[nodiscard]] inline RegistryIterator registryIterator(PredicatesNumber_type max_num_preds) { return { *this, max_num_preds }; }

        /**************************************************************************************************************/

        void set_excluded_label(const AbstractState& src, ActionLabel_type action_label) override;

        void set_excluded_op(const AbstractState& src, ActionOpID_type action_op) override;

        void store_entailments(const AbstractState& src, UpdateOpID_type update_op_id, std::vector<std::pair<PredicateIndex_type, bool>>&& entailments_) override;

        void add_transition_witness(const AbstractState& src, UpdateOpID_type update_op_id, StateID_type src_witness, StateID_type suc_witness) override;

        void set_excluded(const AbstractState& src, UpdateOpID_type update_op_id, const AbstractState& excluded_successor) override;

        void refine_for(const AbstractState& source) override;

        [[nodiscard]] bool is_excluded_label(const AbstractState& source, ActionLabel_type action_label) const override;

        [[nodiscard]] bool is_excluded_op(const AbstractState& source, ActionOpID_type action_op) const override;

        [[nodiscard]] PredicatesNumber_type load_entailments(PredicateState& target, const AbstractState& source, UpdateOpID_type update_op, const PredicateRelations* predicate_relations) const override;

        [[nodiscard]] std::unique_ptr<StateID_type> transition_exists(const AbstractState& source, UpdateOpID_type update_op_id, const AbstractState& successor) const override;

        /**************************************************************************************************************/

        [[nodiscard]] inline PredicateTraverser init_predicate_traverser() const final {
#ifdef LAZY_PA
            PLAJA_ABORT
#else
            return PredicateTraverser::init(stateRegistryPa->get_predicates_size());
#endif
        }

    };

}

#endif //PLAJA_INC_STATE_SPACE_REG_H
