//
// This file is part of the PlaJA code base (2019 - 2023).
//

#ifndef PLAJA_SOURCE_NODE_H
#define PLAJA_SOURCE_NODE_H

#include <list>
#include <unordered_map>
#include <unordered_set>
#include <vector>
#include "../../../utils/default_constructors.h"
#include "../../using_search.h"
#include "../using_predicate_abstraction.h"

namespace SEARCH_SPACE_PA {

    struct TransitionWitness final {
    private:
        StateID_type source;
        StateID_type successor;
    public:
        TransitionWitness(StateID_type src, StateID_type suc):
            source(src)
            , successor(suc) {}

        ~TransitionWitness() = default;
        DEFAULT_CONSTRUCTOR_ONLY(TransitionWitness)
        DELETE_ASSIGNMENT(TransitionWitness)

        [[nodiscard]] inline StateID_type _source() const { return source; }

        [[nodiscard]] inline StateID_type _successor() const { return successor; }
    };

    template<typename ExclusionInfo_t>
    struct UpdateOpNode final {
    private:
        std::vector<std::pair<PredicateIndex_type, bool>> entailments; // truth values entailed for source state and update (should still use static entailments and binary relations)
        std::list<TransitionWitness> transitionWitnesses;
        std::unordered_set<ExclusionInfo_t> excludedSuccessors;
    public:
        inline void store_entailments(std::vector<std::pair<PredicateIndex_type, bool>>&& entailments_) { entailments = std::move(entailments_); }

        inline void add_transition_witness(StateID_type src_witness, StateID_type suc_witness) { transitionWitnesses.emplace_back(src_witness, suc_witness); }

        inline void set_excluded(ExclusionInfo_t excluded_successor) { excludedSuccessors.insert(excluded_successor); }

        //
        [[nodiscard]] inline const std::vector<std::pair<PredicateIndex_type, bool>>& _entailments() const { return entailments; }

        //
        [[nodiscard]] inline const std::list<TransitionWitness>& _transition_witnesses() const { return transitionWitnesses; }

        [[nodiscard]] inline std::list<TransitionWitness>& _transition_witnesses() { return transitionWitnesses; }

        [[nodiscard]] inline bool is_excluded(ExclusionInfo_t successor) const { return excludedSuccessors.count(successor); }

        [[nodiscard]] inline const std::unordered_set<ExclusionInfo_t>& _excluded_successors() const { return excludedSuccessors; }
    };

    template<typename ExclusionInfo_t>
    struct SourceNode final {
    private:
        std::unordered_set<ActionLabel_type> excludedLabels;
        std::unordered_set<ActionOpID_type> excludedOps;
        std::unordered_map<UpdateOpID_type, UpdateOpNode<ExclusionInfo_t>> transitionsPerOp;
    public:
        SourceNode() = default;
        ~SourceNode() = default;
        DEFAULT_CONSTRUCTOR_ONLY(SourceNode)
        DELETE_ASSIGNMENT(SourceNode)

        inline void set_excluded_label(ActionLabel_type action_label) { excludedLabels.insert(action_label); }

        [[nodiscard]] inline bool is_excluded_label(ActionLabel_type action_label) const { return excludedLabels.count(action_label); }

        inline void set_excluded_op(ActionOpID_type action_op) { excludedOps.insert(action_op); }

        [[nodiscard]] inline bool is_excluded_op(ActionOpID_type action_op) const { return excludedOps.count(action_op); }

        [[nodiscard]] inline UpdateOpNode<ExclusionInfo_t>* get_update_op(UpdateOpID_type update_op_id) {
            auto rlt = transitionsPerOp.find(update_op_id);
            return rlt == transitionsPerOp.end() ? nullptr : &rlt->second;
        }

        [[nodiscard]] inline const UpdateOpNode<ExclusionInfo_t>* get_update_op(UpdateOpID_type update_op_id) const {
            auto rlt = transitionsPerOp.find(update_op_id);
            return rlt == transitionsPerOp.end() ? nullptr : &rlt->second;
        }

        [[nodiscard]] inline UpdateOpNode<ExclusionInfo_t>& add_update_op(UpdateOpID_type update_op_id) { return transitionsPerOp.emplace(update_op_id, UpdateOpNode<ExclusionInfo_t>()).first->second; }

        [[nodiscard]] inline std::unordered_map<UpdateOpID_type, UpdateOpNode<ExclusionInfo_t>>& _transitions_per_op() { return transitionsPerOp; }

        inline void add_transition_witness(UpdateOpID_type update_op_id, StateID_type src_witness, StateID_type suc_witness) { add_update_op(update_op_id).add_transition_witness(src_witness, suc_witness); }

        inline void set_excluded(UpdateOpID_type update_op_id, ExclusionInfo_t excluded_successor) { add_update_op(update_op_id).set_excluded(excluded_successor); }
    };

}

#endif //PLAJA_SOURCE_NODE_H
