//
// This file is part of the PlaJA code base (2019 - 2020).
//

#ifndef PLAJA_NN_SAT_CHECKER_H
#define PLAJA_NN_SAT_CHECKER_H

#include <list>
#include <vector>
#include <memory>
#include "../../../../include/marabou_include/milp_solver_bound_tightening_type.h"
#include "../../../../include/ct_config_const.h"
#include "../../../../stats/forward_stats.h"
#include "../../../../assertions.h"
#include "../../../factories/forward_factories.h"
#include "../../../information/jani2nnet/forward_jani_2_nnet.h"
#include "../../../information/jani2nnet/using_jani2nnet.h"
#include "../../../using_search.h"
#include "../../../smt/base/forward_solution_check.h"
#include "../../../smt_nn/forward_smt_nn.h"
#include "../../../states/forward_states.h"
#include "../../pa_states/forward_pa_states.h"
#include "../smt/forward_smt_nn_pa.h"

class AdversarialAttack;

/** NN-Sat checking for PPA. */
class NNSatChecker {
    friend struct NNSatCheckerFactory; //
    friend class NNSatCheckerTest;

protected:
    // construction data
    std::shared_ptr<ModelMarabouPA> modelMarabou;
    const Jani2NNet* jani2NNet;
    const SolutionChecker* solutionChecker;

    // checking modes
    std::unique_ptr<MARABOU_IN_PLAJA::SMTSolver> solverMarabou;
    MILPSolverBoundTighteningType perSrcBounds;
    MILPSolverBoundTighteningType perSrcLabelBounds;
#ifdef USE_ADVERSARIAL_ATTACK
    std::unique_ptr<AdversarialAttack> adversarialAttack;
    bool attackSolutionGuess;
#endif

    // local data
    PLAJA::StatsBase* sharedStatistics;

    // cache
    ActionLabel_type set_action_label;
    OutputIndex_type set_output_index;
    ActionOpID_type set_action_op_id;
    UpdateIndex_type set_update_index;
    const ActionOpMarabouPA* set_action_op;
    std::unique_ptr<const AbstractState> set_source_state;
    std::unique_ptr<const AbstractState> set_target_state;

    std::unique_ptr<SolutionCheckerInstance> lastSolution;

#ifdef ENABLE_QUERY_CACHING
    std::unique_ptr<QueryToJson> queryToJson;
#endif

    void save_solution_if(bool rlt);

    virtual bool check_(); // internal auxiliary function to reduce code duplication

public:
    NNSatChecker(const Jani2NNet& jani_2_nnet, const PLAJA::Configuration& config);
    virtual ~NNSatChecker();
    DELETE_CONSTRUCTOR(NNSatChecker)

    inline void set_solution_checker(const SolutionChecker* solution_checker) { solutionChecker = solution_checker; }

    void set_op_applicability_cache(OpApplicability& op_app);

    [[nodiscard]] const Jani2NNet& _jani_2_nnet() const { return *jani2NNet; }

    void increment();

    void push();
    void pop();

    void set_label(ActionLabel_type action_label);
    void unset_label();
    void set_operator(ActionOpID_type action_op_id);
    void unset_operator();
    void set_update(UpdateIndex_type update_index);
    void unset_update();

    /**
     * Add constraints res. predicates of the source state.
     * In particular also location values if in usage.
     * @param source
     */
    void add_source_state(const AbstractState& source);
    void unset_source_state();

#ifdef ENABLE_TERMINAL_STATE_SUPPORT
    void add_non_terminal_condition();
#else
    inline void add_non_terminal_condition() { PLAJA_ABORT }
#endif

    /**
     * @return true if the guard is not trivially true (and thus the query has been modified).
     */
    bool add_guard();

    /**
     * @return true if guard is not trivially true (and thus the query has *not* become trivially unsat).
     */
    bool add_guard_negation();

    void add_update();
    void add_target_state(const AbstractState& target);
    void unset_target_state();

    /**
     * @return false if the query has become unsat.
     */
    bool add_output_interface(); // internally prior to each check

    [[nodiscard]] bool is_action_selectable(ActionLabel_type action_label) const; // short-cut to jani2nnet interface

    [[nodiscard]] inline bool has_solution() const { return lastSolution != nullptr; }

    [[nodiscard]] const StateValues& get_last_solution() const;

    [[nodiscard]] std::unique_ptr<StateValues> release_solution();

    [[nodiscard]] const StateValues* get_target_solution() const;

    [[nodiscard]] std::unique_ptr<StateValues> release_target_solution();

    /**
     * Check whether the current input query is sat (true) or unsat (false).
     * @param unknown, value to return in case input query could not be solved.
     * @return
     */
    bool check();
    bool check_relaxed();

#ifdef ENABLE_STALLING_DETECTION
    void stalling_detection(const State& source);
#endif

#ifdef USE_ADVERSARIAL_ATTACK
    bool adversarial_attack();
#endif

    void print_statistics();

};

namespace NN_SAT_CHECKER { extern std::unique_ptr<NNSatChecker> construct_checker(const Jani2NNet& jani_2_nnet, const PLAJA::Configuration& config); }

#endif //PLAJA_NN_SAT_CHECKER_H
