//
// This file is part of the PlaJA code base (2019 - 2020).
//

#include "predicate_state.h"
#include "../../../parser/ast/expression/expression.h"
#include "../../../utils/utils.h"
#include "../../states/state_values.h"

PredicateState::PredicateState(PredicatesNumber_type num_predicates, std::size_t num_locs, const PredicatesExpression* predicates):
    PaStateBase(predicates, num_predicates, num_locs)
    , locationState(num_locs > 0 ? new LocationState(num_locs) : nullptr)
#ifdef LAZY_PA
    , predicateValues(num_predicates, PA::unknown_value())
#else
    , predicateValues(num_predicates)
#endif
    , predicateStates(num_predicates, PredicateValueState::UNDEFINED)
    CONSTRUCT_IF_DEBUG(unsat(false)) {
}

PredicateState::~PredicateState() = default;

/**********************************************************************************************************************/

void PredicateState::set_location_state(const PredicateState::LocationState& location_state) const {
    PLAJA_ASSERT(has_location_state())
    PLAJA_ASSERT(location_state.size() == locationState->size())

    const auto& num_locs = locationState->size();
    for (VariableIndex_type state_index = 0; state_index < num_locs; ++state_index) {
        locationState->assign_int<true>(state_index, location_state.get_int(state_index));
    }

}

PLAJA::integer PredicateState::location_value(VariableIndex_type loc_index) const {
    PLAJA_ASSERT(has_location_state())
    return locationState->get_int(loc_index);
}

/**********************************************************************************************************************/

bool PredicateState::has_entailment_information() const { return true; }

PA::EntailmentValueType PredicateState::entailment_value(PredicateIndex_type pred_index) const {
    if constexpr (PLAJA_GLOBAL::lazyPA) {
        return is_entailed(pred_index) ? predicateValues[pred_index] : PA::unknown_value();
    } else {
        PLAJA_ASSERT(is_entailed(pred_index))
        return predicateValues[pred_index];
    }
}

/**********************************************************************************************************************/

void PredicateState::set(PredicateIndex_type pred_index, bool value) {
    PLAJA_ASSERT(not is_unsat())
    PLAJA_ASSERT(pred_index < predicateStates.size())

    PredicateValueState& predicate_state = predicateStates[pred_index];

    switch (predicate_state) {
        case UNDEFINED: {
            predicateValues[pred_index] = value;
            predicate_state = PredicateValueState::SET;
            break;
        }
        case ENTAILED: {
            PLAJA_ASSERT(predicateValues[pred_index] == value) // not intended
            predicate_state = PredicateValueState::SET_AND_ENTAILED;
            break;
        }
        default: { PLAJA_ABORT } // not intended
    }

    add_to_layer(pred_index);
}

void PredicateState::set_entailed(PredicateIndex_type pred_index, bool value) {
    PLAJA_ASSERT(not is_unsat())
    PLAJA_ASSERT(pred_index < predicateStates.size())

    PredicateValueState& predicate_state = predicateStates[pred_index];

    switch (predicate_state) {
        case UNDEFINED: {
            predicateValues[pred_index] = value;
            predicate_state = PredicateValueState::ENTAILED;
            break;
        }
        case SET: {
            PLAJA_ASSERT(predicateValues[pred_index] == value) // not intended
            predicate_state = PredicateValueState::SET_AND_ENTAILED;
            PLAJA_ABORT_IF_CONST(PLAJA_GLOBAL::useFineGrainedPredicateStateStack) // not supported when using fine-grained stack
            break;
        }
        case ENTAILED: {
            PLAJA_ASSERT(predicateValues[pred_index] == value) // not intended
            return;
        }
        case SET_AND_ENTAILED: { return; }
    }

    add_to_layer(pred_index);
}

//

bool PredicateState::operator==(const PredicateState& other) const {
    const auto num_preds = get_size_predicates();
    if (num_preds != other.get_size_predicates()) { return false; }

    for (PredicateIndex_type pred_index = 0; pred_index < num_preds; ++pred_index) {
        if (predicateValues[pred_index] != other.predicateValues[pred_index]) { return false; }
        if (predicateStates[pred_index] != other.predicateStates[pred_index]) { return false; }
    }

    return true;
}

bool PredicateState::equals(const PredicateState& other) const {
    if (get_size_predicates() != other.get_size_predicates()) { return false; }

    for (auto it = init_pred_state_iterator(); !it.end(); ++it) {
        if (it.is_defined() != other.is_defined(it.predicate_index())) { return false; }
        if (it.is_defined() and it.predicate_value() != other.predicate_value(it.predicate_index())) { return false; }
    }

    return true;
}

#ifndef NDEBUG

bool PredicateState::matches(const StateBase& state) const {

    if (has_location_state() and not is_same_location_state(state)) { return false; }

    for (auto it = init_pred_state_iterator(); !it.end(); ++it) { // NOLINT(readability-use-anyofallof)

        if (it.is_set() and it.predicate_value() != it.predicate().evaluate_integer(state)) { return false; }

        if (it.is_entailed() and it.entailment_value() != it.predicate().evaluate_integer(state)) { return false; }

    }

    return true;
}

#endif

#if 0

void PredicateState::dump() const {
#ifndef NDEBUG
    std::cout << "[" << unsat << ";";
#else
    std::cout << "[";
#endif
    const auto l = predicateStates.size();
    for (std::size_t i = 0; i < l;) {

        switch (predicateStates[i]) {
            case UNDEFINED: {
                std::cout << "UNDEFINED";
                break;
            }
            case SET: {
                std::cout << "SET" << PLAJA_UTILS::spaceString << predicateValues[i];
                break;
            }
            case ENTAILED: {
                std::cout << "ENTAILED" << PLAJA_UTILS::spaceString << predicateValues[i];
                break;
            }
            case SET_AND_ENTAILED: {
                std::cout << "SET & ENTAILED" << PLAJA_UTILS::spaceString << predicateValues[i];
                break;
            }
            default: { PLAJA_ABORT }
        }

        if (++i < l) {
            std::cout << PLAJA_UTILS::commaString << PLAJA_UTILS::spaceString;
        }
    }
    std::cout << "]" << std::endl;
}

#endif