//
// This file is part of the PlaJA code base (2019 - 2020).
//

#ifndef PLAJA_PREDICATE_STATE_H
#define PLAJA_PREDICATE_STATE_H

#pragma GCC diagnostic push
#pragma ide diagnostic ignored "Simplify"

#include <forward_list>
#include <memory>
#include "../../../utils/utils.h"
#include "pa_state_base.h"

// #define USE_FINE_GRAINED_PREDICATE_STATE_STACK // only cache the predicates modified per stack layer

#ifdef USE_FINE_GRAINED_PREDICATE_STATE_STACK
namespace PLAJA_GLOBAL { constexpr bool useFineGrainedPredicateStateStack = true; }
#else
namespace PLAJA_GLOBAL { constexpr bool useFineGrainedPredicateStateStack = false; }
#endif

class PredicateState final: public PaStateBase {

    using LocationState = StateValues;

private:
    enum PredicateValueState {
        UNDEFINED,
        SET,
        ENTAILED, // predicate (value) is entailed by other predicate or structure of the underlying sat problem
        SET_AND_ENTAILED
    };

    // locations
    std::unique_ptr<LocationState> locationState;

    // predicates
    std::vector<PA::TruthValueType> predicateValues;
    std::vector<PredicateValueState> predicateStates;

    // stacks
#ifdef USE_FINE_GRAINED_PREDICATE_STATE_STACK
    std::forward_list<std::forward_list<PredicateIndex_type>> predicateStatesStack;
#else
    std::forward_list<std::vector<PredicateValueState>> predicateStatesStack;
#endif

    STMT_IF_DEBUG(bool unsat) // indicates whether the predicate state is unsat

public:
    explicit PredicateState(PredicatesNumber_type num_predicates, std::size_t num_locs, const PredicatesExpression* predicates = nullptr);
    ~PredicateState() final;
    PredicateState(const PredicateState& predicate_state) = delete;
    PredicateState(PredicateState&& predicate_state) = default;
    DELETE_ASSIGNMENT(PredicateState)

    /******************************************************************************************************************/

    // locations

    void set_location_state(const LocationState& location_state) const;

    [[nodiscard]] inline bool has_location_state() const { return locationState.get(); }

    [[nodiscard]] inline const LocationState& get_location_state() const {
        PLAJA_ASSERT(has_location_state())
        return *locationState;
    }

    [[nodiscard]] inline LocationState& get_location_state() {
        PLAJA_ASSERT(has_location_state())
        return *locationState;
    }

    [[nodiscard]] PLAJA::integer location_value(VariableIndex_type loc_index) const final;

    /******************************************************************************************************************/

    // predicates

    FCT_IF_DEBUG([[nodiscard]] inline bool is_unsat() const { return unsat; })

    // getter:

    [[nodiscard]] inline bool is_undefined(PredicateIndex_type pred_index) const {
        PLAJA_ASSERT(not is_unsat()) // sanity check for usage
        PLAJA_ASSERT(pred_index < predicateStates.size())
        return predicateStates[pred_index] == PredicateValueState::UNDEFINED;
    }

    [[nodiscard]] inline bool is_defined(PredicateIndex_type pred_index) const { return not is_undefined(pred_index); }

    [[nodiscard]] bool is_set(PredicateIndex_type pred_index) const final {
        PLAJA_ASSERT(not is_unsat()) // sanity check for usage
        PLAJA_ASSERT(pred_index < predicateStates.size())
        const auto& pred_state = predicateStates[pred_index];
        return pred_state == PredicateValueState::SET or pred_state == PredicateValueState::SET_AND_ENTAILED;
    }

    [[nodiscard]] bool has_entailment_information() const final;

    [[nodiscard]] bool is_entailed(PredicateIndex_type pred_index) const final {
        PLAJA_ASSERT(not is_unsat()) // sanity check for usage
        PLAJA_ASSERT(pred_index < predicateStates.size())
        const auto& pred_state = predicateStates[pred_index];
        return pred_state == PredicateValueState::ENTAILED or pred_state == PredicateValueState::SET_AND_ENTAILED;
    }

    [[nodiscard]] PA::EntailmentValueType entailment_value(PredicateIndex_type pred_index) const final;

    // TODO: This returns the entailed value even if not set.
    //  This is not equivalent to the behavior or the AbstractState implementation.
    //  Prone to result in bugs.
    /**
     * Shall only be called if predicate is *defined*.
     * @param pred_index
    * @return
    */
    [[nodiscard]] PA::TruthValueType predicate_value(PredicateIndex_type pred_index) const final {
        PLAJA_ASSERT(is_defined(pred_index) or PA::is_unknown(predicateValues[pred_index]))
        return predicateValues[pred_index];
    }

    /** stack interface ***********************************************************************************************/

    inline void push_layer() {
        if constexpr (PLAJA_GLOBAL::useFineGrainedPredicateStateStack) {
            predicateStatesStack.emplace_front();
        } else {
            predicateStatesStack.push_front(predicateStates);
        }
    }

#ifdef USE_FINE_GRAINED_PREDICATE_STATE_STACK
    inline void add_to_layer(PredicateIndex_type pred_index) {
        if (not predicateStatesStack.empty()) { predicateStatesStack.front().push_front(pred_index); } // add to last layer if present
    }
#else

    inline void add_to_layer(PredicateIndex_type /*pred_index*/) {}

#endif

    inline void reset_layer() {
        PLAJA_ASSERT(not predicateStatesStack.empty())
        if constexpr (PLAJA_GLOBAL::useFineGrainedPredicateStateStack) {
            for (PredicateIndex_type pred_index: predicateStatesStack.front()) { unset(pred_index); }
        } else {
            predicateStates = predicateStatesStack.front();
        }
    }

    inline void pop_layer() {
        PLAJA_ASSERT(not predicateStatesStack.empty())
        if constexpr (PLAJA_GLOBAL::useFineGrainedPredicateStateStack) {
            for (PredicateIndex_type pred_index: predicateStatesStack.front()) { unset(pred_index); }
        } else {
            predicateStates = std::move(predicateStatesStack.front());
        }
        predicateStatesStack.pop_front();
    }

    /** setter ********************************************************************************************************/

    void set(PredicateIndex_type pred_index, bool value);

    void set_entailed(PredicateIndex_type pred_index, bool value);

    /**
     * Shall only be called if predicate is *defined*.
     * @param pred_index
     */
    inline void unset(PredicateIndex_type pred_index) {
        PLAJA_ASSERT(is_defined(pred_index))
        predicateStates[pred_index] = PredicateValueState::UNDEFINED;
    }

    /******************************************************************************************************************/

    [[nodiscard]] bool operator==(const PredicateState& other) const;

    [[nodiscard]] inline bool operator!=(const PredicateState& other) const { return not operator==(other); }

    /**
     * Compares truth values of defined predicates.
     * @param other
     * @return
     */
    [[nodiscard]] bool equals(const PredicateState& other) const;

    /**
     * Concrete state matches predicate state?
     * Unlike is_same_predicate_state intended to be called with "non-closed" states.
     * Reflects entailment values.
     */
    FCT_IF_DEBUG([[nodiscard]] bool matches(const StateBase& state) const;)

    /******************************************************************************************************************/

    // template<bool CONST, typename PredicateState_t>
    class PredicateStateIterator final: public PaStateBase::PredicateStateIterator {
        friend PredicateState;

    private:

        explicit PredicateStateIterator(const PredicateState& predicate_state):
            PaStateBase::PredicateStateIterator(predicate_state, 0) {}

    public:
        ~PredicateStateIterator() override = default;
        DELETE_CONSTRUCTOR(PredicateStateIterator)

        [[nodiscard]] inline const PredicateState& operator()() const { return PLAJA_UTILS::cast_ref<PredicateState>(*paState); }

        [[nodiscard]] inline bool is_undefined() const { return operator()().is_undefined(predicate_index()); }

        [[nodiscard]] inline bool is_defined() const { return operator()().is_defined(predicate_index()); }

        // template<bool is_const = CONST>
        // [[nodiscard]] inline typename std::enable_if<is_const, void>::type set(PredicateIndex_type pred_index, bool value) { predicateState.set(pred_index, value); }

    };

    using PredicateStateIteratorConst = PredicateStateIterator; // <true, const PredicateState>
    [[nodiscard]] inline PredicateStateIteratorConst init_pred_state_iterator() const { return PredicateStateIteratorConst(*this); }

};

#pragma GCC diagnostic pop

#endif //PLAJA_PREDICATE_STATE_H