//
// This file is part of the PlaJA code base (2019 - 2021).
//

#include <memory>
#include "search_pa_path.h"
#include "../factories/predicate_abstraction/search_engine_config_pa.h"
#include "heuristic/state_queue.h"
#include "pa_states/abstract_path.h"
#include "search_space/search_space_pa_base.h"
#include "base/expansion_pa_base_parameterized.h"

SearchPAPath::SearchPAPath(const SearchEngineConfigPA& config):
    SearchPABase(config) {
    expansionPa = std::make_unique<ExpansionPABaseParameterized<SearchPAPath, true, false>>(config, searchSpace.get(), *modelZ3);
}

SearchPAPath::~SearchPAPath() = default;


//

#include "base/initialize_pa_base.h"

SearchEngine::SearchStatus SearchPAPath::initialize() { return InitializePABase::initialize<SearchPAPath>(this); }

SearchEngine::SearchStatus SearchPAPath::step() {
    if (frontier->empty()) {
        if (searchSpace->goal_path_frontier_empty()) { std::cout << "Goal unreachable!" << std::endl; }
        return SearchEngine::SOLVED;
    }

    auto rlt = expansionPa->step(frontier->pop());
    // update frontier
    for (const StateID_type state_id: expansionPa->_newly_reached()) { frontier->push(searchSpace->get_node(state_id)); }
    //
    return rlt;
}

//

HeuristicValue_type SearchPAPath::search_path(StateID_type src_id) {
    frontier->clear();
    pathLength = -1;
    searchStatistics->set_attr_int(PLAJA::StatsInt::PATH_LENGTH, pathLength);
    searchSpace->reset();

    auto rlt = SearchEngine::SearchStatus::IN_PROGRESS;
    auto node = searchSpace->get_node(src_id);
    node.set_reached_init();
    frontier->push(node);

    while (SearchEngine::SearchStatus::IN_PROGRESS == rlt) {
        // PLAJA_ASSERT(searchSpace->goal_path_frontier_empty())
        if (frontier->empty()) { break; }
        rlt = expansionPa->step(frontier->pop());
        // update frontier
        for (const StateID_type state_id: expansionPa->_newly_reached()) { frontier->push(searchSpace->get_node(state_id)); }
    }

    // Set
    if (not searchSpace->goal_path_frontier_empty()) {
        searchSpace->compute_goal_path_states();
        return searchSpace->get_node(src_id)._goal_distance();
    } else { // Set dead ends.
        /* while (not frontier->empty()) { searchSpace->compute_dead_ends_backwards(frontier->pop()); } */
        PLAJA_ABORT; // Do not expect that to happen on the considered benchmarks.
        return HEURISTIC::deadEnd;
    }
}

HeuristicValue_type SearchPAPath::get_goal_distance(const AbstractState& pa_state) {
    auto node = searchSpace->add_node(searchSpace->set_abstract_state(pa_state));
    if (node.known_goal_distance()) {
        statsHeuristic->inc_attr_unsigned(PLAJA::StatsUnsigned::HEURISTIC_CACHE_HITS);
        return node._goal_distance();
    } else { return search_path(node._id()); }
}