//
// This file is part of the PlaJA code base (2019 - 2022).
//

#ifndef PLAJA_SEARCH_NODE_PA_H
#define PLAJA_SEARCH_NODE_PA_H

#include <algorithm>
#include "../../../utils/utils.h"
#include "../../using_search.h"
#include "../pa_states/abstract_state.h"

namespace SEARCH_SPACE_PA {

    struct SearchNodeInfo final {
        friend struct SearchNode;
        friend struct SearchNodeView;

    private:
        std::vector<std::pair<StateID_type, UpdateOpID_type>> parents;
        StateID_type goalPathSuccessor;
        SearchDepth_type startDistance;
        HeuristicValue_type goalDistance;
        unsigned int currentStep;

        inline void set_current_step(unsigned int current_step) { currentStep = current_step; }

        [[nodiscard]] inline unsigned int _current_step() const { return currentStep; }

    public:
        SearchNodeInfo():
            parents()
            , goalPathSuccessor(STATE::none)
            , startDistance(static_cast<SearchDepth_type>(-1))
            , goalDistance(HEURISTIC::unknown)
            , currentStep(0) {}

        ~SearchNodeInfo() = default;

        DEFAULT_CONSTRUCTOR_ONLY(SearchNodeInfo)
        DELETE_ASSIGNMENT(SearchNodeInfo)

        // inline void set_parent(StateID_type parent_id, UpdateOpID_type parent_update_op) { parentID = parent_id; parentUpdateOp = parent_update_op; }
        inline void add_parent(StateID_type parent_id, UpdateOpID_type parent_update_op) {
            PLAJA_ASSERT(not has_goal_path() or std::all_of(parents.cbegin(), parents.cend(), [parent_id](const std::pair<StateID_type, UpdateOpID_type> known_parent) { return parent_id != known_parent.first; }))
            if (std::any_of(parents.cbegin(), parents.cend(), [parent_id, parent_update_op](const std::pair<StateID_type, UpdateOpID_type> known_parent) { return parent_id == known_parent.first and parent_update_op == known_parent.second; })) {
                return;
            }
            parents.emplace_back(parent_id, parent_update_op);
        }

        [[nodiscard]] inline StateID_type _parent_id() const {
            PLAJA_ASSERT(not parents.empty())
            return parents.front().first;
        }

        [[nodiscard]] inline UpdateOpID_type _parent_update_op() const {
            PLAJA_ASSERT(not parents.empty())
            return parents.front().second;
        }

        // [[nodiscard]] inline StateID_type has_parent() const { PLAJA_ASSERT((parentID != STATE::none and parentUpdateOp != ACTION::noneUpdateOp) or (parentID == STATE::none and parentUpdateOp == ACTION::noneUpdateOp)) return parentID != STATE::none; }
        [[nodiscard]] inline bool has_parent() const { return not parents.empty(); }

        inline void set_start_distance(SearchDepth_type start_distance) { startDistance = start_distance; }

        inline void set_goal_distance(HeuristicValue_type goal_distance) {
            PLAJA_ASSERT(goalDistance == HEURISTIC::unknown or goal_distance < goalDistance)
            goalDistance = goal_distance;
        }

        inline void set_goal_path_successor(StateID_type goal_path_successor, HeuristicValue_type goal_distance) {
            PLAJA_ASSERT(goal_path_successor != STATE::none)
            goalPathSuccessor = goal_path_successor;
            set_goal_distance(goal_distance);
        }

        inline void set_goal() {
            PLAJA_ASSERT(not has_goal_path())
            PLAJA_ASSERT(goalDistance == HEURISTIC::unknown)
            goalPathSuccessor = STATE::none;
            goalDistance = 0;
        }

        inline void set_dead_end() {
            PLAJA_ASSERT(not has_goal_path())
            PLAJA_ASSERT(goalDistance == HEURISTIC::unknown)
            goalPathSuccessor = STATE::none;
            goalDistance = HEURISTIC::deadEnd;
        }

        [[nodiscard]] inline SearchDepth_type _start_distance() const { return startDistance; }

        [[nodiscard]] inline HeuristicValue_type _goal_distance() const { return goalDistance; }

        [[nodiscard]] inline StateID_type _goal_path_successor() const { return goalPathSuccessor; }

        [[nodiscard]] inline bool known_goal_distance() const { return goalDistance != HEURISTIC::unknown; }

        [[nodiscard]] inline bool is_goal() const { return goalDistance == 0; }

        [[nodiscard]] inline bool is_dead_end() const { return goalDistance == HEURISTIC::deadEnd; }

        [[nodiscard]] inline bool has_goal_path() const {
            PLAJA_ASSERT(goalDistance == HEURISTIC::unknown or goalDistance == HEURISTIC::deadEnd or goalDistance == 0 or goalPathSuccessor != STATE::none)
            return goalDistance != HEURISTIC::unknown and goalDistance != HEURISTIC::deadEnd;
        }

        class ParentsIterator {
            friend SEARCH_SPACE_PA::SearchNodeInfo;

        private:
            std::vector<std::pair<StateID_type, UpdateOpID_type>>::const_iterator it;
            std::vector<std::pair<StateID_type, UpdateOpID_type>>::const_iterator itEnd;

            explicit ParentsIterator(const std::vector<std::pair<StateID_type, UpdateOpID_type>>& parents):
                it(parents.cbegin())
                , itEnd(parents.cend()) {}

        public:
            ~ParentsIterator() = default;

            DELETE_CONSTRUCTOR(ParentsIterator)

            inline void operator++() { ++it; }

            [[nodiscard]] inline bool end() const { return it == itEnd; }

            [[nodiscard]] inline StateID_type parent_state() const { return it->first; }

            [[nodiscard]] inline UpdateOpID_type parent_update_op() const { return it->second; }

        };

        [[nodiscard]] inline SEARCH_SPACE_PA::SearchNodeInfo::ParentsIterator _parents() const { return ParentsIterator(parents); }

#ifndef NDEBUG

        void dump() const {
            for (auto& [p_state, p_op]: parents) { std::cout << "Parent: " << p_state << " " << p_op << std::endl; }
            std::cout << "Goal path successor: " << goalPathSuccessor << std::endl;
            std::cout << "Start distance: " << startDistance << std::endl;
            std::cout << "Goal distance: " << goalDistance << std::endl;
            std::cout << "Current step: " << currentStep << std::endl;
        }

#endif

    };

/**********************************************************************************************************************/

    struct SearchNodeView {

    protected:
        std::unique_ptr<AbstractState> paState;
        unsigned currentStep;
        const SearchNodeInfo* info; // ptr to enable assignment

    public:

        [[nodiscard]] inline bool has_been_reached() const {
            PLAJA_ASSERT(info->_current_step() <= currentStep)
            return info->_current_step() > 0;
        }

        [[nodiscard]] inline bool is_reached() const {
            PLAJA_ASSERT(info->_current_step() <= currentStep)
            return info->_current_step() >= currentStep;
        }

        SearchNodeView(std::unique_ptr<AbstractState> pa_state, unsigned current_step, const SearchNodeInfo& info_):
            paState(std::move(pa_state))
            , currentStep(current_step)
            , info(&info_) {
            PLAJA_ASSERT(currentStep >= 1)
        }

        SearchNodeView(const SearchNodeView& other) = delete;
        SearchNodeView(SearchNodeView&& other) = default;

        SearchNodeView& operator=(const SearchNodeView& other) {
            if (this == &other) { return *this; }
            paState = other.paState->to_ptr();
            currentStep = other.currentStep;
            info = other.info;
            return *this;
        }

        SearchNodeView& operator=(SearchNodeView&& other) = default;

        virtual ~SearchNodeView() = default; // base class for modifiable node, though node view is stored on stack -> destruction can be made non-virtual

        [[nodiscard]] inline const AbstractState& _state() const { return *paState; }

        [[nodiscard]] inline StateID_type _id() const { return paState->get_id_value(); }

        [[nodiscard]] inline bool is_start() const {
            PLAJA_ASSERT(is_reached())
            return info->_start_distance() == 0;
        }

        [[nodiscard]] inline SearchDepth_type _start_distance() const {
            PLAJA_ASSERT(has_been_reached())
            return info->_start_distance();
        }

        [[nodiscard]] inline StateID_type _parent_id() const {
            PLAJA_ASSERT(has_been_reached())
            return info->_parent_id();
        }

        [[nodiscard]] inline UpdateOpID_type _parent_update_op() const {
            PLAJA_ASSERT(has_been_reached())
            return info->_parent_update_op();
        }

        [[nodiscard]] inline SearchNodeInfo::ParentsIterator _parents() const {
            PLAJA_ASSERT(has_been_reached())
            return info->_parents();
        }

        [[nodiscard]] inline HeuristicValue_type _goal_distance() const {
            PLAJA_ASSERT(has_been_reached())
            return info->_goal_distance();
        }

        [[nodiscard]] inline StateID_type _goal_path_successor() const {
            PLAJA_ASSERT(has_been_reached())
            return info->_goal_path_successor();
        }

        [[nodiscard]] inline bool is_goal() const {
            PLAJA_ASSERT(is_reached())
            return info->is_goal();
        }

        [[nodiscard]] inline bool has_goal_path() const {
            PLAJA_ASSERT(has_been_reached())
            return info->has_goal_path();
        }

        [[nodiscard]] inline bool known_goal_distance() const { return info->known_goal_distance(); }

        [[nodiscard]] inline bool is_dead_end() const { return info->is_dead_end(); }

        STMT_IF_DEBUG(void dump() const {
            PLAJA_LOG_DEBUG(PLAJA_UTILS::string_f("Search Node: %i (current step global(%i))", _id(), currentStep))
            info->dump();
        })
    };

/**********************************************************************************************************************/

    struct SearchNode final: public SearchNodeView {

    private:

        inline SearchNodeInfo& _info() { return const_cast<SearchNodeInfo&>(*info); } // hack to reduce code duplication

    public:
        SearchNode(std::unique_ptr<AbstractState> pa_state, unsigned current_step, SearchNodeInfo& info_):
            SearchNodeView(std::move(pa_state), current_step, info_) { PLAJA_ASSERT(currentStep >= 1) }

        ~SearchNode() final = default;

        SearchNode(const SearchNode& other) = delete;
        SearchNode(SearchNode&& other) = default;
        SearchNode& operator=(const SearchNode& other) = default;
        SearchNode& operator=(SearchNode&& other) = default;

        inline void set_reached_init() {
            PLAJA_ASSERT(not is_reached())
            PLAJA_ASSERT(not info->has_parent() or (info->_current_step() > 0))
            PLAJA_ASSERT(info->_current_step() <= currentStep)
            _info().set_current_step(currentStep);
            _info().set_start_distance(0);
        }

        inline void set_reached(StateID_type parent_id, UpdateOpID_type parent_update_op, SearchDepth_type start_distance) {
            PLAJA_ASSERT(not is_reached())
            PLAJA_ASSERT(not info->has_parent() or (info->_current_step() > 0))
            PLAJA_ASSERT(info->_current_step() <= currentStep)
            _info().set_current_step(currentStep);
            _info().add_parent(parent_id, parent_update_op);
            _info().set_start_distance(start_distance);
        }

        inline void set_unreached() {
            _info().set_current_step(0);
            _info().parents.clear();
        }

        inline void add_parent(StateID_type parent_id, UpdateOpID_type parent_update_op) {
            PLAJA_ASSERT(is_reached())
            _info().add_parent(parent_id, parent_update_op);
        }

        inline void set_goal_path_successor(StateID_type goal_path_successor, HeuristicValue_type goal_distance) {
            PLAJA_ASSERT(has_been_reached())
            _info().set_goal_path_successor(goal_path_successor, goal_distance);
        }

        inline void set_goal() {
            PLAJA_ASSERT(is_reached())
            _info().set_goal();
        }

        inline void set_dead_end() {
            PLAJA_ASSERT(is_reached())
            _info().set_dead_end();
        }

    };

}

#endif //PLAJA_SEARCH_NODE_PA_H
