//
// This file is part of the PlaJA code base (2019 - 2020).
//

#ifndef PLAJA_SEARCH_SPACE_PA_BASE_H
#define PLAJA_SEARCH_SPACE_PA_BASE_H

#include <list>
#include <memory>
#include "../../../include/ct_config_const.h"
#include "../../../stats/forward_stats.h"
#include "../../../utils/fd_imports/segmented_vector.h"
#include "../../../utils/structs_utils/forward_struct_utils.h"
#include "../../factories/forward_factories.h"
#include "../../using_search.h"
#include "../inc_state_space/forward_inc_state_space_pa.h"
#include "../pa_states/abstract_state.h"
#include "../pa_states/forward_pa_states.h"
#include "../smt/forward_smt_pa.h"
#include "../forward_pa.h"
#include "forward_search_space.h"
#include "search_node_pa.h"

class SearchSpacePABase {

protected:

    const ModelZ3PA* modelZ3; // underlying Z3 Model
    std::unique_ptr<InitialStatesEnumeratorPa> initialStatesEnumerator;
    StateID_type initialState; // (first) model initial state
    std::unordered_set<StateID_type> initialStates; // abstract initial states
    std::vector<StateID_type> cachedStartStates; // TODO redundant cache, though let's leave it like that for now, as Transport seems to be insanely lucky with the reverted start state expansion order resulting from unordered-set
    PredicatesNumber_type numPreds;
    //
    std::shared_ptr<StateRegistryPa> stateRegPa;
    bool useIncSearch;
    bool cacheWitnesses;
#ifdef USE_INC_REG
    std::shared_ptr<SEARCH_SPACE_PA::IncStateSpaceReg> incStateSpace;
#endif
#ifdef USE_MATCH_TREE
    std::shared_ptr<SEARCH_SPACE_PA::IncStateSpaceMt> incStateSpace;
#endif
    std::shared_ptr<SEARCH_SPACE_PA::SharedSearchSpace> sharedSearchSpace;
    //
    PLAJA::StatsBase* statsHeuristic;

public:
    explicit SearchSpacePABase(const PLAJA::Configuration& config, const ModelZ3PA& model_z3);
    virtual ~SearchSpacePABase();
    DELETE_CONSTRUCTOR(SearchSpacePABase)

    // short-cuts:
    [[nodiscard]] UpdateOpID_type compute_update_op(const PATransitionStructure& transition) const;
    [[nodiscard]] UpdateOpID_type compute_update_op(ActionOpID_type op, UpdateIndex_type update) const;
    [[nodiscard]] UpdateOpStructure compute_update_op(UpdateOpID_type update_op) const;
    [[nodiscard]] const class ActionOp& retrieve_action_op(ActionOpID_type op_id) const;

    void increment();
    void reset(); // for incremental usage

    [[nodiscard]] PredicateTraverser init_predicate_traverser() const;

    void compute_initial_states();
    StateID_type add_initial_state(const PredicateState& pa_initial_state);

    [[nodiscard]] inline std::size_t get_number_initial_states() const { return initialStates.size(); }

    [[nodiscard]] inline StateID_type _initialState() const {
        PLAJA_ASSERT(initialState != STATE::none)
        return initialState;
    }

    [[nodiscard]] inline const std::unordered_set<StateID_type>& _initialStates() const { return initialStates; }

    /** GET STATES ***************************************************************************************************/

    [[nodiscard]] std::unique_ptr<AbstractState> get_abstract_state(PredicatesNumber_type num_preds, StateID_type state_id) const;

    [[nodiscard]] std::unique_ptr<AbstractState> get_abstract_state(StateID_type state_id) const;

    [[nodiscard]] inline std::unique_ptr<AbstractState> get_abstract_initial_state() const {
        PLAJA_ASSERT(initialState != STATE::none)
        return get_abstract_state(initialState);
    }

    [[nodiscard]] std::unique_ptr<AbstractState> set_abstract_state(const PredicateState& pa_state_values) const;
    [[nodiscard]] std::unique_ptr<AbstractState> set_abstract_state(const AbstractState& different_reg_state) const;

    void unregister_state_if_possible(StateID_type id);

    FCT_IF_DEBUG([[nodiscard]] PredicateState get_pa_state_values(StateID_type state_id) const;)

    FCT_IF_DEBUG([[nodiscard]] PredicateState get_pa_initial_state_values() const;)

/** incremental interface *********************************************************************************************/

    inline StateID_type add_witness(const StateValues& witness) { return stateRegPa->add_witness(witness); }

    [[nodiscard]] inline State lookup_witness(StateID_type witness) const { return stateRegPa->lookup_witness(witness); }

    [[nodiscard]] inline std::unique_ptr<State> get_witness(StateID_type witness) const { return stateRegPa->get_witness(witness); }

    [[nodiscard]] StateID_type retrieve_witness(StateID_type source, UpdateOpID_type update_op_id, StateID_type successor) const;

    void set_excluded_label(const AbstractState& source, ActionLabel_type action_label);
    void set_excluded_op(const AbstractState& source, ActionOpID_type action_op);
    void store_entailments(const PATransitionStructure& transition, const std::unordered_map<PredicateIndex_type, bool>& entailments);
    void add_transition(const PATransitionStructure& transition, const StateValues& witness_src, const StateValues& witness_suc);
    void exclude_transition(const PATransitionStructure& transition);
    void refine_transitions(const AbstractState& source);

    [[nodiscard]] bool is_excluded_label(const AbstractState& source, ActionLabel_type action_label) const;
    [[nodiscard]] bool is_excluded_op(const AbstractState& source, ActionOpID_type action_op) const;

    PredicatesNumber_type load_entailments(PredicateState& target, const PATransitionStructure& transition, const class PredicateRelations* predicate_relations);

    [[nodiscard]] std::unique_ptr<StateID_type> transition_exists(const AbstractState& source, UpdateOpID_type update_op_id, const AbstractState& successor) const;
    [[nodiscard]] std::unique_ptr<StateID_type> transition_exists(const AbstractState& source, ActionOpID_type action_op_id, UpdateIndex_type update_index, const AbstractState& successor) const;
    [[nodiscard]] std::unique_ptr<StateID_type> transition_exists(StateID_type source, UpdateOpID_type update_op_id, StateID_type successor) const;

/** search node *******************************************************************************************************/

private:
    unsigned int currentStep;
    segmented_vector::SegmentedVector<SEARCH_SPACE_PA::SearchNodeInfo> searchInformation;

public:
    [[nodiscard]] inline std::size_t information_size() const { return searchInformation.size(); } // mainly for debugging purposes

    /**
     * Add node to search space if necessary.
     * Can also be used as normal get_node.
     * @param state
     * @return
     */
    [[nodiscard]] inline SEARCH_SPACE_PA::SearchNode add_node(std::unique_ptr<AbstractState>&& abstract_state) {
        const StateID_type state_id = abstract_state->get_id_value();
        if (state_id >= searchInformation.size()) { searchInformation.resize(state_id + 1); }
        return { std::move(abstract_state), currentStep, searchInformation[state_id] };
    }

    [[nodiscard]] inline SEARCH_SPACE_PA::SearchNode get_node(StateID_type state_id) {
        PLAJA_ASSERT(state_id < searchInformation.size())
        return { get_abstract_state(state_id), currentStep, searchInformation[state_id] };
    }

    [[nodiscard]] inline SEARCH_SPACE_PA::SearchNodeView get_node(StateID_type state_id) const {
        PLAJA_ASSERT(state_id < searchInformation.size())
        PLAJA_ASSERT(SEARCH_SPACE_PA::SearchNodeView(get_abstract_state(state_id), currentStep, searchInformation[state_id]).is_reached())
        return { get_abstract_state(state_id), currentStep, searchInformation[state_id] };
    }


/** goals states *******************************************************************************************************/

private:
    std::list<StateID_type> goalPathFrontier;
    std::size_t numGoalPathStates;

public:
    inline void add_goal_path_frontier(StateID_type state_id) { goalPathFrontier.push_back(state_id); }

    [[nodiscard]] inline bool goal_path_frontier_empty() const { return goalPathFrontier.empty(); }

    [[nodiscard]] inline StateID_type goal_frontier_state() const {
        PLAJA_ASSERT(not goal_path_frontier_empty())
        return *goalPathFrontier.cbegin();
    }

    [[nodiscard]] inline std::size_t number_of_goal_path_states() const { return numGoalPathStates; }

/** post search ********************************************************************************************************/

private:
    [[nodiscard]] AbstractPath extract_path_backwards(StateID_type state_id) const;
    [[nodiscard]] AbstractPath extract_goal_path_forward(StateID_type state_id) const; // !empty path if unsafe source or safe

public:

    std::size_t compute_goal_path_states(); // returns the number of goal path states
    void compute_dead_ends_backwards(StateID_type state_id);
    [[nodiscard]] AbstractPath extract_goal_path(StateID_type state_id) const;

    [[nodiscard]] std::size_t number_of_initial_goal_path_states() const;

    [[nodiscard]] inline bool has_goal_path(StateID_type state_id) const { return get_node(state_id).has_goal_path(); }

/** stats *************************************************************************************************************/

public:

    // Statistics:
    void update_statistics() const;
    void print_statistics() const;
    void stats_to_csv(std::ofstream& file) const;
    static void stat_names_to_csv(std::ofstream& file);

};

#endif //PLAJA_SEARCH_SPACE_PA_BASE_H
