//
// This file is part of the PlaJA code base (2019 - 2021).
//

#include "solution_checker.h"
#include "../../parser/ast/model.h"
#include "../../globals.h"
#include "../factories/configuration.h"
#include "../information/model_information.h"
#include "../non_prob_search/policy/policy_restriction.h"
#include "../successor_generation/action_op.h"
#include "../successor_generation/successor_generator_c.h"
#include "pa_states/abstract_state.h"

SolutionChecker::SolutionChecker(const PLAJA::Configuration& config, const Model& model_ref, const Jani2NNet* nn_interface):
    modelInfo(&model_ref.get_model_information())
    , policy(nn_interface ? new PolicyRestriction(config, *nn_interface) : nullptr) {

    if (not config.has_sharable(PLAJA::SharableKey::SUCC_GEN)) {
        config.set_sharable<SuccessorGeneratorC>(PLAJA::SharableKey::SUCC_GEN, std::make_shared<SuccessorGeneratorC>(config, model_ref));
    }

    successorGenerator = config.get_sharable_as_const<SuccessorGeneratorC>(PLAJA::SharableKey::SUCC_GEN);

    PLAJA_ASSERT(&successorGenerator->get_model() == &model_ref)

}

SolutionChecker::~SolutionChecker() = default;

/* */

const Model& SolutionChecker::get_model() const { return successorGenerator->get_model(); }

/* */

const ActionOp& SolutionChecker::get_action_op(ActionOpID_type action_op_id) const { return successorGenerator->get_action_op(action_op_id); }

/* */

const StateValues& SolutionChecker::get_models_initial_values() const { return modelInfo->get_initial_values(); }

StateValues SolutionChecker::get_solution_constructor(const AbstractState& pa_state) const {
    // set locations in constructor
    auto solution = modelInfo->get_initial_values();
    for (auto loc_it = pa_state.init_loc_state_iterator(); !loc_it.end(); ++loc_it) {
        solution.assign_int<false>(loc_it.state_index(), loc_it().location_value(loc_it.state_index()));
    }
    return solution;
}

/* */

StateValues SolutionChecker::compute_successor_solution(const StateValues& solution, const ActionOp& action_op, UpdateIndex_type update_index) const {
    PLAJA_ASSERT(update_index != ACTION::noneUpdate)
    PLAJA_ASSERT(not action_op.has_zero_probability(update_index, solution))
    return successorGenerator->compute_successor(action_op.get_update(update_index), solution);
}

bool SolutionChecker::agrees(const StateBase& solution, const ActionOp& action_op, UpdateIndex_type update_index, const StateBase& successor) const {
    PLAJA_ASSERT(not successorGenerator->has_transient_variables())
    PLAJA_ASSERT(update_index != ACTION::noneUpdate)
    PLAJA_ASSERT(not action_op.has_zero_probability(update_index, solution))
    return action_op.get_update(update_index).agrees(solution, successor);
}

/* */

#ifdef ENABLE_TERMINAL_STATE_SUPPORT

bool SolutionChecker::check_terminal(const StateBase& state) const { return successorGenerator->is_terminal(state); }

const Expression* SolutionChecker::get_terminal() const { return successorGenerator->get_terminal_state_condition(); }

#endif

/* */

bool SolutionChecker::check_applicability(const StateBase& source, const ActionOp& action_op, const AbstractState* abstract_src) const {
    PLAJA_ASSERT_EXPECT(not check_terminal(source))

    if (not action_op.is_enabled(source)) {

        STMT_IF_DEBUG(
            if (PLAJA_GLOBAL::do_additional_outputs()) {
                PLAJA_LOG_DEBUG("Guard no sat.")
                source.dump(true);
                action_op.dump(nullptr, true);
            }
        )

        return false;
    }

    if (abstract_src and not abstract_src->is_abstraction(source)) {

        STMT_IF_DEBUG(

            if (PLAJA_GLOBAL::do_additional_outputs()) {
                PLAJA_LOG_DEBUG("Source not in abstract state.")
                source.dump(true);
                abstract_src->dump();
            }

        )

        return false;
    }

    return true;
}

bool SolutionChecker::check_transition(const StateValues& source, const AbstractState& abstract_src, const ActionOp& action_op, UpdateIndex_type update_index, const AbstractState& abstract_target, const StateValues* target) const {
    if (not check_applicability(source, action_op, &abstract_src) or action_op.has_zero_probability(update_index, source)) { return false; }

    if (target) {

        STMT_IF_DEBUG(PLAJA_GLOBAL::push_additional_outputs_frame(true))
        PLAJA_ASSERT(agrees(source, action_op, update_index, *target))
        STMT_IF_DEBUG(PLAJA_GLOBAL::pop_additional_outputs_frame())

        if (not abstract_target.is_abstraction(*target)) {

            STMT_IF_DEBUG(
                if (PLAJA_GLOBAL::do_additional_outputs()) {
                    PLAJA_LOG_DEBUG("Target not in abstract state.")
                    target->dump(true);
                    abstract_target.dump();
                }
            )

            return false;
        }

        if (not modelInfo->is_valid(*target)) { // Target may actually be invalid.

            STMT_IF_DEBUG(

                if (PLAJA_GLOBAL::do_additional_outputs()) {
                    PLAJA_LOG_DEBUG("Target not in bounds.")
                    target->dump(true);
                }

            )

            return false;
        }

        return true;

    } else {

        const auto successor = successorGenerator->compute_successor(action_op.get_update(update_index), source);
        return abstract_target.is_abstraction(successor) && modelInfo->is_valid(successor);

    }

}

bool SolutionChecker::check_policy(const StateBase& source, ActionLabel_type action_label, const AbstractState* abstract_src) const {
    PLAJA_ASSERT(policy)
    PLAJA_ASSERT_EXPECT(not check_terminal(source))

    if (not policy->is_enabled(source, action_label)) {
        STMT_IF_DEBUG(if (PLAJA_GLOBAL::do_additional_outputs()) { policy->dump_cache(); })
        return false;
    }

    if (abstract_src and not abstract_src->is_abstraction(source)) {

        STMT_IF_DEBUG(

            if (PLAJA_GLOBAL::do_additional_outputs()) {
                PLAJA_LOG_DEBUG("Source not in abstract state.")
                source.dump(true);
                abstract_src->dump();
            }

        )

        return false;
    }

    return true;
}

bool SolutionChecker::check_policy_applicability(const StateValues& source, const ActionOp& action_op, const AbstractState* abstract_src) const {
    return check_applicability(source, action_op, nullptr) && check_policy(source, action_op._action_label(), abstract_src);
}

bool SolutionChecker::check_policy_transition(const StateValues& source, const AbstractState& abstract_src, const ActionOp& action_op, UpdateIndex_type update_index, const AbstractState& abstract_target, const StateValues* target) const {
    return check_transition(source, abstract_src, action_op, update_index, abstract_target, target) && check_policy(source, action_op._action_label(), nullptr);
}

/* Interface. */

bool SolutionChecker::check_state(const StateValues& concrete_state) const { return modelInfo->is_valid(concrete_state); }

bool SolutionChecker::check_abstraction(const StateValues& concrete_state, const AbstractState& abstract_state) const { return check_state(concrete_state) and abstract_state.is_abstraction(concrete_state); }

bool SolutionChecker::check(const StateValues& solution, const AbstractState* pa_src, ActionLabel_type action_label, ActionOpID_type action_op_id, UpdateIndex_type update_index, const AbstractState* pa_target, const StateValues* target) const {

    PLAJA_ASSERT(check_state(solution)) // Should hold ...

    PLAJA_ASSERT_EXPECT(not check_terminal(solution)) // Let's revisit if this occurs ...
    if constexpr (PLAJA_GLOBAL::enableTerminalStateSupport) { if (check_terminal(solution)) { return false; } }

    PLAJA_ASSERT(policy) // So far non-NN case is not supported. // TODO reset chosenWithTolerance flag in policy if we do not call the policy
    if (pa_target) {
        PLAJA_ASSERT(action_op_id != ACTION::noneOp)
        PLAJA_ASSERT(action_label == ACTION::noneLabel or action_label == get_action_op(action_op_id)._action_label())
        PLAJA_ASSERT(update_index != ACTION::noneUpdate)
        return check_policy_transition(solution, *pa_src, get_action_op(action_op_id), update_index, *pa_target, target); // transition test
    } else {
        if (action_op_id != ACTION::noneOp) {
            PLAJA_ASSERT(action_label == ACTION::noneLabel or action_label == get_action_op(action_op_id)._action_label())
            return check_policy_applicability(solution, get_action_op(action_op_id), pa_src);  // applicability test
        } else {
            PLAJA_ASSERT(action_label != ACTION::noneLabel)
            return check_policy(solution, action_label, pa_src);  // selection test
        }
    }

}

/* */

PLAJA::floating SolutionChecker::policy_selection_delta() const { return policy ? policy->get_selection_delta() : 0; }

bool SolutionChecker::policy_chosen_with_tolerance() const { return policy and policy->is_chosen_with_tolerance(); }

