//
// This file is part of the PlaJA code base (2019 - 2021).
//

#ifndef PLAJA_SOLUTION_CHECKER_H
#define PLAJA_SOLUTION_CHECKER_H

#include <list>
#include <memory>
#include <vector>
#include "../../include/ct_config_const.h"
#include "../../parser/ast_forward_declarations.h"
#include "../../utils/default_constructors.h"
#include "../../assertions.h"
#include "../factories/forward_factories.h"
#include "../information/jani2nnet/forward_jani_2_nnet.h"
#include "../information/jani2nnet/using_jani2nnet.h"
#include "../information/forward_information.h"
#include "../non_prob_search/policy/forward_policy.h"
#include "../states/forward_states.h"
#include "../successor_generation/forward_successor_generation.h"
#include "pa_states/forward_pa_states.h"

class SolutionChecker final {

private:
    const ModelInformation* modelInfo;
    std::shared_ptr<const SuccessorGeneratorC> successorGenerator;
    std::unique_ptr<PolicyRestriction> policy;

public:
    /**
     * @param nn_interface : Policy is not supported iff nn_interface is NULL.
     * @param config : Successor generator is constructed internally iff not provided.
     */
    SolutionChecker(const PLAJA::Configuration& config, const Model& model, const Jani2NNet* nn_interface);
    ~SolutionChecker();
    DELETE_CONSTRUCTOR(SolutionChecker)

    [[nodiscard]] const Model& get_model() const;

    [[nodiscard]] inline const ModelInformation& get_model_info() const { return *modelInfo; }

    [[nodiscard]] inline bool has_policy() const { return policy.get(); }

    /* getter */
    [[nodiscard]] const ActionOp& get_action_op(ActionOpID_type action_op_id) const;

    // solution interface:
    [[nodiscard]] const StateValues& get_models_initial_values() const;

    [[nodiscard]] inline const StateValues& get_solution_constructor() const { return get_models_initial_values(); }

    [[nodiscard]] StateValues get_solution_constructor(const AbstractState& pa_state) const;
    // TODO: might try to reuse last solution instead
    //  intention: keep solution values (of preceding queries) for variables non-relevant to the current query;
    //  however: unlikely to be beneficial as source variables are always present in solver structures due to the predicates (in PA context).

    [[nodiscard]] StateValues compute_successor_solution(const StateValues& solution, const ActionOp& action_op, UpdateIndex_type update_index) const;
    [[nodiscard]] bool agrees(const StateBase& solution, const ActionOp& action_op, UpdateIndex_type update_index, const StateBase& successor) const;

    /* Check interface. */

    [[nodiscard]] bool check_state(const StateValues& concrete_state) const;
    [[nodiscard]] bool check_abstraction(const StateValues& concrete_state, const AbstractState& abstract_state) const;

#ifdef ENABLE_TERMINAL_STATE_SUPPORT
    [[nodiscard]] bool check_terminal(const StateBase& state) const;
    [[nodiscard]] const Expression* get_terminal() const;
#else

    [[nodiscard]] inline constexpr bool check_terminal(const StateBase&) const { return false; }

    [[nodiscard]] inline const Expression* get_terminal() const { PLAJA_ABORT }

#endif

    [[nodiscard]] bool check_applicability(const StateBase& source, const ActionOp& action_op, const AbstractState* abstract_src) const;
    [[nodiscard]] bool check_transition(const StateValues& source, const AbstractState& abstract_src, const ActionOp& action_op, UpdateIndex_type update_index, const AbstractState& abstract_target, const StateValues* target = nullptr) const;
    [[nodiscard]] bool check_policy(const StateBase& source, ActionLabel_type action_label, const AbstractState* abstract_src) const;
    [[nodiscard]] bool check_policy_applicability(const StateValues& source, const ActionOp& action_op, const AbstractState* abstract_src) const;
    [[nodiscard]] bool check_policy_transition(const StateValues& source, const AbstractState& abstract_src, const ActionOp& action_op, UpdateIndex_type update_index, const AbstractState& abstract_target, const StateValues* target = nullptr) const;

    [[nodiscard]] bool check(const StateValues& solution, const AbstractState* pa_src, ActionLabel_type action_label, ActionOpID_type action_op_id, UpdateIndex_type update_index, const AbstractState* pa_target, const StateValues* target = nullptr) const;

    /** Delta between label of interest and label chosen with precision. */
    [[nodiscard]] PLAJA::floating policy_selection_delta() const;

    /** Did policy check pass due to tolerance? */
    [[nodiscard]] bool policy_chosen_with_tolerance() const;

};

#endif //PLAJA_SOLUTION_CHECKER_H
