//
// This file is part of the PlaJA code base (2019 - 2023).
//

#include "model_smt.h"
#include "../../../parser/ast/expression/expression.h"
#include "../../../parser/ast/expression/non_standard/predicates_expression.h"
#include "../../../parser/ast/expression/non_standard/state_condition_expression.h"
#include "../../../parser/ast/model.h"
#include "../../../parser/visitor/extern/to_normalform.h"
#include "../../factories/configuration.h"
#include "../../information/jani2nnet/input_feature_to_jani.h"
#include "../../information/jani2nnet/jani_2_nnet.h"
#include "../../information/model_information.h"
#include "../../information/property_information.h"
#include "../../states/state_values.h"
#include "../../successor_generation/successor_generator_c.h"
#include "smt_constraint.h"
#include "smt_solver.h"

ModelSmt::ModelSmt(std::shared_ptr<PLAJA::SmtContext>&& context, const PLAJA::Configuration& config): // NOLINT(modernize-pass-by-value)
    context(std::move(context))
    , model(nullptr)
    , modelInfo(nullptr)
    , nnInterface(nullptr)
    , predicates(nullptr)
    , cachedOpApp(nullptr)
    , ignoreLocations(false)
    , maxStep(1) {

    if (config.has_sharable_const(PLAJA::SharableKey::MODEL)) {
        model = config.get_sharable_const<Model>(PLAJA::SharableKey::MODEL);
    } else { PLAJA_ASSERT(config.has_sharable_const(PLAJA::SharableKey::PROP_INFO)) }

    if (config.has_sharable_const(PLAJA::SharableKey::PROP_INFO)) {
        const auto* prop_info = config.get_sharable_const<PropertyInformation>(PLAJA::SharableKey::PROP_INFO);
        if (not model) { model = prop_info->get_model(); }
        else { PLAJA_ASSERT(model == prop_info->get_model()) }

        nnInterface = prop_info->get_nn_interface();
        set_start(prop_info->get_start() ? prop_info->get_start()->deepCopy_Exp() : nullptr);
        set_reach(prop_info->get_reach() ? prop_info->get_reach()->deepCopy_Exp() : nullptr);
        if constexpr (PLAJA_GLOBAL::enableTerminalStateSupport) { set_non_terminal(prop_info->get_terminal() ? prop_info->get_terminal()->deepCopy_Exp() : nullptr, true); }

        predicates = prop_info->get_predicates();
    }

    modelInfo = &model->get_model_information();

    // only use locations if there is more than one location state?
    if (not ignoreLocations) {
        ignoreLocations = true;
        for (auto loc_it = modelInfo->locationIndexIterator(); !loc_it.end(); ++loc_it) {
            if (loc_it.range() > 1) {
                ignoreLocations = false;
                break;
            }
        }
    }

    if (nnInterface) { set_nn_interface(nnInterface); } // Reset to properly trigger computation of extended structures.

    if (config.has_sharable(PLAJA::SharableKey::SUCC_GEN)) {
        successorGenerator = config.get_sharable<SuccessorGeneratorC>(PLAJA::SharableKey::SUCC_GEN);
        PLAJA_ASSERT(&successorGenerator->get_model() == model)
    } else {
        successorGenerator = config.set_sharable(PLAJA::SharableKey::SUCC_GEN, std::make_shared<SuccessorGeneratorC>(config, *model));
    }

}

ModelSmt::~ModelSmt() = default;

/**********************************************************************************************************************/

void ModelSmt::set_nn_interface(const Jani2NNet* nn_interface) {

    PLAJA_ASSERT(nn_interface)

    PLAJA_ASSERT_EXPECT(not nnInterface or nn_interface == nnInterface)

    nnInterface = nn_interface;

    // Cache input locations.
    for (InputIndex_type input_index = 0; input_index < nnInterface->get_num_input_features(); ++input_index) {
        const auto* input_feature = nnInterface->get_input_feature(input_index);
        if (input_feature->is_location()) { inputLocations.push_back(input_feature->stateIndex); }
    }

    PLAJA_ASSERT(inputLocations.empty() or not ignoreLocations)

}

void ModelSmt::set_start(std::unique_ptr<Expression>&& start_ptr) { start = std::move(start_ptr); }

void ModelSmt::set_reach(std::unique_ptr<Expression>&& reach_ptr) { reach = std::move(reach_ptr); }

#ifdef ENABLE_TERMINAL_STATE_SUPPORT

void ModelSmt::set_non_terminal(std::unique_ptr<Expression>&& terminal, bool negate) {
    PLAJA_ASSERT(PLAJA_GLOBAL::enableTerminalStateSupport)
    if (terminal) {
        auto terminal_state_con = StateConditionExpression::transform(std::move(terminal));
        PLAJA_ASSERT(terminal_state_con->get_size_loc_values() == 0) // Not supported.
        nonTerminal = terminal_state_con->set_constraint();
        if (negate) { TO_NORMALFORM::negate(nonTerminal); }
        if constexpr (not PLAJA_GLOBAL::reachMayBeTerminal) {
            std::list<std::unique_ptr<Expression>> reach_non_terminal_conjuncts;
            reach_non_terminal_conjuncts.push_back(nonTerminal->deepCopy_Exp());
            reach_non_terminal_conjuncts.push_back(reach->deepCopy_Exp());
            reachNonTerminal = TO_NORMALFORM::construct_conjunction(std::move(reach_non_terminal_conjuncts));
        } else { reachNonTerminal = nullptr; }
    } else {
        nonTerminal = nullptr;
        reachNonTerminal = nullptr;
    }
    nonTerminalSmt.clear();

}

#else

void ModelSmt::set_non_terminal(std::unique_ptr<Expression>&&, bool) { PLAJA_ABORT }

#endif

/**********************************************************************************************************************/

std::size_t ModelSmt::get_number_automata_instances() const { return get_model().get_number_automataInstances(); }

std::size_t ModelSmt::get_array_size(VariableID_type var_id) const { return get_model_info().get_array_size(var_id); }

bool ModelSmt::initial_state_is_subsumed_by_start() const { return start and start->evaluate_integer(modelInfo->get_initial_values()); }

const Expression* ModelSmt::get_terminal() const { return successorGenerator->get_terminal_state_condition(); }

std::size_t ModelSmt::get_number_predicates() const {
    PLAJA_ASSERT(predicates)
    return predicates->get_number_predicates();
}

const Expression* ModelSmt::get_predicate(std::size_t index) const {
    PLAJA_ASSERT(predicates)
    return predicates->get_predicate(index);
}

bool ModelSmt::predicate_is_bound(std::size_t index) const { return predicates->predicate_is_bound(index); }

/**********************************************************************************************************************/

const ActionOp& ModelSmt::get_action_op_concrete(ActionOpID_type op_id) const { return successorGenerator->get_action_op(op_id); }

/**********************************************************************************************************************/

std::unique_ptr<Expression> ModelSmt::optimize_expression(const Expression& expr, PLAJA::SmtSolver& solver, bool prepare) const {
    PLAJA_LOG_DEBUG("Optimizing expression for SMT encoding ...")

    auto expr_copy = expr.deepCopy_Exp();

    auto conjuncts = TO_NORMALFORM::split_conjunction(std::move(expr_copy), false);
    PLAJA_LOG_DEBUG(PLAJA_UTILS::string_f("%i conjuncts ...", conjuncts.size()))

    std::list<std::unique_ptr<Expression>> conjuncts_processed;

    for (auto& conjunct: conjuncts) {

        if (prepare) { TO_NORMALFORM::to_dnf(conjunct); }
        auto disjuncts = TO_NORMALFORM::split_disjunction(std::move(conjunct), false);

        PLAJA_ASSERT(not disjuncts.empty())

        if (disjuncts.size() == 1) { conjuncts_processed.push_back(std::move(disjuncts.front())); }
        else {

            PLAJA_LOG_DEBUG(PLAJA_UTILS::string_f("%i disjuncts in conjunct ...", disjuncts.size()))

            auto it = disjuncts.begin();
            for (; it != disjuncts.end();) {

                solver.push();

                add_to_solver(solver, **it, 0);

                const auto pre_rlt = solver.pre_check(); // Expecting many simple constraints, so pre-check makes sense.
                if (not pre_rlt) {
                    if (solver.check()) { ++it; }
                    else { it = disjuncts.erase(it); }
                } else if (*pre_rlt) { ++it; }
                else { it = disjuncts.erase(it); }

                solver.pop();

            }

            PLAJA_ASSERT(not disjuncts.empty())

            PLAJA_LOG_DEBUG(PLAJA_UTILS::string_f("%i disjuncts after preprocessing ...", disjuncts.size()))

            conjuncts_processed.push_back(TO_NORMALFORM::construct_disjunction(std::move(disjuncts)));
        }

    }

    PLAJA_ASSERT(conjuncts.size() == conjuncts_processed.size())

    auto expr_opt = TO_NORMALFORM::construct_conjunction(std::move(conjuncts_processed));

    return expr_opt;
}

/* */

#ifdef ENABLE_TERMINAL_STATE_SUPPORT

void ModelSmt::add_terminal(PLAJA::SmtSolver& solver, StepIndex_type step) const {
    if (nonTerminalSmt.empty()) { return; }
    PLAJA_ASSERT(step < nonTerminalSmt.size())
    nonTerminalSmt[step]->add_negation_to_solver(solver);
}

void ModelSmt::exclude_terminal(PLAJA::SmtSolver& solver, StepIndex_type step) const {
    if (nonTerminalSmt.empty()) { return; }
    PLAJA_ASSERT(step < nonTerminalSmt.size())
    nonTerminalSmt[step]->add_to_solver(solver);
}

#else

void ModelSmt::add_terminal(PLAJA::SmtSolver&, StepIndex_type) const {}

void ModelSmt::exclude_terminal(PLAJA::SmtSolver&, StepIndex_type) const {}

#endif