//
// This file is part of the PlaJA code base (2019 - 2022).
//

#include "solution_enumerator_naive.h"
#include "../../../exception/not_supported_exception.h"
#include "../../../parser/ast/expression/expression.h"
#include "../../../utils/utils.h"
#include "../../information/model_information.h"
#include "../../states/state_values.h"

SolutionEnumeratorNaive::SolutionEnumeratorNaive(const ModelInformation& model_info, const Expression& state_condition): modelInformation(model_info), stateCondition(state_condition) {}

SolutionEnumeratorNaive::~SolutionEnumeratorNaive() = default;

void SolutionEnumeratorNaive::compute_solutions_rec(VariableIndex_type state_index, StateValues& state_values) { // NOLINT(misc-no-recursion)
    if (state_index == modelInformation.get_int_state_size()) {
        // sanity: floating not supported:
        PLAJA_ASSERT(modelInformation.get_int_state_size() == modelInformation.get_state_size())
        PLAJA_ASSERT(modelInformation.get_floating_state_size() == 0)
        //
        if (stateCondition.evaluate_integer(state_values)) { solutions.push_back(state_values); }
    } else {
        auto lower_bound = modelInformation.get_lower_bound_int(state_index);
        auto upper_bound = modelInformation.get_upper_bound_int(state_index);
        auto next_state_index = state_index + 1;
        for (auto value = lower_bound; value <= upper_bound; ++value) {
            state_values.assign_int<false>(state_index, value);
            compute_solutions_rec(next_state_index, state_values);
        }
    }
}

std::list<StateValues> SolutionEnumeratorNaive::compute_solutions(const Expression& state_condition, const ModelInformation& model_info) {
    static bool warning_printed(false);
    if (not warning_printed) {
        PLAJA_LOG(PLAJA_UTILS::to_red_string("Warning: Naive solution enumeration."))
        PLAJA_LOG(PLAJA_UTILS::to_red_string("Consider building SMT-Module (dependency: Z3) to enable scalable enumeration methods."))
        warning_printed = true;
    }
    if (model_info.get_floating_state_size() > 0) { throw NotSupportedException("Naive solution enumeration cannot handle continuous state spaces."); }
    SolutionEnumeratorNaive solution_enumerator(model_info, state_condition);
    StateValues state_values(model_info.get_int_state_size(), 0);
    solution_enumerator.compute_solutions_rec(0, state_values);
    return std::move(solution_enumerator.solutions);
}