//
// This file is part of the PlaJA code base (2019 - 2020).
//

#include <z3++.h>
#include "solution_enumerator_z3.h"
#include "../../states/state_values.h"
#include "../model/model_z3.h"
#include "../solver/smt_solver_z3.h"
#include "../solver/solution_z3.h"
#include "../vars_in_z3.h"

SolutionEnumeratorZ3::SolutionEnumeratorZ3(Z3_IN_PLAJA::SMTSolver& solver_, const ModelZ3& model_z3):
    solver(&solver_)
    , modelZ3(&model_z3)
    , stateIndexes(model_z3.get_src_vars().cache_state_indexes(false))
    , stateVars(nullptr) { // only for customized state indexes
}

SolutionEnumeratorZ3::~SolutionEnumeratorZ3() = default;

//

void SolutionEnumeratorZ3::set_variables_of_interest(const std::vector<VariableID_type>& var_ids, bool use_target_vars) {
    // compute id to expr mapping
    stateVars = std::make_unique<Z3_IN_PLAJA::StateVarsInZ3>(modelZ3->get_model(), modelZ3->get_context());
    const auto& state_vars = use_target_vars ? modelZ3->get_target_vars() : modelZ3->get_src_vars();
    for (const VariableID_type var_id: var_ids) { stateVars->set(var_id, state_vars.to_z3(var_id)); }
    //
    stateIndexes = stateVars->cache_state_indexes(false);
}

bool SolutionEnumeratorZ3::get_solution(StateValues& solution_state) {
    if (not solver->check()) { return false; } // problem is unsat
    auto solution = solver->get_solution();
    // new solution found, extract and constraint for next solution ...
    stateIndexes->extract_solution(*solution, solution_state);
    // add solution constraint:
    solver->add(!stateIndexes->encode_solution(*solution));
    //
    return true; // solution found
}

std::list<StateValues> SolutionEnumeratorZ3::compute_solutions(Z3_IN_PLAJA::SMTSolver& solver, const ModelZ3& model_z3, const StateValues& constructor) {
    std::list<StateValues> solutions;

    SolutionEnumeratorZ3 solution_enumerator(solver, model_z3);
    solutions.push_back(constructor);
    while (solution_enumerator.get_solution(solutions.back())) { solutions.push_back(constructor); }
    solutions.pop_back(); // pop last state which does not correspond to a new state

    return solutions;
}