//
// This file is part of the PlaJA code base (2019 - 2021).
//

#ifndef PLAJA_PER_ACTIONS_ITERATOR_H
#define PLAJA_PER_ACTIONS_ITERATOR_H

#include "../../using_search.h"


// forward declaration:
template<typename ActionOp_t, typename ActionOpView_t>
class ActionLabelIteratorBase;
template<typename ActionOp_t, typename ActionOpView_t>
class ActionOpIteratorBase;


template<typename ActionOp_t, typename ActionOpView_t>
class PerActionIteratorBase {
    friend class SuccessorGeneratorC;
    friend class SuccessorGeneratorAbstract;
    friend class SuccessorGeneratorPA;
    friend class ActionLabelIteratorBase<ActionOp_t, ActionOpView_t>;
    friend class ActionOpIteratorBase<ActionOp_t, ActionOpView_t>;

private:
    typename std::list<ActionOp_t*>::const_iterator it;
    typename std::list<ActionOp_t*>::const_iterator itEnd;

    explicit PerActionIteratorBase(const std::list<ActionOp_t*>& operator_list): it(operator_list.cbegin()), itEnd(operator_list.cend()) {}

public:
    inline void operator++() { ++it; }
    [[nodiscard]] inline bool end() const { return it == itEnd; }
    [[nodiscard]] inline const ActionOpView_t* operator()() const { return static_cast<ActionOpView_t*>(*it); }
    [[nodiscard]] inline const ActionOpView_t* operator->() const { return static_cast<ActionOpView_t*>(*it); }
    [[nodiscard]] inline const ActionOpView_t& operator*() const { return *static_cast<ActionOpView_t*>(*it); }


    // routines (for PA) to cache guard enabled value internally:
    template<typename Solver_t, bool incremental>
    bool guard_enabled(Solver_t& solver) const {
        ActionOp_t& action_op = **it;
        bool rlt;
        if constexpr(incremental) {
            rlt = action_op.guard_enabled_inc(solver); // does not pop afterwards
        } else {
            rlt = action_op.guard_enabled(solver);
        }
        action_op.set_guard_enabled(rlt);
        return rlt;
    }

};

/**********************************************************************************************************************/

template<typename ActionOp_t, typename ActionOpView_t>
class ActionLabelIteratorBase {
    friend class SuccessorGeneratorC;
    friend class SuccessorGeneratorAbstract;
    friend class SuccessorGeneratorPA;

private:
    std::list<std::pair<ActionLabel_type, const std::list<ActionOp_t*>*>> applicable_actions;
    typename std::list<std::pair<ActionLabel_type, const std::list<ActionOp_t*>*>>::const_iterator it;
    typename std::list<std::pair<ActionLabel_type, const std::list<ActionOp_t*>*>>::const_iterator itEnd;

    explicit ActionLabelIteratorBase(const std::list<ActionOp_t*>& silent_ops, const std::vector<std::list<ActionOp_t*>>& labeled_ops) {
        // silent
        if (!silent_ops.empty()) {
            applicable_actions.emplace_back(ACTION::silentAction, &silent_ops);
        }
        // labeled
        ActionLabel_type action_label = 0;
        for (const auto& labeled_ops_per_action: labeled_ops) {
            if (!labeled_ops_per_action.empty()) {
                applicable_actions.emplace_back(action_label, &labeled_ops_per_action);
            }
            ++action_label;
        }

        it = applicable_actions.cbegin();
        itEnd = applicable_actions.cend();
    }

public:
    inline void operator++() { ++it; }
    [[nodiscard]] inline bool end() const { return it == itEnd; }

    [[nodiscard]] inline ActionLabel_type action_label() const { return it->first; }
    [[nodiscard]] inline std::size_t num_applicable_ops() const { return it->second->size(); }
    [[nodiscard]] inline PerActionIteratorBase<ActionOp_t, ActionOpView_t> iterator() const { return PerActionIteratorBase<ActionOp_t, ActionOpView_t>(*it->second); }

};

/**********************************************************************************************************************/

template<typename ActionOp_t, typename ActionOpView_t>
class ActionOpIteratorBase {
    friend class SuccessorGeneratorC;
    friend class SuccessorGeneratorAbstract;
    friend class SuccessorGeneratorPA;

private:
    ActionLabel_type action_label;
    const std::vector<std::list<ActionOp_t*>>& labeledOps;
    PerActionIteratorBase<ActionOp_t, ActionOpView_t> it;

    // aux:
    inline void increment_until_op() {
        if (it.end()) {
            auto num_labels = labeledOps.size();
            for (++action_label; it.end() && action_label < num_labels; ++action_label) {
                it = PerActionIteratorBase<ActionOp_t, ActionOpView_t>(labeledOps[action_label]);
            }
        }
    }

    explicit ActionOpIteratorBase(const std::list<ActionOp_t*>& silent_ops, const std::vector<std::list<ActionOp_t*>>& labeled_ops):
            action_label(-1),
            labeledOps(labeled_ops),
            it(silent_ops) {
        increment_until_op();
    }

public:
    inline void operator++() {
        ++it;
        increment_until_op();
    }

    [[nodiscard]] inline bool end() const { return action_label == labeledOps.size(); }

    [[nodiscard]] inline const ActionOpView_t* operator()() const { return it(); }
    [[nodiscard]] inline const ActionOpView_t* operator->() const { return it.operator->(); }
    [[nodiscard]] inline const ActionOpView_t& operator*() const { return *it; }

};

#endif //PLAJA_PER_ACTIONS_ITERATOR_H
