//
// This file is part of the PlaJA code base (2019 - 2020).
//

#ifndef PLAJA_SUCCESSOR_GENERATOR_C_H
#define PLAJA_SUCCESSOR_GENERATOR_C_H

#include <list>
#include <memory>
#include <unordered_set>
#include <vector>
#include "../../assertions.h"
#include "../states/forward_states.h"
#include "../successor_generation/forward_successor_generation.h"
#include "base/iterators.h"
#include "base/successor_generator_base.h"

struct ActionOpStructure;

namespace MARABOU_IN_PLAJA { class SuccessorGenerator; }

/** caching ActionOps at initialization/construction time */
class SuccessorGeneratorC final: public SuccessorGeneratorBase {
    friend class SuccessorGeneratorAbstract; //
    friend class SuccessorGeneratorPA; //
    friend class SuccessorGeneratorToZ3; //
    friend MARABOU_IN_PLAJA::SuccessorGenerator;

private:
    // action op construction structures
    std::vector<std::unique_ptr<ActionOp>> nonSyncOps;
    std::vector<std::vector<std::unique_ptr<ActionOp>>> syncOps;

    // actual operator list at exploration time
    std::list<const ActionOp*> silentOps; // operators with silent label
    std::vector<std::list<const ActionOp*>> labeledOps; // labeled (non-silent) operators per action label

    // per state exploration:
    const StateBase* cached_exploration_state;
    std::vector<unsigned int> enabledSynchronisations; // number of (enabled) operators per sync
    std::vector<std::vector<std::list<EdgeID_type>>> enabledEdges; // enabled edges per automaton and per non-silent action label

    // action op construction sub-routines:
    void initialize_action_ops();

    void compute_enabled_synchronisations(); // compute number of enabled operators per sync
    void compute_sync_ops(ActionOpStructure& action_op_structure, SyncIndex_type sync_index, std::size_t condition_index); // recursively compute synchronisation ops

    // exploration time routines:
    void reset_exploration_structures();
    void compute_enabled_sync_ops(SyncIndex_type sync_index, const std::vector<AutomatonSyncAction>& condition, ActionOpIndex_type action_op_index, std::size_t condition_index); // recursively compute enabled synchronisation ops

    // internal auxiliaries for extraction:
    [[nodiscard]] const ActionOp& get_action_op(SyncID_type sync_id, ActionOpIndex_type action_op_index) const;
    [[nodiscard]] const ActionOp& explore_action_op(const StateBase& state, SyncID_type sync_id, ActionOpIndex_type action_op_index) const;

public:
    explicit SuccessorGeneratorC(const PLAJA::Configuration& config, const Model& model);
    ~SuccessorGeneratorC() override;
    DELETE_CONSTRUCTOR(SuccessorGeneratorC)

    void explore(const StateBase& state);
    [[nodiscard]] bool is_applicable(ActionLabel_type action_label) const; // with respect to explore

    // iterators (initialized by explore):

    using PerActionIterator = PerActionIteratorBase<const ActionOp, const ActionOp>;

    [[nodiscard]] inline PerActionIterator silentActionIterator() const { return PerActionIterator(silentOps); }

    [[nodiscard]] inline PerActionIterator labeledActionIterator(ActionLabel_type action_label) const {
        PLAJA_ASSERT(0 <= action_label && action_label <= labeledOps.size())
        return PerActionIterator(labeledOps[action_label]);
    }

    using ActionIterator = ActionLabelIteratorBase<const ActionOp, const ActionOp>;

    [[nodiscard]] inline ActionIterator actionIterator() const { return ActionIterator(silentOps, labeledOps); }

    using ActionOpIterator = ActionOpIteratorBase<const ActionOp, const ActionOp>;

    [[nodiscard]] inline ActionOpIterator actionOpIterator() const { return ActionOpIterator(silentOps, labeledOps); }

#if 0
    // all action op iterator

    class AllActionOpIterator {
        friend class SuccessorGeneratorC;
    private:
        SyncIndex_type sync_index;
        const std::vector<std::vector<std::unique_ptr<ActionOp>>>& syncOps;
        std::vector<std::unique_ptr<ActionOp>>::const_iterator it;
        std::vector<std::unique_ptr<ActionOp>>::const_iterator itEnd;
        // aux:
        void increment_until_op();
        //
        explicit AllActionOpIterator(const std::vector<std::unique_ptr<ActionOp>>& non_sync_ops, const std::vector<std::vector<std::unique_ptr<ActionOp>>>& sync_ops);
    public:
        inline void operator++() { ++it; increment_until_op(); }
        [[nodiscard]] inline bool end() const { return sync_index == syncOps.size(); }
        [[nodiscard]] inline const ActionOp* operator()() const { return it->get(); }
        [[nodiscard]] inline const ActionOp* operator->() const { return it->get(); }
        [[nodiscard]] inline const ActionOp& operator*() const { return **it; }

    };

    [[nodiscard]] inline AllActionOpIterator allActionOpIterator() const { return AllActionOpIterator(nonSyncOps, syncOps); }
#endif

    // extraction (invalidates iterators if non-const):

    [[nodiscard]] const ActionOp& get_action_op(const ActionOpStructure& action_op_structure) const;
    [[nodiscard]] const ActionOp& get_action_op(ActionOpID_type action_op_id) const;
    [[nodiscard]] std::list<const ActionOp*> extract_ops_per_label(ActionLabel_type action_label) const;

    [[nodiscard]] ActionLabel_type get_action_label(const ActionOpStructure& action_op_structure) const;
    [[nodiscard]] ActionLabel_type get_action_label(ActionOpID_type action_op_id) const;

    [[nodiscard]] const ActionOp& explore_action_op(const StateBase& state, const ActionOpStructure& action_op_structure);
    [[nodiscard]] const ActionOp& explore_action_op(const StateBase& state, ActionOpID_type action_op_id);

    [[nodiscard]] bool is_applicable(const StateBase& state, const ActionOp& action_op) const;

    using SuccessorGeneratorBase::compute_successor;
    [[nodiscard]] std::unique_ptr<State> compute_successor(const State& source, const ActionOp& action_op, UpdateIndex_type update_index) const;
    [[nodiscard]] std::unique_ptr<State> compute_successor(const State& source, ActionOpID_type action_op_id, UpdateIndex_type update_index) const;
    [[nodiscard]] std::unique_ptr<State> compute_successor(const State& source, UpdateOpID_type update_op_id) const;

    [[nodiscard]] bool is_applicable(const StateBase& state, ActionLabel_type action_label) const;
    [[nodiscard]] std::list<const ActionOp*> get_action(const StateBase& state, ActionLabel_type action_label) const;
    [[nodiscard]] std::list<const ActionOp*> explore_action(const StateBase& state, ActionLabel_type action_label);

};

/**********************************************************************************************************************/

// extern for interfacing with Marabou-related structures:
namespace SUCCESSOR_GENERATOR_C {
    [[nodiscard]] extern std::list<const ActionOp*> extract_ops_per_label(const SuccessorGeneratorC& successor_generator, ActionLabel_type action_label);
}

#endif //PLAJA_SUCCESSOR_GENERATOR_C_H
